#!/usr/bin/perl
#
#  Brian Albere
#
#  The script that controls the Array Mapper pipeline.  It handles setting up
#  the infrastructure needed by the pipeline, it controls the flow or steps
#  through the pipeline, it checks the success / failure of each step of the
#  pipeline, and it generates the report files containing the pipeline
#  statistics.
#
#  Every time a parallel process is started, the EPIMPI.c is
#  inititiated on as many cores as specified in the softwareParameters.xml.
#  Each instance of the c program will create an instance of the perl script
#  passed to the c program.  Each parallel instance of the c program and the
#  perl scripts have their own unique logs.  The log files will be named:
#
#  [whatever name or path Commander.pl passes]_
#  [the processes machine's host name]_
#  [the processes' mpi rank].log
#
#  The log files will be stored in separate folders depending on what part of
#  the pipeline the log is associated with.
#
#  Also, each sub process in the pipeline will create a status file who's name
#  is exactly the same as its log file, except the file extension will be
#  .status.  These will also be stored in separate folders depending on the part
#  of the pipeline.  These must have the first word within the files be either
#  ERROR or SUCCESS.  The next line must contain a message about the error or
#  success.  These are so subprocesses can communicate to their callers
#  whether or not they successfully acheived their goals, or if an error
#  occured.  If everything goes smoothly, the calling processes should delete
#  their sub-processes status files in order to keep clean directories.  Also,
#  on the subsequent lines of the status files, there will be statistics about
#  the parallel instances' computations.  All lines before a line saying
#  "BIOLOGY" will be concatenated into one performance report file by the
#  Commander.  All subsequent lines will be concatenated into one biology report
#  file by the Commander.
#
#  The status files will have content in the following format:
#
#  "SUCCESS" or "ERROR" followed by a '\n' and then a message on the following
#  line explaining the status
#  all lines afterwards and before the line
#  that says "BIOLOGY" will be stats about the performance of the sub-process.
#  All lines after "BIOLOGY" will be stats about the results of the sub-process.
#
#  Edit history
#  06/05/12 received from Brian.
#  06/10/12 rdb: added compile time options for execution environments:
#                local - for running small tests on a single processor 
#                inquiry - for running on unh inquiry cluster
#                c0 - for running on unhcs c0 cluster
#           Main goal was the "local" option, so can easily experiment with
#               the code and test changes.
#  06/12/12 rdb: -Added status code checking for all "system" calls and 
#                 similar methods, such as mkdir, so failure is reported.
#                -Also replaced mkdir with mkpath from File::Path package
#                -Reorganized conventions for storing output; mostly with
#                 changes in the xml files. But, also needed changes in this
#                 file to handle the Commander.log file within the new 
#                 conventions, where ALL output is in a single directory.
#                -Changed exitEarly to also write to STDERR
#                -Did some formatting changes in logging output.
#  06/13/12 rdb: -Whenever possible replaced "system" calls with equivalent
#                 Perl builtin functions (like using "unlink" instead of 
#                 system( "rm ..." )
#  06/14/12 rdb: Added commandline options for reduced time testing 
#                        -b n  : max # bases files to process
#                        -l n  : max # snp locations to process
#                        -x local | inquiry | c0  : eXecution options
#                                "local" is especially important for testing
#  06/21/12 rdb: -Completed implementation of -b and -x
#                
#
use warnings;
use strict;
use Getopt::Std;
use Time::HiRes qw( time );
use File::Path qw( mkpath remove_tree );

require( "/Users/rdb/bioinf/genotypeMapper/bin/XMLParser.pl" );

#time the entire pipeline
my $start = time();
my $end;

#variables----------------------------------------------------------------------

my $execMode = "local";    # test mode with single local processor
#my $execMode = "c0";       # UNH CS c0 cluster
#my $execMode = "inquiry";  # Inquiry cluster (and others in Discovery network)
                            # This is "normal" default

my $skipDipTest = 0;     # Don't bother with dip test, use if R isn't installed

my $parallelExec = 1; # rdb:  if 1, use mpi when appropriate

my $true = 1;
my $false = 0;
my $processTag = "COMMANDER";
my $errorTag = "ERROR";
my $successTag = "SUCCESS";
my $infoTag = "INFO";
my $bioTag = "BIOLOGY";
my $usageMessage = "
ArrayMapper version 1.0
Written by Brian Albere
University of New Hampshire
Durham, NH

Usage: ./Commander.pl [-s softwareParametersFileName] [-u userParametersFileName]

    -s fileName         An XML file designating parameters specific to the
                        infrastructure of the software.  These parameters should
                        not be available to the users of the software, the
                        software administrators should alter these parameters.
    -u fileName         An XML file designating parameters that will affect the
                        computations and output of the software.  These should
                        be provided and altered by the users of the software.
                                    
    Program Execution:
        1.  Open up the XML parameter files for input
        2.  Create folders for the rest of the pipeline to use
        3.  For each step of the pipeline
                A.  Check whether the parameters say the step should execute
                B.  Execute the step using MPI and EPIMPI
                C.  Verify the step executed successfully
                D.  Add the step's statistics to the report file
        4.  Finalize the report file
        5.  Terminate";

#global application variables
our( $opt_s, $opt_u, $opt_b, $opt_l, $opt_x );  # command line options

#global Commander variables
my $softwareParametersFile;
my $userParametersFile;
my $maxBasesFiles = 0; # max # bases files to process (for testing) 0 => no max
my $maxSnps = 0;       # max # snps to process (for testing) 0 => no max

my $outDir = "";                    # directory for all output; specified
                                    #    in userParameters.xml
my $outDirSuffix = "";              # for -b and -l options, we change name
                                    #    of output director to reflect 
                                    #    incomplete data usage
                                    
my $smoothedBasesDir = "";          # location of smoothed bases files
#other variables
my $maxNumProcesses = 1;
my $mpiMachineFile;

my $main_logFolder;
my $epimpi_logFolder;
my $spatialSmooth_logFolder;
my $homoIndex_logFolder;
my $figmdp_logFolder;
my $dipTest_logFolder;
my $finalGenotypeMapping_logFolder;

my $main_statusFolder;
my $epimpi_statusFolder;
my $spatialSmooth_statusFolder;
my $homoIndex_statusFolder;
my $figmdp_statusFolder;
my $dipTest_statusFolder;
my $finalGenotypeMapping_statusFolder;

my $execute_spatialSmooth;
my $execute_homoIndex;
my $execute_figmdp;
my $execute_dipTest;

#program execution--------------------------------------------------------------

#are there valid command line arguments?
&getopts( "s:u:l:b:x:" );
&checkUsage();
&getArguments();

print STDERR "****************************** $execMode\n";
if ( $execMode eq "local" )
{
   $parallelExec = 0;
}

#make sure the parameter parser is up and running
if( !&openParameterFiles( $softwareParametersFile, $userParametersFile ))
{
    my $error = "Could not initialize XML parameter parser with these files:
    $softwareParametersFile
    $userParametersFile";

    &printError( $error );
    exit;
}

$outDir = &uParameter( our $output_dir, our $output_dir_default );
$outDir =~ s/\/$//;   # if specif ends in /, delete it
if ( $outDir eq "." )
{
    # if cwd specified and there is a suffix, just use suffix
    $outDir = "$outDirSuffix";
}
else
{
    # append suffix to dir name
    $outDir .= "$outDirSuffix";
}

# check if output directory exists, create if it doesn't
if( !(-d $outDir ))
{
    print STDOUT "*** Creating output directory: $outDir ***\n";
    &makeFolder( $outDir );
}
#make sure log and output files are opened, and make sure 
#the folders that separate processes/steps use are generated
&initializeLogFile();
&printInfo( "ArrayMapper initializing" ) ;

#------------------------
&initializeReportFiles();
&initializeProcessFolders();

#generate the machine file for MPI that designates
#machine names and cores per machine
if ( $parallelExec )
{
   &generateMPIMachineFile();
}

#make sure that the steps of the pipeline that
#are on and off make sense
&turnOnAndOffPipelineSteps();

#make sure all the file names of the bases files
#meet the desired format
&validateBasesFileNameFormat( $maxBasesFiles );

&printInfo( "Beginning the pipeline steps" );

#THE PIPELINE STARTS HERE
&printInfo( "------ Spatial Smoothing -----------" );
&spatialSmooth( $maxBasesFiles );
&printInfo( "------ Homo Index -----------" );
&homoIndex();
&printInfo( "------ FIGMDP -----------" );
&figmdp();
&printInfo( "------ Dip Test -----------" );

if ( !$skipDipTest )
{
    &dipTest();
}
&printInfo( "------ Final Mapping -----------" );
&finalGenotypeMapping();

#end of program
&exitSuccessfully();

#-------------------------------------------------------------------------------
#
#  Look at what parts of the pipeline are turned on in the XML parameters, and
#  then make sure that all steps are turned off up to the latest step that's
#  turned off in the parameters.
#
sub turnOnAndOffPipelineSteps
{
    $execute_spatialSmooth = &uParameter( our $spatialSmooth,
                                         our $spatialSmooth_default );
    $execute_homoIndex = &uParameter( our $homoIndex,
                                     our $homoIndex_default );
    $execute_figmdp = &uParameter( our $figmdp, our $figmdp_default );
    $execute_dipTest = &uParameter( our $dipTest, our $dipTest_default );
    
    if( !$execute_dipTest  )
    {
        &printInfo( "The FinalGenotypeMapping step will be the first to execute" );
        $execute_figmdp = 0;
        $execute_homoIndex = 0;
        $execute_spatialSmooth = 0;
    }
    elsif( !$execute_figmdp )
    {
        &printInfo( "The DipTest step will be the first to execute" );
        $execute_homoIndex = 0;
        $execute_spatialSmooth = 0;
    }
    elsif( !$execute_homoIndex )
    {
        &printInfo( "The FIGMDP step will be the first to execute" );
        $execute_spatialSmooth = 0;
    }
    elsif( !$execute_spatialSmooth )
    {
        &printInfo( "The HomoIndex step will be the first to execute" );
    }
    else
    {
        &printInfo( "The SpatialSmooth step will be the first to execute" );
    }
}

#-------------------------------------------------------------------------------
#
#  Make sure the bases file names have the correct format
#
sub validateBasesFileNameFormat
{
    my ( $maxBaseFiles ) = @_;
    
    my $dir = &uParameter( our $bases_files_dir, our $bases_files_dir_default );
    my @files = glob( $dir . "/*.bases" );
    
    if ( $maxBaseFiles > 0  )
    {
        print STDERR "----- Reducing from ", scalar( @files ), 
                " to $maxBaseFiles bases files ------ \n";
        while ( $maxBaseFiles < @files )
        {
            pop( @files );
        }
    }
    #go through each bases file
    foreach my $fileName ( @files )
    {
        my @chars = split( //, $fileName );
        my $extensionTime = 0;
        my $extension = "";
        my $lastChar = "";
        
        #go through each char in the file name
        foreach my $char ( @chars )
        {
            #we've reached the file extension already
            if( $extensionTime )
            {
                $extension .= $char;
            }
            #just reached the beginning of the file extension
            elsif( $char =~ /\./ )
            {
                if( $lastChar eq "_" )
                {
                    my $error = "The char before the file extension in a bases";
                    $error .= " file name can't be a '_': $fileName";
                    &exitEarly( $error );
                }
                elsif( $lastChar eq "" )
                {
                    my $error = "There are no chars before the file extension ";
                    $error .= "in a bases file: $fileName";
                    &exitEarly( $error );
                }
                
                $extensionTime = 1;
            }
            #no spaces allowed
            elsif( $char =~ /\s/ )
            {
                my $error = "White space is not allowed in a bases file name: ";
                $error .= "$fileName";
                &exitEarly( $error );
            }
            #two '_' not allowed in a row
# rdb ----------- Why do this? 
#             elsif( $lastChar eq "_" && $char eq "_" )
#             {
#                 my $error = "Two '_' are not allowed in a row in a bases file ";
#                 $error .= "name: $fileName";
#                 &exitEarly( $error );
#             }
            
            $lastChar = $char;
        }
        
        #is the extension correct?
        if( $extension ne "bases" )
        {
            my $error = "Invalid file extension in a bases file name: ";
            $error .= "$fileName";
            &exitEarly( $error );
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Execute the SpatialSmooth step
#
sub spatialSmooth
{
    #should this step of the pipeline execute?
    if( !$execute_spatialSmooth )
    {
        return;
    }
    
    #spatialSmooth variables
    my $ss = &sParameter( our $spatialSmooth_script,
                         our $spatialSmooth_script_default );
    my $ss_log = &sParameter( our $spatialSmooth_log_file,
                             our $spatialSmooth_log_file_default );
    my $ss_status = &sParameter( our $spatialSmooth_status_file,
                                our $spatialSmooth_status_file_default );
    my $parent_allele = &uParameter( our $parent_allele_file,
                                    our $parent_allele_file_default );
    my $ndf = &uParameter( our $ndf_file, our $ndf_file_default );
    my $bases_files = &uParameter( our $bases_files_dir,
                                  our $bases_files_dir_default );
    my $x_dim = &uParameter( our $x_slide_dimensions,
                            our $x_slide_dimensions_default );
    my $y_dim = &uParameter( our $y_slide_dimensions,
                            our $y_slide_dimensions_default );
    my $zones = &uParameter( our $num_zones, our $num_zones_default );
    my $smooth_number = &uParameter( our $smooth_num, our $smooth_num_default );
    my $logInstanceFiles = &sParameter( our $logEachSSInstanceFiles,
                                       our $logEachSSInstanceFiles_default );
    my $smoothDir = &sParameter( our $spatialSmooth_output_dir,
                                 our $spatialSmooth_output_dir_default );
    my $output_dir = "$outDir/$smoothDir";
    $smoothedBasesDir = $output_dir;
    
    #remove any old results or status files
    &cleanFolder( "$main_statusFolder/$spatialSmooth_statusFolder/",
                   "*.status" ); 
    &cleanFolder( "$output_dir", "*.bases" ); 
    
    #make the spatialSmooth log file in the correct folder
    $ss_log = "$main_logFolder/$spatialSmooth_logFolder/$ss_log";
    
    #make the spatialSmooth status file in the correct folder
    $ss_status = "$main_statusFolder/$spatialSmooth_statusFolder/$ss_status";
    
    #print report info
    my $string = "
[STEP]\tSpatialSmooth ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
[PARAMETER]\t$x_dim - xDim - The x microarray slide dimension size
[PARAMETER]\t$y_dim - yDim - The y microarray slide dimension size
[PARAMETER]\t$zones - numZones - The number of zones the microarray slides are 
                        broken into during the SpatialSmooth algorithm
[PARAMETER]\t$smooth_number - smoothNumber - The smooth number used during 
                        the SpatialSmooth algorithm";
    
    &printToBioStats( $string );
    &printToPerfStats( $string );
    
    #the actual system string that will kick off the parallel processes
    my $system_string = "$ss $ss_log $ss_status "
            . "$parent_allele $ndf $bases_files "
            . "$x_dim $y_dim $zones $smooth_number "
            . "$logInstanceFiles $output_dir "
            . "$maxSnps $maxBasesFiles ";
    
    &initiateEPIMPI( "SpatialSmooth", $system_string );
}

#-------------------------------------------------------------------------------
#
#  Execute the HomoIndex step
#
sub homoIndex
{
    #should this step of the pipeline execute?
    if( !$execute_homoIndex )
    {
        return;
    }

    #homoIndex variables
    my $hi = &sParameter( our $homoIndex_script, our $homoIndex_script_default );
    my $hi_log = &sParameter( our $homoIndex_log_file,
                             our $homoIndex_log_file_default );
    my $hi_status = &sParameter( our $homoIndex_status_file,
                                our $homoIndex_status_file_default );
    my $parent_allele = &uParameter( our $parent_allele_file,
                                    our $parent_allele_file_default );
    my $hi_quality_cut = &uParameter( our $homoIndexQualityScore_cutoff,
                                     our $homoIndexQualityScore_cutoff_default );
    my $hetero_cut = &uParameter( our $hetero_cutoff,
                                 our $hetero_cutoff_default );
    my $homo_cut = &uParameter( our $homo_cutoff, our $homo_cutoff_default );
    my $hi_quality_preFilter = &uParameter( 
                               our $homoIndexQualityScore_preFilter,
                               our $homoIndexQualityScore_preFilter_default );
    my $genoOption = &uParameter( our $genotypeOption,
                                 our $genotypeOption_default );
    my $logInstanceFiles = &sParameter( our $logEachHIInstanceFiles,
                                       our $logEachHIInstanceFiles_default );
    my $homoDir = &sParameter( our $homoIndex_output_dir,
                                 our $homoIndex_output_dir_default );
    my $output_dir = "$outDir/$homoDir";
    
    #which base files should we use based on
    #whether the spatialSmooth process executed?
    my $bases_files;
    if( $execute_spatialSmooth )
    {
        $bases_files = &sDirParameter( $outDir, 
                                    our $spatialSmooth_output_dir,
                                    our $spatialSmooth_output_dir_default );
    }
    else
    {
        $bases_files = &uParameter( our $bases_files_dir,
                                   our $bases_files_dir_default );
    }
    
    #---------------------
    #remove any old results or status files
    &cleanFolder( "$main_statusFolder/$homoIndex_statusFolder", "*.status" );
    
    if( -d $output_dir )
    {
        &cleanFolder( $output_dir, "*.homoIndex" );
    }
    
    #make the HomoIndex log file in the correct folder
    $hi_log = "$main_logFolder/$homoIndex_logFolder/$hi_log";
    
    #make the HomoIndex status file in the correct folder
    $hi_status = "$main_statusFolder/$homoIndex_statusFolder/$hi_status";
    
    #print report info
    my $string = "
[STEP]\tHomoIndex ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
[PARAMETER]\t$hi_quality_cut - homoIndexQualityScoreCutoff - Any HomoIndex Quality Score greater than this number is considered invalid
[PARAMETER]\t$hetero_cut - heteroCutoff - Any HomoIndex values above this number are considered heterozygous
[PARAMETER]\t$homo_cut - homoCutoff - Any HomoIndex values below this number are considered homozygous
[PARAMETER]\t$hi_quality_preFilter - homoIndexQualityScorePreFilter - 1 ( on ), 0 ( off ), genotype mapping alleles are marked invalid if either of their two HomoIndex Quality Scores ( one for the forward DNA strand, and one for the reverse DNA strand )are invalid
[PARAMETER]\t$genoOption - genotypeOption - 1 indicates the genotype mapping alleles will be A, B, H, or NA depending on if each SNP is homozygous or heterozygous, and whether the nucleotides match up with the A and B parents. 0 indicates the genotype mapping alleles will be A, C, T, G, any combination of the two, or NA, which represent the SNP nucleotides in the forward strand, one nucleotide if the SNP is homozygous, and two nucleotides if the SNP is heterozygous";
    
    &printToBioStats( $string );
    &printToPerfStats( $string );
    
    #the actual system string that will kick off the parallel processes
    my $system_string = "$hi $hi_log $hi_status $parent_allele $bases_files ";
    $system_string .= "$hi_quality_cut $hetero_cut $homo_cut ";
    $system_string .= "$hi_quality_preFilter $genoOption ";
    $system_string .= "$logInstanceFiles $output_dir";
    
    &initiateEPIMPI( "HomoIndex", $system_string );
}
#------------------------------------------------------------------------------
# Remove old files from a directory
#    $dir - directory to be cleaned up
#    $pattern - r.e. pattern to match the files in the directory to be deleted
#               if $pattern is not defined, delete entire directory and 
#                  recreate it.
sub cleanFolder()
{
    my ( $dir, $pattern ) = @_;
    
    if ( -e $dir )
    {
        if ( defined( $pattern ))
        {
            my @files = glob( "$dir/$pattern" );
            my $nDeleted = unlink( @files );
            if ( $nDeleted != @files )
            {
                print STDERR "\n\nCommander.cleanFolder: only $nDeleted of ", 
                                scalar( @files ), " were deleted from $dir\n";
            } 
        }
        else
        {
            remove_tree( $dir, {result => \my $list} );
            if ( scalar( @$list ) > 0 )
            {
                print STDERR "Commander.cleanFolder( $dir ) failed to delete: ",
                    scalar( @$list ), " files!\n";
                # can print list with:   print "unlinked $_\n" for @$list;

            }
            else
            {
                &makeFolder( $dir );   # and recreate the folder
            }
        }
    }   
}
#-------------------------------------------------------------------------------
#
#  Execute the FIGMDP step
#
sub figmdp
{
    #should this part of the pipeline execute?
    if( !$execute_figmdp )
    {
        return;
    }

    #figmdp variables
    my $figmdp = &sParameter( our $figmdp_script, our $figmdp_script_default );
    my $figmdp_log = &sParameter( our $figmdp_log_file,
                                 our $figmdp_log_file_default );
    my $figmdp_status = &sParameter( our $figmdp_status_file,
                                    our $figmdp_status_file_default );
    my $hi_files = &sDirParameter( $outDir, 
                                our $homoIndex_output_dir,
                                our $homoIndex_output_dir_default );
    my $hetero_cut = &uParameter( our $hetero_cutoff,
                                 our $hetero_cutoff_default );
    my $homo_cut = &uParameter( our $homo_cutoff, our $homo_cutoff_default );
    my $NAs_per_snp = &uParameter( our $NAs_allowed_per_snp,
                                  our $NAs_allowed_per_snp_default );
    my $NAs_per_individual = &uParameter( 
                                       our $NAs_allowed_per_individual,
                                       our $NAs_allowed_per_individual_default );
    my $f1_filter = &uParameter( our $F1_filter, our $F1_filter_default );
    my $f1_ORorAND = &uParameter( our $F1_filter_OR_or_AND,
                                 our $F1_filter_OR_or_AND_default );
    my $parentFilter = &uParameter( our $parent_filter,
                                   our $parent_filter_default );
    my $parent_ORorAND = &uParameter( our $parent_filter_OR_or_AND,
                                     our $parent_filter_OR_or_AND_default );
    my $na_snp_filter = &uParameter( our $NA_snps_filter,
                                    our $NA_snps_filter_default );
    my $na_individual_filter = &uParameter( our $NA_individuals_filter,
                                           our $NA_individuals_filter_default );
    my $duplicate_filter = &uParameter( our $duplicates_filter,
                                        our $duplicates_filter_default );
    my $outputF1SNPs = &uParameter( our $output_F1_filter,
                                   our $output_F1_filter_default );
    my $outputParentSNPs = &uParameter( our $output_parent_filter,
                                       our $output_parent_filter_default );
    my $outputNAFilterSNPs = &uParameter( our $output_NA_snps_filter,
                                         our $output_NA_snps_filter_default );
    my $outputNAFilterIndividuals = &uParameter( 
                                     our $output_NA_individuals_filter,
                                     our $output_NA_individuals_filter_default );
    my $outputRemovedDuplicates = &uParameter( 
                                        our $output_duplicates_filter,
                                        our $output_duplicates_filter_default );
    my $figmdpDir = &sParameter( our $figmdp_output_dir,
                                 our $figmdp_output_dir_default );
    my $output_dir = "$outDir/$figmdpDir";
    
    my $genotypeMapping = &sParameter( 
                         our $post_FIGMDP_genotype_mapping_output_file,
                         our $post_FIGMDP_genotype_mapping_output_file_default );
    my $preHomoIndexMapping_F = &sParameter( 
                       our $pre_FIGMDP_homoIndex_mapping_F_output_file,
                       our $pre_FIGMDP_homoIndex_mapping_F_output_file_default );
    my $preHomoIndexMapping_R = &sParameter( 
                       our $pre_FIGMDP_homoIndex_mapping_R_output_file,
                       our $pre_FIGMDP_homoIndex_mapping_R_output_file_default );
    my $postHomoIndexMapping_F = &sParameter( 
                      our $post_FIGMDP_homoIndex_mapping_F_output_file,
                      our $post_FIGMDP_homoIndex_mapping_F_output_file_default );
    my $postHomoIndexMapping_R = &sParameter( 
                      our $post_FIGMDP_homoIndex_mapping_R_output_file,
                      our $post_FIGMDP_homoIndex_mapping_R_output_file_default );
    my $genoOption = &uParameter( our $genotypeOption,
                                 our $genotypeOption_default );
    
    #remove any old results or status files
    &cleanFolder( "$main_statusFolder/$figmdp_statusFolder", "*.status" ); 
    &cleanFolder( $output_dir ); # delete everything in $output_dir
    
    # rdb: note original code included the following:
    #    "rm $output_dir/*.R >&/dev/null";
    # This failed because there are more files than shell allows
    #   (at least with my Mac OS X) for the "rm" command (or any other)
    #   but we never knew it.
    
    #make the figmdp log file in the correct folder
    $figmdp_log = "$main_logFolder/$figmdp_logFolder/$figmdp_log";
    
    #make the figmdp status file in the correct folder
    $figmdp_status = "$main_statusFolder/$figmdp_statusFolder/$figmdp_status";
    
    #print report info
    my $string = "
[STEP]\tFIGMDP +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
[PARAMETER]\t$hetero_cut - heteroCutoff - Any HomoIndex values above this number 
[PARAMETER]\t\tare considered heterozygous
[PARAMETER]\t$homo_cut - homoCutoff - Any HomoIndex values below this number are  
[PARAMETER]\t\tconsidered homozygous
[PARAMETER]\t$NAs_per_snp - NAsPerSNP - If the NASNPFilter is on, SNPs with a  
[PARAMETER]\t\tpercentage of invalid genotype mapping alleles across all  
[PARAMETER]\t\tindividuals higher than this parameter will be removed
[PARAMETER]\t$NAs_per_individual - NAsPerIndividual - If the NAIndividualFilter  
[PARAMETER]\t\tis on, individuals with a percentage of invalid genotype mapping  
[PARAMETER]\t\talleles across all SNPs higher than this parameter will be removed
[PARAMETER]\t$f1_filter - F1Filter - 1 ( on ), 0 ( off ), SNPs with non-heterozygous  
[PARAMETER]\t\tHomoIndex values in any F1 individual are removed
[PARAMETER]\t$f1_ORorAND - F1FilterORorAND - 1 ( OR ), 0 ( AND ), OR indicates  
[PARAMETER]\t\tthe F1 filter will remove SNPs with non-heterozygous HomoIndex  
[PARAMETER]\t\tvalues in either DNA strand of any F1 individual, AND indicates  
[PARAMETER]\t\tthe F1 filter will remove SNPs with non-heterozygous HomoIndex  
[PARAMETER]\t\tvalues in both DNA strands of any F1 individual
[PARAMETER]\t$parentFilter - parentFilter - 1 ( on ), 0 ( off ), SNPs with  
[PARAMETER]\t\tincorrect genotype mapping alleles in the parent clones ( homozygous  
[PARAMETER]\t\tfrom the appropriate parent )will be removed
[PARAMETER]\t$parent_ORorAND - parentFilterORorAND - 1 ( OR ), 0 ( AND ), OR  
[PARAMETER]\t\tindicates the parent filter will remove SNPs with incorrect  
[PARAMETER]\t\tgenotype mapping alleles in either parent clone, AND indicates  
[PARAMETER]\t\tthe parent filter will remove SNPs with incorrect genotype mapping  
[PARAMETER]\t\talleles in both parent clones
[PARAMETER]\t$na_snp_filter - NASNPFilter - 1 ( on ), 0 ( off ), SNPs with a high  
[PARAMETER]\t\tpercentage of invalid genotype mapping alleles across all  
[PARAMETER]\t\tindividuals are removed
[PARAMETER]\t$na_individual_filter - NAIndividualFilter - 1 ( on ), 0 ( off ),  
[PARAMETER]\t\tindividuals with a high percentage of invalid genotype mapping  
[PARAMETER]\t\talleles across all SNPs are removed
[PARAMETER]\t$duplicate_filter - duplicatesFilter - 1 ( on ), 0 ( off ), out of  
[PARAMETER]\t\teach set of duplicate individuals, only the individual with the  
[PARAMETER]\t\tlowest percentage of invalid genotype mapping alleles is kept,  
[PARAMETER]\t\tthe others are removed, duplicate individuals are indicated by the  
[PARAMETER]\t\tsame last number in their bases file names
[PARAMETER]\t$outputF1SNPs - outputF1FilterSNPs - 1 ( on ), 0 ( off ), list the  
[PARAMETER]\t\tSNPs removed by the F1Filter in the bio report
[PARAMETER]\t$outputParentSNPs - outputParentFilterSNPs - 1 ( on ), 0 ( off ),  
[PARAMETER]\t\tlist the SNPs removed by the parentFilter in the bio report
[PARAMETER]\t$outputNAFilterSNPs - outputNASNPFilterSNPs - 1 ( on ), 0 ( off ),  
[PARAMETER]\t\tlist the SNPs removed by the NASNPFilter in the bio report
[PARAMETER]\t$outputNAFilterIndividuals - outputNAIndividualFilterIndividuals  
[PARAMETER]\t\t- 1 ( on ), 0 ( off ), list the individuals removed by the  
[PARAMETER]\t\tNAIndividualFilter in the bio report
[PARAMETER]\t$outputRemovedDuplicates - outputDuplicateFilterIndividuals  
[PARAMETER]\t\t- 1 ( on ), 0 ( off ), list the individuals removed by the  
[PARAMETER]\t\tduplicatesFilter in the bio report
[PARAMETER]\t$genoOption - genotypeOption - 1 indicates the genotype mapping  
[PARAMETER]\t\talleles will be A, B, H, or NA depending on if each SNP is  
[PARAMETER]\t\thomozygous or heterozygous, and whether the nucleotides match up  
[PARAMETER]\t\twith the A and B parents. 0 indicates the genotype mapping alleles  
[PARAMETER]\t\twill be A, C, T, G, any combination of the two, or NA, which  
[PARAMETER]\t\trepresent the SNP nucleotides in the forward strand, one  
[PARAMETER]\t\tnucleotide if the SNP is homozygous, and two nucleotides if the  
[PARAMETER]\t\tSNP is heterozygous";
    
    &printToBioStats( $string );
    &printToPerfStats( $string );
    
    #the actual system string that will kick off the process
    my $system_string = "$figmdp $figmdp_log $figmdp_status $hi_files ";
    $system_string .= "$hetero_cut $homo_cut $NAs_per_snp $NAs_per_individual ";
    $system_string .= "$f1_filter $f1_ORorAND $parentFilter $parent_ORorAND ";
    $system_string .= "$na_snp_filter ";
    $system_string .= "$na_individual_filter $duplicate_filter $outputF1SNPs ";
    $system_string .= "$outputParentSNPs $outputNAFilterSNPs ";
    $system_string .= "$outputNAFilterIndividuals $outputRemovedDuplicates ";
    $system_string .= "$output_dir $genotypeMapping $preHomoIndexMapping_F ";
    $system_string .= "$preHomoIndexMapping_R $postHomoIndexMapping_F ";
    $system_string .= "$postHomoIndexMapping_R $genoOption";
    
    &initiateEPIMPI( "FIGMDP", $system_string );
}

#-------------------------------------------------------------------------------
#
#  Execute the DipTest step
#
sub dipTest
{
    #should this part of the pipeline execute?
    if( !$execute_dipTest )
    {
        return;
    }
    
    if( !( &uParameter( our $pValue_filter, our $pValue_filter_default )))
    {
        my $info = "Not executing the DipTest step because the pValue filter ";
        $info .= "in the FinalGenotypeMapping step is not turned on";
    
        &printInfo( $info );
        return;
    }

    #dipTest variables
    my $dip = &sParameter( our $dipTest_script, our $dipTest_script_default );
    my $dip_log = &sParameter( our $dipTest_log_file,
                              our $dipTests_log_file_default );
    my $dip_status = &sParameter( our $dipTest_status_file,
                                 our $dipTest_status_file_default );
    my $R_files = &sParameter( our $figmdp_output_dir,
                              our $figmdp_output_dir_default );
    my $Rexec = &sParameter( our $Rexec_path, our $Rexec_path_default );
    my $logInstanceFiles = &sParameter( our $logEachDTInstanceFiles,
                                       our $logEachDTInstanceFiles_default );
    
    my $output_dir = "$outDir/$R_files";
    
    #remove any old results or status files
    &cleanFolder( "$main_statusFolder/$dipTest_statusFolder", "*.status" );
    &cleanFolder( $R_files );
    
    #make the dipTest log file in the correct folder
    $dip_log = "$main_logFolder/$dipTest_logFolder/$dip_log";
    
    #make the dipTest status file in the correct folder
    $dip_status = "$main_statusFolder/$dipTest_statusFolder/$dip_status";
    
    #print report info
    my $string = "
[STEP]\tDipTest ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++";
    
    &printToBioStats( $string );
    &printToPerfStats( $string );
    
    #the actual system string that will kick off the parallel processes
    my $system_string = "$dip $dip_log $dip_status $R_files $Rexec ";
    $system_string .= "$logInstanceFiles";
    
    &initiateEPIMPI( "DipTest", $system_string );
}

#-------------------------------------------------------------------------------
#
#  Execute the FinalGenotypeMapping step
#
sub finalGenotypeMapping
{
    #FGM variables
    my $figmdpDir = &sParameter( our $figmdp_output_dir,
                                our $figmdp_output_dir_default );
    $figmdpDir = "$outDir/$figmdpDir";
    
    my $fgm = &sParameter( our $finalGenotypeMapping_script,
                          our $finalGenotypeMapping_script_default );
    my $fgm_log = &sParameter( our $finalGenotypeMapping_log_file,
                              our $finalGenotypeMapping_log_file_default );
    my $fgm_status = &sParameter( our $finalGenotypeMapping_status_file,
                                 our $finalGenotypeMapping_status_file_default );
                                 
# rdb: only compute figmdpDir once
#     my $dip_files = &sParameter( our $figmdp_output_dir,
#                                 our $figmdp_output_dir_default );
#     my $orig_genotypeMapping = &sParameter( $figmdp_output_dir,
#                                            $figmdp_output_dir_default ). "/" .
#              &sParameter( our $post_FIGMDP_genotype_mapping_output_file,
#                          our $post_FIGMDP_genotype_mapping_output_file_default );
#     my $orig_hiMapping_F = &sParameter( $figmdp_output_dir,
#                                        $figmdp_output_dir_default ). "/" .
#           &sParameter( our $post_FIGMDP_homoIndex_mapping_F_output_file,
#                       our $post_FIGMDP_homoIndex_mapping_F_output_file_default );
#     my $orig_hiMapping_R = &sParameter( $figmdp_output_dir,
#                                        $figmdp_output_dir_default ). "/" .
#           &sParameter( our $post_FIGMDP_homoIndex_mapping_R_output_file,
#                       our $post_FIGMDP_homoIndex_mapping_R_output_file_default );

    my $dip_files = $figmdpDir;
    my $orig_genotypeMapping = "$figmdpDir/" .
             &sParameter( our $post_FIGMDP_genotype_mapping_output_file,
                         our $post_FIGMDP_genotype_mapping_output_file_default );

    my $orig_hiMapping_F = "$figmdpDir/" .
          &sParameter( our $post_FIGMDP_homoIndex_mapping_F_output_file,
                      our $post_FIGMDP_homoIndex_mapping_F_output_file_default );
    my $orig_hiMapping_R = "$figmdpDir/" .
          &sParameter( our $post_FIGMDP_homoIndex_mapping_R_output_file,
                      our $post_FIGMDP_homoIndex_mapping_R_output_file_default );
                      
    my $ignore = &uParameter( our $individuals_to_ignore_file,
                             our $individuals_to_ignore_file_default );
    my $pValue_filter_on = &uParameter( our $pValue_filter,
                                       our $pValue_filter_default );
    my $pValueFilterORorAND = &uParameter( our $pValue_filter_OR_or_AND,
                                          our $pValue_filter_OR_or_AND_default );
    my $outputPValueFilterSNPs = &uParameter( our $output_pValue_filter,
                                             our $output_pValue_filter_default );
    my $outputIgnoredIndividuals = &uParameter( 
                                       our $output_ignored_individuals,
                                       our $output_ignored_individuals_default );
    my $sig_pValue = &uParameter( our $significant_pValue,
                                 our $significant_pValue_default );
    my $hetero_cut = &uParameter( our $hetero_cutoff,
                                 our $hetero_cutoff_default );
    my $homo_cut = &uParameter( our $homo_cutoff, our $homo_cutoff_default );
    my $hiQualChoice = &uParameter( our $homoIndexQualityScoreChoice,
                                   our $homoIndexQualityScoreChoice_default );
    my $mapDir = &sParameter( our $finalGenotypeMapping_output_dir,
                                 our $finalGenotypeMapping_output_dir_default );
    my $final_genotypeMapping = &sParameter( 
                               our $final_genotype_mapping_output_file,
                               our $final_genotype_mapping_output_file_default );
    my $homoIndexMapping_F = &sParameter( 
                          our $final_homoIndex_mapping_F_output_file,
                          our $final_homoIndex_mapping_F_output_file_default );
    my $homoIndexMapping_R = &sParameter( 
                          our $final_homoIndex_mapping_R_output_file,
                          our $final_homoIndex_mapping_R_output_file_default );
    my $homoIndexQualScores = &sParameter( 
                      our $final_homoIndex_quality_mapping_output_file,
                      our $final_homoIndex_quality_mapping_output_file_default );
    my $genoOption = &uParameter( our $genotypeOption,
                                 our $genotypeOption_default );
    my $output_dir = "$outDir/$mapDir";
    
    #remove any old results or status files
    &cleanFolder( "$main_statusFolder/$finalGenotypeMapping_statusFolder",
                  "*.status" );
    &cleanFolder( $output_dir ); 
      
    #make the FGM log file in the correct folder
    $fgm_log = "$main_logFolder/$finalGenotypeMapping_logFolder/$fgm_log";
    
    #make the FGM status file in the correct folder
    $fgm_status = "$main_statusFolder/$finalGenotypeMapping_logFolder/$fgm_status";
    
    #print report info
    my $string = "
[STEP]\tFinalGenotypeMapping++++++++++++++++++++++++++++++++++++++++++++++++++++
[PARAMETER]\t$pValue_filter_on - pValueFilter - 1 ( on ), 0 ( off ), SNPs
           \twith invalid pValues are removed 
[PARAMETER]\t$pValueFilterORorAND - pValueFilterORorAND - 1 ( OR ), 0 ( AND ), 
           \tOR indicates the pValueFilter will remove SNPs with invalid 
           \tpValues from either DNA strand, AND indicates the pValueFilter will 
           \tremove SNPs with invalid pValue from both DNA strands
[PARAMETER]\t$outputPValueFilterSNPs - outputPValueFilterSNPs - 1 ( on ), 0 ( off ), 
           \tlist the SNPs removed by the pValueFilter in the bio report
[PARAMETER]\t$outputIgnoredIndividuals - outputIgnoredIndividuals - 1 ( on ), 0 ( off ), 
           \tlist the individuals ignored due to the inputted individualsToIgnore 
           \tfile in the bio report
[PARAMETER]\t$sig_pValue - pValueCutoff - pValues > this number are invalid
[PARAMETER]\t$hetero_cut - heteroCutoff - Any HomoIndex values above this number 
           \tare considered heterozygous
[PARAMETER]\t$homo_cut - homoCutoff - Any HomoIndex values below this number  
           \tare considered homozygous
[PARAMETER]\t$hiQualChoice - homoIndexQualityScoreChoice - 1, 2, 3, each  
           \tgenotype mapping allele has two associated HomoIndex quality scores  
           \t( one for the forward DNA strand, and one for the reverse DNA strand ),  
           \tthis parameter determines which of the two is associate with each  
           \tgenotype mapping allele: 1 indicates the worst of the two, 2  
           \tindicates the best of the two, 3 indicates the average of the two
[PARAMETER]\t$genoOption - genotypeOption - 1 indicates the genotype mapping  
           \talleles will be A, B, H, or NA depending on if each SNP is  
           \thomozygous or heterozygous, and whether the nucleotides match up  
           \twith the A and B parents. 0 indicates the genotype mapping alleles  
           \twill be A, C, T, G, any combination of the two, or NA, which  
           \trepresent the SNP nucleotides in the forward strand, one nucleotide  
           \tif the SNP is homozygous, and two nucleotides if the SNP is heterozygous";

    &printToBioStats( $string );
    &printToPerfStats( $string );
    
    #the actual system string that will kick off the process
    my $system_string = "$fgm $fgm_log $fgm_status $dip_files ";
    $system_string .= "$orig_genotypeMapping $orig_hiMapping_F ";
    $system_string .= "$orig_hiMapping_R $ignore $pValue_filter_on ";
    $system_string .= "$pValueFilterORorAND $outputPValueFilterSNPs ";
    $system_string .= "$outputIgnoredIndividuals $sig_pValue $hetero_cut ";
    $system_string .= "$homo_cut $hiQualChoice $output_dir ";
    $system_string .= "$final_genotypeMapping $homoIndexMapping_F ";
    $system_string .= "$homoIndexMapping_R $homoIndexQualScores ";
    $system_string .= "$genoOption";
    
    &initiateEPIMPI( "FinalGenotypeMapping", $system_string );
}

#-------------------------------------------------------------------------------
#
#  The function to call when you want to kick of a sub process in parallel.  You
#  pass in the name of the process, and all the arguments to the process ( such
#  as the executable name and any arguments that go to the process ).  This
#  function will handle calling EPIMPI and telling EPIMPI the subprocess to
#  kickoff in parallel.  Afterwards, a function will be called to make sure the
#  subprocesses executed successfully, and their statistics will be recorded
#
sub initiateEPIMPI
{
    my ( $processName, $args )= @_;
    
    #EPIMPI variables
    my $mpiexec = &sParameter( our $mpiexec_path, our $mpiexec_path_default );
    my $processes = &sParameter( our $mpi_processes, our $mpi_processes_default );
    my $epimpi = &sParameter( our $epimpi_executable,
                             our $epimpi_executable_default );
    my $epimpi_log = &sParameter( our $epimpi_log_file,
                                 our $epimpi_log_file_default );
    my $epimpi_status = &sParameter( our $epimpi_status_file,
                                    our $epimpi_status_file_default );
    
    #remove any old results or status files
    &cleanFolder( "$main_statusFolder/$epimpi_statusFolder", "*.status" );
    
    #make the EPIMPI log file in the correct folder
    $epimpi_log = "$main_logFolder/$epimpi_logFolder/$epimpi_log";
    
    #make the EPIMPI status file in the correct folder
    $epimpi_status = "$main_statusFolder/$epimpi_statusFolder/$epimpi_status";
    
    #TODO this is hacked in to have the FIGMDP and FinalGenotypeMapping steps
    #not in parallel, but using the framework will
    #hopefully allow changes to make them run in parallel easy
    #to make if desired in the future
    
    if( $processName eq "FIGMDP" || $processName eq "FinalGenotypeMapping"
        || !$parallelExec )
    {
        $processes = 1;
    }
    
    #make sure there are a valid number
    #of processes designated ( no decimals allowed )
    $processes = int( $processes );
    if( $processes > $maxNumProcesses || $processes < 1 )
    {
        &exitEarly( "Invalid desired number of parallel processes: $processes" );
    }
    
    #print report info
    my $string = "[PARAMETER]\t$processes - The number of parallel instances";
    
    &printToBioStats( $string );
    &printToPerfStats( $string );
    
    #the actual system string that will kick off the parallel processes
    
    my $system_string;
    # rdb added runtime test (using compile-time variables for now)
    #    to distinguish between the CS c0 cluster, the inquiry cluster,
    #    and running in test mode with no parallelism.
    if ( $execMode eq "c0" )
    {
      #the c0 version
      $system_string = "$mpiexec -n $processes -f $mpiMachineFile $epimpi ";
      $system_string .= "$epimpi_log $epimpi_status $args";
    } 
    elsif ( $execMode eq "inquiry" )
    {
      #the inquiry version
      $system_string = "$mpiexec -machinefile $mpiMachineFile -n $processes ";
      $system_string .= "$epimpi $epimpi_log $epimpi_status $args";
    }
    else # no parallel execution is to be done
    {
      $system_string = "$args";
    }
    
    #output info
    &printInfo( "Beginning the $processName step using $processes parallel instance( s )" );
    &logInfo( "System string for $processName:" );
    &logInfo( "$system_string" );
    
    #time the step, and execute it
    my $s = time();
    my $status = system $system_string;
    my $e = time();
    my $stepMinutes = ( $e - $s )/ 60;
    
    if ( $status != 0 )
    {
       print STDERR "*** system error code: $status for \n\t\t\t$system_string\n";
    }
    
    &printToPerfStats( "[RUN TIME]\t$stepMinutes - The execution time of the "
                       . "pipeline step in minutes" );
    
    #did everything finish successfully?
    #Also record the statistics
    if ( $parallelExec )  # rdb added
    {
       &processEPIMPIStatusFiles( $processName, $processes, $epimpi_status );
    }
    else
    {
       # do we need to do anything if it's local???
    }
}

#-------------------------------------------------------------------------------
#
#  Use this routine to check the EPIMPI status files after a parallel part of
#  the pipeline has completed.  If all the status files are good, the function
#  will print success and return, if anything is wrong errors will be printed
#  and the program will exit early.
#
#  Pass in the name of the process that just completed, the number of individual
#  cores that the process was spread across, and the beginning of all the status
#  files.
#
#  The format of the $status_prefix should be
#  directory/directory/beginningOfFileName, if it's not, this function will
#  probably not work, there can be any number of preceding directories, and if
#  the files are in the CWD you need ./ in the beginning
#
#  The status files will also contain statistics from all the parallel
#  processes that the Commander will concatenate into the report files
#
sub processEPIMPIStatusFiles
{
    my ( $process_name, $num_processes, $status_prefix ) = @_;
    
    &printInfo( "The $process_name step completed" );
    &printInfo( "Processing the $process_name status files" );
    
    my @prefix_parts = split( '/', $status_prefix );
    my $numParts = @prefix_parts;
    my $directory = "";
    my $file_prefix = $prefix_parts[$numParts - 1];
    
    #whats the directory containing all the status files?
    for( my $i = 0; $i < $numParts - 1; ++$i )
    {
        $directory .= "$prefix_parts[$i]/"
    }
    opendir( DIR, $directory ) or
         &exitEarly( "Could not open status folder: $directory" );
    
    my @file_names = readdir( DIR );
    my $numFiles = @file_names;
    closedir( DIR );
    
    #make sure all the status files are
    #there, make sure they all report success, and
    #record all of their statistics
    my $error_string = "";
        
    for( my $i = 0; $i < $num_processes; ++$i )
    {
        my $found = $false;
        my $status_file;
        my $hostname;
        
        #is the file there?
        #this n^2 implementation might need to change if there become
        #lots of separate processes, ( like more than 36 )but probably
        #not TODO
        for( my $j = 0; $j < $numFiles && !$found; ++$j )
        {
            if( $file_names[ $j ] =~ /$file_prefix(_)(.+)_$i.status/ )
            {
                $found = $true;
                $status_file = $file_names[ $j ];
                $hostname = $2;
            }
        }
        
        #file not found
        if( !$found )
        {
            $error_string .= "Could not find EPIMPI status file with rank $i\n";
        }
        #file found
        else
        {
            my $full_file = $directory . $status_file;
            open( STATUS, $full_file )or
            &exitEarly( "Could not open status file: $full_file" );
            
            my $line1 = <STATUS>;
            
            #empty status file
            if( !defined $line1 )
            {
                $error_string .= "Empty EPIMPI status file: hostname ";
                $error_string .= "$hostname: rank: $i\n";
                
                #delete the status file
                unlink( $full_file );
            }
            else
            {
                my $line2 = <STATUS>;
                chomp( $line2 );
                
                if( !defined $line2 )
                {
                    $error_string .= "EPIMPI status file: hostname $hostname: ";
                    $error_string .= "rank: $i did not contain a message ";
                    $error_string .= "explaining its status\n";
                    
                    #delete the status file
                    unlink( $full_file );
                }
                #error?
                elsif( $line1 =~ /^$errorTag$/ )
                {
                    $error_string .= "EPIMPI error reported from hostname ";
                    $error_string .= "$hostname: rank $i: $line2\n";
                    
                    #delete the status file
                    unlink( $full_file );
                }
                #success
                elsif( $line1 =~ /^$successTag$/ )
                {
                    #note that this doesn't do anything with the
                    #success status message on the second line of
                    #the status file
                    
                    #record all the statistics
                    &printToBioStats( 
                    "
[PARALLEL INSTANCE]\t$hostname:$i _________________________" );

                    &printToPerfStats( "
[PARALLEL INSTANCE]\t$hostname:$i _________________________" );
                    
                    #performance statistics
                    while( defined( ( $line1 = <STATUS> ))&& ( $line1 ne "$bioTag\n" ))
                    {
                        chomp( $line1 );
                        &printToPerfStats( $line1 );
                    }
                    
                    #bio statistics
                    while( $line1 = <STATUS> )
                    {
                        chomp( $line1 );
                        &printToBioStats( $line1 );
                    }
                
                    #delete the status file
                    unlink( $full_file );
                }
                #unknown
                else
                {
                    $error_string .= "Invalid status in EPIMPI status file: ";
                    $error_string .= "hostname: $hostname: rank: $i...";
                    $error_string .= "view the file\n";
                    
                    #note that this doesn't delete the EPIMPI status file
                }
            }
            
            close( STATUS );
        }
    }
    
    #if there were any errors, exit early
    if( $error_string ne "" )
    {
        $error_string .= "The $process_name step failed\n";
        $error_string .= "Quitting early due to errors";
        &exitEarly( $error_string );
    }
    else
    {
        &printInfo( "The $process_name step completed successfully" );
    }
}

#-------------------------------------------------------------------------------
#
#  Create the mpi machine file designating machines and number of cores per
#  machine based on the parameters in the software parameters file
#
sub generateMPIMachineFile
{
    my $machines = &sParameter( our $cluster_machine,
                               our $cluster_machine_default );
    
    $mpiMachineFile = &sParameter( our $mpi_machine_file,
                                  our $mpi_machine_file_default );
                                            
    &logInfo( "Generating MPI Machine File: $mpiMachineFile" );
    
    open( MPIMACHINE, ">$mpiMachineFile" )or
    &exitEarly( "Could not open mpi machine file for output: $mpiMachineFile" );
    
    $maxNumProcesses = 0;
    
    #go through the info for each machine
    for( my $i = 0; $i < scalar @{$machines}; ++$i )
    {
        my $machine = $machines -> [$i];
        my $name = $machine -> {our $cluster_machine_name};
        my $cores = $machine -> {our $cluster_machine_cores};
        
        &logInfo( "\t$name:$cores" );
        print MPIMACHINE "$name:$cores\n";
        
        #remember how many cores available
        $maxNumProcesses += $cores;
    }
    
    close( MPIMACHINE );
    &logInfo( "Successfully Generated MPI Machine File" );
}

#-------------------------------------------------------------------------------
#
#  Open and initialize the log file for appending
#
sub initializeLogFile
{
    my $log_file = &sDirParameter( $outDir, our $commander_log_file,
                               our $commander_log_file_default );
    $main_logFolder = &sDirParameter( $outDir,
                                      our $log_folder, our $log_folder_default );
    
    &checkFolder( $main_logFolder, "" );
    if ( !( -d $main_logFolder ))
    {
        &makeFolder( $main_logFolder );
    }
    open( LOG, ">>$log_file" )
       or
       &exitEarly( "Could not open Commander log file for output: $log_file" );

    #output log info about this run
    my @timeData = localtime( time );
    my $month = $timeData[4] + 1;
    my $day = $timeData[3];
    my $year = $timeData[5] + 1900;
    my $hour = $timeData[2];
    my $minute = $timeData[1];
    
    #header info
    print LOG "


-------------------Run on $month/$day/$year at $hour:$minute-------------------
    softwareParametersFileName: $softwareParametersFile
    userParametersFileName: $userParametersFile\n\n";
}

#-------------------------------------------------------------------------------
#
#  Open and initialize the two report files
#
sub initializeReportFiles
{
    my $bio_file = &sDirParameter( $outDir, our $bio_report_file,
                               our $bio_report_file_default );
    my $perf_file = &sDirParameter( $outDir, our $performance_report_file,
                                our $performance_report_file_default );
    
    if ( !( $bio_file =~ /[a-zA-Z\/.]+/ ))
    {
      &exitEarly( "Bad bio_file name: $bio_file" );
    }
    if ( !( $perf_file =~ /[a-zA-Z\/.]+/ ))
    {
      &exitEarly( "Bad bio_file name: $perf_file" );
    }
    open( BIOREPORT, ">$bio_file" )or
    &exitEarly( "Could not open bio report file for output: $bio_file" );
    
    open( PERFREPORT, ">$perf_file" )or
    &exitEarly( "Could not open performance report file for output: $perf_file" );
    
    #output log info about this run
    my @timeData = localtime( time );
    my $month = $timeData[4] + 1;
    my $day = $timeData[3];
    my $year = $timeData[5] + 1900;
    my $hour = $timeData[2];
    my $minute = $timeData[1];
    
    #header info
    print BIOREPORT "
-------------------Run on $month/$day/$year at $hour:$minute-------------------
    softwareParametersFileName: $softwareParametersFile
    userParametersFileName: $userParametersFile\n\n";
    
    print PERFREPORT "
-------------------Run on $month/$day/$year at $hour:$minute-------------------
    softwareParametersFileName: $softwareParametersFile
    userParametersFileName: $userParametersFile\n\n";
}

#-------------------------------------------------------------------------------
#
#  Add anything extra to the report files, and then close them
#
sub finalizeReportFiles
{
    $end = time();
    my $minutes = ( $end - $start ) / 60;
    
    print PERFREPORT "\n\n[RUN TIME]\t$minutes - The execution time of the "
                     . "entire pipeline in minutes\n";
    
    close( PERFREPORT );
    close( BIOREPORT );
}

#-------------------------------------------------------------------------------
#
#  This function makes sure the correct folders to store files for different
#  steps of the pipeline have been created so that the files can be
#  created and added to
#
sub initializeProcessFolders
{
    #----------------------------------------------------------------------
    #variables
    $epimpi_logFolder = &sParameter( 
                                     our $epimpi_log_folder,
                                     our $epimpi_log_folder_default );
    $spatialSmooth_logFolder = &sParameter( 
                                         our $spatialSmooth_log_folder,
                                         our $spatialSmooth_log_folder_default );
    $homoIndex_logFolder = &sParameter( 
                                        our $homoIndex_log_folder,
                                        our $homoIndex_log_folder_default );
    $figmdp_logFolder = &sParameter( 
                                    our $figmdp_log_folder,
                                    our $figmdp_log_folder_default );
    $dipTest_logFolder = &sParameter( 
                                     our $dipTest_log_folder,
                                     our $dipTest_log_folder_default );
    $finalGenotypeMapping_logFolder = &sParameter( 
                                  our $finalGenotypeMapping_log_folder,
                                  our $finalGenotypeMapping_log_folder_default );
    
    $main_statusFolder = &sDirParameter( $outDir, 
                                     our $status_folder,
                                     our $status_folder_default );
    $epimpi_statusFolder = &sParameter( 
                                       our $epimpi_status_folder,
                                       our $epimpi_status_folder_default );
    $spatialSmooth_statusFolder = &sParameter( 
                                      our $spatialSmooth_status_folder,
                                      our $spatialSmooth_status_folder_default );
    $homoIndex_statusFolder = &sParameter( 
                                            our $homoIndex_status_folder,
                                            our $homoIndex_status_folder_default );
    $figmdp_statusFolder = &sParameter( 
                                        our $figmdp_status_folder,
                                        our $figmdp_status_folder_default );
    $dipTest_statusFolder = &sParameter( 
                                        our $dipTest_status_folder,
                                        our $dipTest_status_folder_default );
    $finalGenotypeMapping_statusFolder = &sParameter( 
                               our $finalGenotypeMapping_status_folder,
                               our $finalGenotypeMapping_status_folder_default );
    
    #make sure the folders exists or create them
    &checkFolder( $main_logFolder, "" );
    &checkFolder( $main_logFolder, $epimpi_logFolder );
    &checkFolder( $main_logFolder, $spatialSmooth_logFolder );
    &checkFolder( $main_logFolder, $homoIndex_logFolder );
    &checkFolder( $main_logFolder, $figmdp_logFolder );
    &checkFolder( $main_logFolder, $dipTest_logFolder );
    &checkFolder( $main_logFolder, $finalGenotypeMapping_logFolder );
    
    &checkFolder( $main_statusFolder, "" );
    &checkFolder( $main_statusFolder, $epimpi_statusFolder );
    &checkFolder( $main_statusFolder, $spatialSmooth_statusFolder );
    &checkFolder( $main_statusFolder, $homoIndex_statusFolder );
    &checkFolder( $main_statusFolder, $figmdp_statusFolder );
    &checkFolder( $main_statusFolder, $dipTest_statusFolder );
    &checkFolder( $main_statusFolder, $finalGenotypeMapping_statusFolder );
}

#-------------------------------------------------------------------------------
#
#  Take in two parameters, when concatenated together create a whole path/name
#  to a directory, if the directory exists, do nothing, if not, create the
#  directory; also logs the action.
#
sub checkFolder
{
    my ( $prefix_folder, $folder_name )= @_;
    my $path = $prefix_folder . "/" . $folder_name;
    
    #does the folder exist?
    if( !(-d $path ))
    {
        &logInfo( "Creating $folder_name Folder: $path" );
        &makeFolder( $path );
    }
}
#-------------------------------------------------------------------------------
#
#  Given a directory path name, if it exists, do nothing, else make the
#      directory.
#
sub makeFolder
{
    my ( $path ) = @_;
    if ( -d $path )  # don't make it if it's already there
    {
        return;
    }
    mkpath( $path, {error => \my $err} );
    if ( @$err > 0 )  # $err -> array of hash tables descr all errors
    {
        # To keep it simple, we just report failure
        my $errHash = $$err[ 0 ]; # should only be one error, with 1 arg
        my $message = $$errHash{ $path };
        $message = "Unable to make directory: $path; $message";
        print STDERR( "$message\n" );
        &exitEarly( $message ); 
    }
}
#-------------------------------------------------------------------------------
#
#  Make sure the appropriate command line arguments have been passed to this
#  program, if not use defaults
#
sub checkUsage
{
    if( !defined $opt_s )
    {
        $opt_s = "softwareParameters.xml";
    }
    if ( !defined $opt_u )
    {
        $opt_u = "userParameters.xml";
    }
}

#-------------------------------------------------------------------------------
#
#  What are the command line arguments?
#
sub getArguments
{
    $softwareParametersFile = $opt_s;
    $userParametersFile = $opt_u;
    if ( defined( $opt_b ))
    {
        $maxBasesFiles = $opt_b;
        $outDirSuffix .= "-b$maxBasesFiles"; # change output directory name
    }
    if ( defined( $opt_l ))
    {
        $maxSnps = $opt_l;
        $outDirSuffix .= "-l$maxSnps"; # change output directory name
        print STDERR "----Max snp spec: $maxSnps\n";
    }
    if ( defined( $opt_x ))
    {
        $execMode = $opt_x;
    }
}

#-------------------------------------------------------------------------------
#
#  Print info messages for the user to see, also records the messages in the log
#
sub printInfo
{
    my ( $message )= @_;
    my @messageLines = split( "\n", $message );
    
    foreach my $line ( @messageLines )
    {
        print STDOUT "[$processTag: $infoTag]\t$line\n";
        print LOG "[USER $infoTag]\t$line\n";
    }
}

#-------------------------------------------------------------------------------
#
#  Print error messages for the user to see, also records the messages in the
#  log
#
sub printError
{
    my ( $message )= @_;
    my @messageLines = split( "\n", $message );
    
    foreach my $line ( @messageLines )
    {
        print STDERR "[$processTag: $errorTag]\t$line\n";
        print LOG "[USER $errorTag]\t$line\n";
    }
}

#-------------------------------------------------------------------------------
#
#  Print messages to the log
#
sub logInfo
{
    my ( $message ) = @_;
    my @messageLines = split( "\n", $message );
    
    foreach my $line ( @messageLines )
    {
        if ( defined( fileno( LOG )))
        {
            print LOG "[LOG]\t$line\n";
        }
        else
        {
            print STDERR "---[LOG not open]---\t$line\n";    
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print lines to the biology report file
#
sub printToBioStats
{
    my ( $message )= @_;
    print BIOREPORT "$message\n";
}

#-------------------------------------------------------------------------------
#
#  Print lines to the performance report file
#
sub printToPerfStats
{
    my ( $message )= @_;
    print PERFREPORT "$message\n";
}

#-------------------------------------------------------------------------------
#
#  Call this method if an error occurs and you want to exit the program early,
#  pass in a message to output, this will attempt to close any open streams, and
#  the program will terminate
#
#  Also, the error will be outputted to the log
#
sub exitEarly
{
    my ( $message )= @_;
    my @messageLines = split( "\n", $message );
    
    &finalizeReportFiles();
    
    foreach my $line ( @messageLines )
    {
        print STDERR "[$processTag: $errorTag]\t$line\n";
        print LOG "[$errorTag]\t$line\n";
    }
    
    close( DIR );
    close( STATUS );
    close( MPIMACHINE );
    close( LOG );
    
    exit;
}

#-------------------------------------------------------------------------------
#
#  Should be called right before the end of the program, closes open files
#
sub exitSuccessfully
{
    &finalizeReportFiles();
    
    my $string = "Successful termination of the pipeline";
    
    print STDOUT "[$processTag: $successTag]\t$string\n";
    print LOG "[$successTag]\t$string\n";
    close( LOG );
}
