#!/usr/bin/perl
#
#  Brian Albere
#
#  A script to be run in parallel with my MPI framework.  A simple script when
#  compared to the others, just kicks off a set of R scripts using system calls
#  to R, and makes sure they all create a results file.  The result files should
#  be created in the same directory as the R files, and each one should only
#  contain one line with the snp name, follwed by a tab, followed by the p
#  value for that snp
#
#  This script assumes that R has been installed on the current machine, and any
#  libraries that the R scripts need have been installed
#
#  NOTE - info messages are printed to STDERR for the user to see because
#  anything printed to STDOUT during MPI is not put to the screen, but STDERR
#  is...In other words, not everything printed to STDERR in this file is
#  necessarily an error...This may not be entirely true, it could be that STDOUT
#  is being buffered and everything would be printed eventually, but STDERR
#  seems to work well, so just keeping this
#
#  Edit history
#  06/05/12 received from Brian.
#  06/12/12 rdb
#           -changed exitEarly to also send error message to STDERR
#           -handle uniprocessor case, where MPI env variables aren't set

use warnings;
use strict;

#variables----------------------------------------------------------------------

my $true = 1;
my $false = 0;
my $processTag = "DIPTEST";
my $infoTag = "INFO";
my $errorTag = "ERROR";
my $successTag = "SUCCESS";
my $bioTag = "BIOLOGY";
my $usageMessage = "
Usage: DipTest.pl [logFile] [statusFile] [RFiles] [Rexec] [logProcessedFiles]

    logFile             The name of the file where log information should be
                        stored during this script, this is only the beginning of
                        the file name, the hostname and the mpi rank will be
                        added to the name along with .log
    statusFile          The name of a status file where this script will put
                        information about the overall success of its execution,
                        this is only the beginning of the file name, the
                        hostname and the mpi rank will be added to the name
                        along with .status
    RFiles              The directory containing all the R files to execute
    Rexec               The path to the R executable
    logProcessedFiles   1 (on) means log which files this parallel instance
                        handles, 0 (off) means don't do this output
                                    
    Program Execution:
        1.  Figure out which R scripts to kickoff based on what process this is
            within all the MPI processes
        2.  Tell each script to execute in R and make sure each script creates
            a results .dip file";
        
#command line arguments
my $logFile;
my $statusFile;
my $RFilesDir;
my $Rexec;
my $logProcessedFiles;

#other variables
my $rank;
my $size;
my $hostname;
my @RFiles;
my @myRFiles;
my $bioStatsString = "";
my $perfStatsString = "";
my $handleUserOutput = $false;
my $numPartsInProgressBar = 50;
my $progressPartsOutputSoFar = 0;

#program execution--------------------------------------------------------------

#are there valid command line arguments?
&checkUsage();
&getArguments();

#assumes these environment variables have been set
$rank = $ENV{MPI_RANK};
$size = $ENV{MPI_SIZE};
$hostname = $ENV{MPI_HOSTNAME};

if ( !defined( $hostname ))     
{
   $hostname = "local";   # executing locally ( not in parallel ) 
   $rank = 0;
   $size = 1;
}

#open the log file and set up a header
&initializeLogFile();

#open the status file
$statusFile = $statusFile . "_$hostname" . "_$rank.status";

if(!open(STATUS, ">$statusFile"))
{
    &printError("Could not open status file: $statusFile, quitting");
    close(LOG);
    exit;
}

#does the R files directory exist?
opendir(TEMP, $RFilesDir) or
&exitEarly("Could not open directory $RFilesDir for input");

close(TEMP);
@RFiles = glob($RFilesDir . "/*.R");

#randomize which parallel instance handles the user output
my @timeData = localtime(time);
my $day = $timeData[3];
srand($day);

my $randIndex;
if($size > scalar @RFiles)
{
    $randIndex = scalar @RFiles;
}
else
{
    $randIndex = $size;
}

my $i = int(rand($randIndex));
if($i == $rank)
{
    $handleUserOutput = $true;
    &printInfo("The parallel instance handling the user output: $rank");
}

#are there even any files to process?
if(scalar @RFiles < 1)
{
    &exitEarly("There are no R scripts to process in the input directory");
}

#is there anything this process should do?
&computeMyFiles();

if(scalar @myRFiles > 0)
{
    &processRFiles();
    &exitSuccessfully("Successfully Processed All R Files");
}
else
{
    &exitSuccessfully("No files processed, more cores available than needed");
}

#-------------------------------------------------------------------------------
#
#  Open and initialize the log file for appending
#
sub initializeLogFile
{
    $logFile = $logFile . "_$hostname" . "_$rank.log";
    
    if(!open(LOG, ">>$logFile"))
    {
        &printError("Could not open log file for output: $logFile, quitting");
        exit;
    }

    #output log info about this run
    my @timeData = localtime(time);
    my $month = $timeData[4] + 1;
    my $day = $timeData[3];
    my $year = $timeData[5] + 1900;
    my $hour = $timeData[2];
    my $minute = $timeData[1];

    #header info
    print LOG "


-------------------Run on $month/$day/$year at $hour:$minute-------------------
    logFile:\t$logFile
    statusFile:\t$statusFile
    RFiles:\t$RFilesDir
    Rexec:\t$Rexec
    logProcessedFiles:\t$logProcessedFiles\n\n";
}

#-------------------------------------------------------------------------------
#
#  Call this function to fill in the array that contains which files this
#  process should handle since there will be multiple instances of this
#  process running in parallel
#
sub computeMyFiles
{
    my $numRfiles = @RFiles;
    
    #more cores than there are files to process
    if($size > $numRfiles)
    {
        if($rank < $numRfiles)
        {
            $myRFiles[0] = $RFiles[$rank];
        }
    }
    else
    {
        my $filesPerCore = int($numRfiles / $size);
        my $startIndex = $rank * $filesPerCore;
        
        for(my $i = 0; $i < $filesPerCore; ++$i)
        {
            $myRFiles[$i] = $RFiles[$startIndex];
            ++$startIndex;
        }
        
        if($rank < $numRfiles - ($size * $filesPerCore))
        {
            $myRFiles[$filesPerCore] = $RFiles[($size * $filesPerCore) + $rank];
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Execute all the R scripts
#
sub processRFiles
{
    my $total = @RFiles;
    my $n = @myRFiles;
    
    &printInfo("Processing $total R scripts");
    &recordBioStat($n,
    "MY NUM DIP TESTS",
    "The num of dip tests executed by this parallel instance");
    &outputProgressBar();
    
    if($logProcessedFiles)
    {
        &logInfo("This parallel instance is processing the following $n R scripts");
    }
    
    for(my $i = 0; $i < $n; ++$i)
    {
        &updateProgress($n, $i);
        &processRFile($myRFiles[$i]);
    }
    
    #done with all files
    &updateProgress($n, $n);
}

#-------------------------------------------------------------------------------
#
#  Kickoff the R script with a system call, and then make sure the resulting
#  dip file was created!
#
sub processRFile
{
    my ($file_name) = @_;
    my @fileParts = split("/", $file_name);
    my $numParts = @fileParts;
    my $abrev_file = $fileParts[$numParts - 1];
    
    if($logProcessedFiles)
    {
        &logInfo("Processing $abrev_file");
    }
    
    @fileParts = split(/\./, $abrev_file);
    my $nameBeforePeriod = $RFilesDir . "/" . $fileParts[0];
    
    #A generic R call, --vanilla means don't try to load
    #up any previously used info from previous runs and
    #don't try and save any variables from this run, then
    #the R script is provided, and then I redirect the output
    #to dev null because I don't care about the output (it's
    #basically just an echoing of all the commands in
    #the inputted R script file
    system("$Rexec --vanilla --file=$file_name >/dev/null");
    
    #was the resulting dip file created?
    if(!open(EXIST, "$nameBeforePeriod.dip"))
    {
        &exitEarly("R script $file_name created no output .dip file");
    }
    else
    {
        close(EXIST);
    }
}

#-------------------------------------------------------------------------------
#
#  Output a progress bar for the user to see, only do it if this is parallel
#  instance that should handle it
#
#  Note that this bypasses the log and print methods below that the rest of the
#  script uses
#
sub outputProgressBar
{
    if($handleUserOutput)
    {
        my $bar = "0% ";
        
        for(my $i = 0; $i < $numPartsInProgressBar; ++$i)
        {
            $bar .= "_";
        }
        
        $bar .= " 100%";
        
        print STDERR "[$processTag: $infoTag]\t$bar\n";
        print STDERR "[$processTag: $infoTag]\t   ";
        $progressPartsOutputSoFar = 0;
    }
}

#-------------------------------------------------------------------------------
#
#  Update the progress bar, this decides whether to update the progress bar,
#  pass in the total number of things that need to be done, and the number of
#  things completed so far
#
#  Note that this bypasses the log and print methods below that the rest of the
#  script uses
#
sub updateProgress
{
    if($handleUserOutput)
    {
        my ($total, $completed) = @_;
        
        my $stepSize = 100 / $numPartsInProgressBar;
        my $percentage = ($completed / $total) * 100;
        
        while($progressPartsOutputSoFar * $stepSize < $percentage)
        {
            print STDERR ">";
            ++$progressPartsOutputSoFar;
        }
        
        if($completed >= $total)
        {
            print STDERR "\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Make sure the appropriate command line arguments have been passed to this
#  program, if not, print out the usage message and quit
#
sub checkUsage
{
    if(scalar @ARGV != 5)
    {
        print STDERR "$usageMessage\n\n";
        exit;
    }
}

#-------------------------------------------------------------------------------
#
#  Set variables to either their defaults or their inputted command line
#  argument
#
sub getArguments
{
    #these arguments are not optional, therefore no defaults possible
    $logFile = $ARGV[0];
    $statusFile = $ARGV[1];
    $RFilesDir = $ARGV[2];
    $Rexec = $ARGV[3];
    $logProcessedFiles = $ARGV[4];
}

#-------------------------------------------------------------------------------
#
#  Print some info to the screen for the user to see, note that only one
#  parallel instance is allowed to print info the screen, info is also recorded
#  in the log
#
sub printInfo
{
    my ($message) = @_;
    my @messageLines = split("\n", $message);
    
    foreach my $line (@messageLines)
    {
        print LOG "[USER $infoTag]\t$line\n";
        
        if($handleUserOutput)
        {
            print STDERR "[$processTag: $infoTag]\t$line\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print error messages for the user to see, also records the messages in the
#  log, note that only one parallel instance is allowed to print info to the
#  screen
#
sub printError
{
    my ($message) = @_;
    my @messageLines = split("\n", $message);
    
    foreach my $line (@messageLines)
    {
        print LOG "[USER $errorTag]\t$line\n";
        
        if($handleUserOutput)
        {
            print STDERR "[$processTag: $errorTag]\t$line\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print messages to the log
#
sub logInfo
{
    my ($message) = @_;
    my @messageLines = split("\n", $message);
    
    foreach my $line (@messageLines)
    {
        print LOG "[LOG]\t$line\n";
    }
}

#-------------------------------------------------------------------------------
#
#  Record biology statistics, pass in a value for the statistic, pass in a tag,
#  and pass in a descriptor
#
sub recordBioStat
{
    my ($value, $tag, $desc) = @_;
    
    $bioStatsString .= "[$tag]\t$value - $desc\n";
}

#-------------------------------------------------------------------------------
#
#  Record performance statistics, pass in a value for the statistic, pass in a
#  tag, and pass in a descriptor
#
sub recordPerfStat
{
    my ($value, $tag, $desc) = @_;
    
    $perfStatsString .= "[$tag]\t$value - $desc\n";
}

#-------------------------------------------------------------------------------
#
#  Exit due to an error, close all open files and update the status file with an
#  error, also output the statistics to the status file
#
sub exitEarly
{
    my ($message) = @_;
    
    print STATUS "$errorTag\n$message\n$perfStatsString";
    print STATUS "$bioTag\n$bioStatsString";
    print LOG "[$errorTag]\t$message\n";
    print STDERR "[$errorTag]\t$message\n";  # rdb added
    
    close(LOG);
    close(STATUS);
    close(TEMP);
    close(EXIST);
    
    exit;
}

#-------------------------------------------------------------------------------
#
#  Finish the script by outputting success to the log and the status files, and
#  close both files, should be the last thing called, and then the program
#  should exit correctly, statistics will also be output to the status file
#
sub exitSuccessfully
{
    my ($message) = @_;
    
    print STATUS "$successTag\n$message\n$perfStatsString";
    print STATUS "$bioTag\n$bioStatsString";
    print LOG "[$successTag]\t$message\n";
    
    close(LOG);
    close(STATUS);
    close(STATUS);
    close(EXIST);
}
