/**
 *  Brian Albere
 *
 *  Use this c program to execute perl processes in parallel through MPI.  An
 *  individual log file will be created for this process, along with an
 *  individual status file for recording the success of the program so that a
 *  calling process can see if this executed successfully.  The status file will
 *  also contain any statistics generated by the perl script in the perl
 *  script's status file
 *
 *  The command line parameters should be:
 *  thisExecutableName thisLogFile thisStatusFile perlScriptName perlLogFile
 *  perlStatusFile otherPerlArgumentsSeparatedBySpaces
 *
 *  The log files and status files passed to the program are only the file
 *  paths and the beginning of the file names (no extensions), the file prefixes
 *  will have the hostname and MPI rank added to the names along with the file
 *  extensions
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <unistd.h>
#include <errno.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>  //required header file
#include <time.h>

//forward declarations----------------------------------------------------------
void executePerl(char **argv);
void processPerlStatusFile(char *file, char *hostname, int rank);
void initializeLogFile(int argc, char **argv);
void printInfo(char *message);
void printError(char *message);
void logInfo(char *message);
void exitEarly(char *message);
void exitSuccessfully(char *message);

//constants---------------------------------------------------------------------
#define MAX_HOSTNAME_LENGTH 512
#define MAX_DIGITS_IN_MPI_RANK 3
#define MAX_CHARS_IN_LINE 256

//variables---------------------------------------------------------------------
static FILE *log = NULL;
static FILE *status = NULL;
static char *processTag = "EPIMPI";
static char *infoTag = "INFO";
static char *errorTag = "ERROR";
static char *successTag = "SUCCESS";
static char hostname[MAX_HOSTNAME_LENGTH] = "unknown";
static int rank = -1;
static int size = -1;
static int mpiInitialized = 0;

//------------------------------------------------------------------------------
/**
 *  Set up the log and status files, set up the environment variables for the
 *  perl script to use, then kick off the perl script and call the appropriate
 *  method to handle checking whether the perl script ran successfully
 */
int main(int argc, char **argv)
{
    //are there the correct number of arguments?
    if(argc < 6)
    {
        char error[MAX_CHARS_IN_LINE];
        sprintf(error, "Incorrect number of command line arguments: %d", argc);
        
        //have to print manually because log and status files not open yet
        printError(error);
        exitEarly(error);
    }

    //what's this machine's hostname
    if(gethostname(hostname, MAX_HOSTNAME_LENGTH) == -1)
    {
        //have to print manually because log and status files not open yet
        printError("Unable to access this machine's hostname");
        exitEarly("Unable to access this machine's hostname");
    }

    //initialize mpi
    MPI_Init(&argc, &argv);
    mpiInitialized = 1;
    
    // get number of processes in the process group
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    // get this process's rank (ID) within the process group
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    //open the log file
    int hnLength = strlen(hostname);
    int lfLength = strlen(argv[1]);
    char logFile[lfLength + 1 + hnLength + 1 + MAX_DIGITS_IN_MPI_RANK + 5];
    sprintf(logFile, "%s_%s_%d.log", argv[1], hostname, rank);
    log = fopen(logFile, "a");

    if(log == NULL)
    {
        char error[MAX_CHARS_IN_LINE];
        sprintf(error, "Unable to open log file for output: %s", logFile);
        
        //have to print manually because log and status files not open yet
        printError(error);
        exitEarly(error);
    }
    
    initializeLogFile(argc, argv);

    //output some log info
    char procInfo[MAX_CHARS_IN_LINE];
    char hostnameInfo[MAX_CHARS_IN_LINE];
    char rankInfo[MAX_CHARS_IN_LINE];
    
    sprintf(procInfo, "Overall number of parallel instances: %d", size);
    sprintf(hostnameInfo, "Hostname: %s", hostname);
    sprintf(rankInfo, "Rank: %d", rank);
    
    logInfo(procInfo);
    logInfo(hostnameInfo);
    logInfo(rankInfo);

    //open the status file
    int stLength = strlen(argv[2]);
    char statusFile[stLength + 1 + hnLength + 1 + MAX_DIGITS_IN_MPI_RANK + 8];
    sprintf(statusFile, "%s_%s_%d.status", argv[2], hostname, rank);
    status = fopen(statusFile, "w");
    
    if(status == NULL)
    {
        char error[MAX_CHARS_IN_LINE];
        sprintf(error, "Unable to open status file for output: %s", statusFile);
        
        //have to print manually because status file not open yet
        printError(error);
        exitEarly(error);
    }

    // insert barrier to keep all processes together (probably not needed?)
    MPI_Barrier(MPI_COMM_WORLD);
    
    //set up the environment variables for the perl script
    char buf1[9 + MAX_DIGITS_IN_MPI_RANK + 1];
    sprintf(buf1, "MPI_SIZE=%d", size);
    if(putenv(buf1))
        exitEarly("Could not generate env variable MPI_SIZE for Perl script");
    
    char buf2[9 + MAX_DIGITS_IN_MPI_RANK + 1];
    sprintf(buf2, "MPI_RANK=%d", rank);
    if(putenv(buf2))
        exitEarly("Could not generate env variable MPI_RANK for Perl script");
    
    char buf3[13 + MAX_HOSTNAME_LENGTH + 1];
    sprintf(buf3, "MPI_HOSTNAME=%s", hostname);
    if(putenv(buf3))
        exitEarly("Could not generate env variable MPI_HOSTNAME for Perl script");

    //make sure no previous instances of the perl script are running
    char *processNameBegin = strrchr(argv[3], '/');
    if(processNameBegin == NULL)
        exitEarly("Perl script file name does not contain any '/'s");
    else
    {
        ++processNameBegin;
        char buf4[MAX_CHARS_IN_LINE];
        sprintf(buf4, "killall %s >&/dev/null", processNameBegin);
        system(buf4);
    }
    
    //make sure everyone is starting their Perl script at the same time
    MPI_Barrier(MPI_COMM_WORLD);

    executePerl(argv + 3);
    
    //only the parent process should execute code from here
    //down because the child process is replaced by the 
    //Perl sub-process
    
    //see if the perl script finished successfully, if it
    //returns, than the script finished successfully
    processPerlStatusFile(argv[5], hostname, rank);
    exitSuccessfully("Successful execution of EPIMPI and its Perl script");
    return 0;
}

//------------------------------------------------------------------------------
/**
 *  Call this once ready to execute the perl script.  The name of the perl file
 *  should be the first string in the passed in array, and the rest of the array
 *  should be the parameters to the perl script
 */
void executePerl(char **argv)
{
    //see if the perl script even exists
    FILE *test = fopen(argv[0], "r");
    if(test == NULL)
        exitEarly("The passed in Perl script does not exist");
    else
        fclose(test);

    //flush the output buffers so the seperate forked
    //processes don't duplicate the info in the output files
    fflush(log);
    fflush(status);    
    
    //fork to create a copy of the current process
    pid_t pid = fork();
    
    //error in creating the child process
    if(pid == -1)
        exitEarly("Unable to create child process");
    //if child process...
    if(pid == 0)
    {
        //now start a Perl program
        execvp(argv[0], argv);
        
        //the code should only reach this point if an
        //error occured during the execvp call, just 
        //exit with -1, not by calling exitEarly
        //because this is the child process and the
        //bad exit will be seen by the parent process
        //and the error will be communicated back to the
        //Commander
        exit(-1);
    }
    else
    {
        logInfo("The Perl script was successfully started by the child process");
    
        int s;
        pid_t w;
        
        //wait for Perl program to finish
        do
        {
            w = waitpid(pid, &s, WUNTRACED | WCONTINUED);
            
            if(w == -1)
                exitEarly("The waitpid function returned an error");

            if(WIFEXITED(s))
            {
                char info[MAX_CHARS_IN_LINE];
                sprintf(info, "The child process exited with status: %d",
                        WEXITSTATUS(s));
                logInfo(info);
            }
            else if(WIFSIGNALED(s))
            {
                char info[MAX_CHARS_IN_LINE];
                sprintf(info, "The child process was killed by signal: %d",
                        WTERMSIG(s));
                logInfo(info);
            }
            else if(WIFSTOPPED(s))
            {
                char info[MAX_CHARS_IN_LINE];
                sprintf(info, "The child process was stopped by signal: %d",
                        WSTOPSIG(s));
                logInfo(info);
            }
            else if(WIFCONTINUED(s))
                logInfo("The child process continued");
        }
        while(!WIFEXITED(status) && !WIFSIGNALED(status));
    }
}

//------------------------------------------------------------------------------
/**
 *  This should be called after the perl script has ran, it checks the success
 *  of the perl script by looking at its status file, and then the status file
 *  for this process will be updated to reflect the success of the perl script
 *
 *  This will also clean up the perl status files if appropriate
 *
 *  This will also pass on the statistics from the Perl script to the status
 *  file of EPIMPI
 */
void processPerlStatusFile(char *file, char *hostname, int rank)
{
    int hnLength = strlen(hostname);
    int pstLength = strlen(file);
    char perlStatus[pstLength + 1 + hnLength + 1 + MAX_DIGITS_IN_MPI_RANK + 8];
    sprintf(perlStatus, "%s_%s_%d.status", file, hostname, rank);
    FILE *perlStatusFile = fopen(perlStatus, "r");
    
    //did the perl program create a status file?
    if(perlStatusFile == NULL)
        exitEarly("The Perl script failed to create a status file");
    else
    {
        char s[8];
        
        //empty status file
        if(fscanf(perlStatusFile, "%s\n", s) == EOF)
        {
            char *error = "The Perl script's status file was empty";
            fclose(perlStatusFile);
            remove(perlStatus);
            exitEarly(error);
        }
        //perl script succeeded
        else if(strcmp(s, successTag) == 0)
        {
            //alter the EPIMPI status file and pass on the
            //Perl script's statistics
            char successMessage[MAX_CHARS_IN_LINE];
            fgets(successMessage, MAX_CHARS_IN_LINE, perlStatusFile);
            fprintf(status, "%s\n%s", successTag, successMessage);
            
            int keepGoing = 1;
            while(keepGoing)
            {
                char line[MAX_CHARS_IN_LINE];
                if(fgets(line, MAX_CHARS_IN_LINE, perlStatusFile) == NULL)
                    keepGoing = 0;
                else
                    fprintf(status, "%s", line);
            }
            
            fclose(perlStatusFile);
            remove(perlStatus);
        }
        //perl script failed
        else if(strcmp(s, errorTag) == 0)
        {
            char line[MAX_CHARS_IN_LINE];
            fgets(line, MAX_CHARS_IN_LINE, perlStatusFile);
            char error[MAX_CHARS_IN_LINE * 2];
            sprintf(error, "Perl script error: %s", line);
            
            fclose(perlStatusFile);
            remove(perlStatus);
            exitEarly(error);
        }
        //unknown status
        else
        {
            fclose(perlStatusFile);
            exitEarly("Invalid status in Perl status file, view the file");
            //note that it doesn't delete the perl status file
        }
    }
}

//------------------------------------------------------------------------------
/**
 *  Open the log file and append information about this run.
 *
 *  This assumes that the static log FILE variable has been opened successfully
 *
 *  Also outputs all the parameters passed to the program
 */
void initializeLogFile(int argc, char **argv)
{
    time_t t;
    char *tString;
    int i;

    t = time(NULL);
    
    if(t == -1)
        tString = "Unknown Time";
    else
        tString = ctime(&t);

    int timeLength = strlen(tString);
    tString[timeLength - 1] = '\0';

    fprintf(log, "\n\n\n-----------------Run on %s-----------------\n", tString);

    for(i = 0; i < argc; ++i)
    {
        fprintf(log, "\t%s\n", argv[i]);
    }
    
    fprintf(log, "\n");
}

//------------------------------------------------------------------------------
/**
 *  Print info for the user to see, also printed to the log
 */
void printInfo(char *message)
{
    fprintf(stdout, "[%s - %s - %s: %s]\t%s\n", processTag, hostname, rank,
                                                infoTag, message);
    
    if(log != NULL)
        fprintf(log, "[USER %s]\t%s\n", infoTag, message);
}

//------------------------------------------------------------------------------
/**
 *  Print an error for the user to see, also printed to the log
 */
void printError(char *message)
{
    fprintf(stderr, "[%s - %s - %d: %s]\t%s\n", processTag, hostname, rank,
                                                errorTag, message);
    if(log != NULL)
        fprintf(log, "[USER %s]\t%s\n", errorTag, message);
}

//------------------------------------------------------------------------------
/**
 *  Print info to the log
 */
void logInfo(char *message)
{
    if(log != NULL)
        fprintf(log, "[LOG]\t%s\n", message);
}

//------------------------------------------------------------------------------
/**
 *  Call this if an error occurs that should end the program, it will print/log
 *  the passed in error message, close the log and status files, close MPI, and
 *  exit the program with an error status
 */
void exitEarly(char *message)
{
    if(status != NULL)
    {
        fprintf(status, "%s\n%s", errorTag, message);
        fclose(status);
    }
    
    if(log != NULL)
    {
        fprintf(status, "[%s]\t%s\n", errorTag, message);
        fclose(log);
    }
    
    // exit MPI
    if(mpiInitialized)
        MPI_Finalize();
    
    exit(-1);
}

//------------------------------------------------------------------------------
/**
 *  Call this right before the end of the program in order to record success in
 *  the status file and the log, and close those files, and close MPI, should be
 *  the last thing called, and then the program should exit normally
 */
void exitSuccessfully(char *message)
{
    if(status != NULL)
        fclose(status);
    
    //doesn't print anything to the status file because that should have been
    ///done when EPIMPI was examining the status file of the Perl script
    
    if(log != NULL)
    {
        fprintf(log, "[%s]\t%s\n", successTag, message);
        fclose(log);
    }
    
    // exit MPI
    if(mpiInitialized)
        MPI_Finalize();
}
