#!/usr/bin/perl
#
#  Brian Albere
#
#  Goes through all the homoIndex data for all the individuals, and filters
#  out certain sections of the data (could be entire SNPs across all the
#  individuals, entire individuals, etc.).  Currently this is not a parallel
#  process, but it runs relatively quickly so it might not need to be.  The
#  slowest part is reading in all the .homoIndex files, so if there is a 
#  dramatic increase in the amount of individuals some sort of parallelization
#  might need to occur
#
#  The output is a genotype mapping file which contains one line per individual.
#  The beginning of the line is the name of the individual and then the rest of
#  the line is tab delimited genotypes for each snp (the genotypes
#  are either H, A, B, or NA).  The first line is a header that
#  is tab delimited (with a blank in the first spot) and lists out the order
#  of the snps and their names.
#
#  The output also consists of R files for each snp in each strand, and these
#  files should be run in R by the next step to run dip tests for each snp and
#  generate p values for each snp
#
#  The HomoIndex values for each SNP and individual are output in the same table
#  format as the initial genotype mapping file
#
#  NOTE - info messages are printed to STDERR for the user to see because
#  anything printed to STDOUT during MPI is not put to the screen, but STDERR
#  is...In other words, not everything printed to STDERR in this file is
#  necessarily an error...This may not be entirely true, it could be that STDOUT
#  is being buffered and everything would be printed eventually, but STDERR
#  seems to work well, so just keeping this
#
#  NOTE - I think some of the iterations through all the data could be avoided,
#  in other words, some function are a little repetitive in finding information
#  in the data by going through it all because the information could maybe have
#  been found in a previous iteration through the data that was focused on
#  something else, however I don't think that much computation time is added and
#  I think the program is easier to understand.
#
#  Edit history
#  06/05/12 received from Brian.
#  06/12/12 rdb
#           -changed exitEarly to also send error message to STDERR
#           -handle uniprocessor case, where MPI env variables aren't set

use warnings;
use strict;

#variables----------------------------------------------------------------------

my $true = 1;
my $false = 0;
my $processTag = "FIGMDP";
my $infoTag = "INFO";
my $errorTag = "ERROR";
my $successTag = "SUCCESS";
my $bioTag = "BIOLOGY";
my $usage_message = "
Usage: FIGMDP.pl [logFile] [statusFile] [homoIndexFiles] [heteroCutoff]
                 [homoCutoff] [NAsAllowedPerSNP] [NAsAllowedPerIndividual]
                 [F1Filter] [F1FilterORorAND] [parentFilter]
                 [parentFilterORorAND] [NASNPFilter] [NAIndividualFilter]
                 [duplicateFilter] [outputF1FilterSNPs] [outputNAFilterSNPs]
                 [outputNAFilterIndividuals] [outputRemovedDuplicates]
                 [outputDir] [initialGenotypeMappingFile]
                 [preHomoIndexMapping_F] [preHomoIndexMapping_R]
                 [postHomoIndexMapping_F] [postHomoIndexMapping_R]
                 [genotypeOption]

    logFile                     The name of the file where log information
                                should be stored during this script, this is
                                only the beginning of the file name, the
                                hostname and the mpi rank will be added to the
                                name along with .log
    statusFile                  The name of a status file where this script will
                                put information about the overall success of its
                                execution, this is only the beginning of the
                                file name, the hostname and the mpi rank will be
                                added to the name along with .status
    homoIndexFiles              The directory containing the HomoIndex results
                                to read in as input
    heteroCutoff                HomoIndex values greater than or equal to this
                                value are considered heterozygous
    homoCutoff                  HomoIndex values less than or equal to this
                                value are considered homozygous
    NAsAllowedPerSNP            The percentage of invalid genotypes allowed per
                                SNP across all individuals before the entire
                                SNP is removed
    NAsAllowedPerIndividual     The percentage of invalid genotypes allowed per
                                individual across all SNPs before the entire
                                individual is removed
    F1Filter                    1 (on) means entire SNPs are removed if an F1
                                individual has a non-heterozygous HomoIndex
                                value at that SNP, 0 (off) means the filter
                                will not be applied
    F1FilterORorAND             1 (OR) means the F1 filter will remove SNPs if
                                they are non-heterozygous in either strand of
                                any F1, 0 (AND) means the F1 filter will only
                                remove SNPs if they are non-heterozygous in both
                                strands of any F1s
    parentFilter                1 (on) means entire SNPs are removed if a
                                parent clone doesnt have the correct genotype
                                mapping allele at that SNP (homozygous for that
                                parent), 0 (off) means the filter will not be
                                applied
    parentFilterORorAND         1 (OR) means the parent filter will remove
                                SNPs if they have an invalid genotype mapping
                                allele in either parent, 0 (AND) means the
                                filter will only remove snps if they have
                                invalid genotype mapping alleles in both parents
    NASNPFilter                 1 (on) means entire SNPs are removed if there
                                is too high a percentage of invalid genotypes
                                within a SNP across all individuals, 0 (off)
                                means the filter will not be applied
    NAIndividualFilter          1 (on) means entire individuals are removed if
                                there is too high a percentage of invalid
                                genotypes within an individual across all SNPs,
                                0 (off) means the filter will not be applied
    duplicatesFilter            1 (on) means only the best individual (the
                                individual with the lowest percentage of invalid
                                genotypes across all SNPs) will not be removed
                                among a set of duplicate individuals, 0 (off)
                                means this filter will not be applied
    outputF1FilterSNPs          1 (on) means output the SNP Identifiers to the
                                bio stats of every SNP that the F1 filter
                                removed, 0 (off) means don't do this output
    outputParentFilterSNPs      1 (on) means output the SNP Identifiers to the
                                bio stats of every SNP that the parent filter
                                removed, 0 (off) means don't do this output
    outputNAFilterSNPs          1 (on) means output the SNP Identifiers to the
                                bio stats of every SNP removed by the NA SNP
                                filter, 0 (off) means don't do this output
    outputNAFilterIndividuals   1 (on) means output the individuals to the bio
                                stats that are removed by the NA individuals
                                filter, 0 (off) means don't do this output
    outputRemovedDuplicates     1 (on) means output the individuals to the bio
                                stats that are removed because they are the
                                worst of a set of duplicates, 0 (off) means
                                don't do this output
    outputDir                   The directory to store the resulting genotype
                                mapping file and the R scripts
    initialGenotypeMappingFile  The name of the initial genotype mapping file
    preHomoIndexMapping_F       The name of the text file in the format of a
                                table, similar to the format of the pipeline's
                                final genotype mapping file, has all the
                                HomoIndex values of every SNP in every
                                individual in the forward strand before this
                                step's filters are applied
    preHomoIndexMapping_R       The name of the text file in the format of a
                                table, similar to the format of the pipeline's
                                final genotype mapping file, has all the
                                HomoIndex values of every SNP in every
                                individual in the reverse strand before this
                                step's filters are applied
    postHomoIndexMapping_F      The name of the text file in the format of a
                                table, similar to the format of the pipeline's
                                final genotype mapping file, has all the
                                HomoIndex values of every SNP in every
                                individual in the forward strand after this
                                step's filters are applied
    postHomoIndexMapping_R      The name of the text file in the format of a
                                table, similar to the format of the pipeline's
                                final genotype mapping file, has all the
                                HomoIndex values of every SNP in every
                                individual in the reverse strand after this
                                step's filters are applied
    genotypeOption              1 means the genotype mapping alleles will either
                                be A, B, H, or NA depending on how each SNP's
                                nucleotides match up with the parent
                                nucleotides, 0 means the genotype mapping
                                alleles will either be A, C, T, G, any
                                combination of two, or NA, which is basically
                                each SNP's nucleotides on the corresponding set
                                of homologous chromosomes, if the SNP is homo
                                the genotype will be one nucleotide, and
                                if the SNP is hetero the genotype will be two
                                nucleotides, note that the nucleotides are the
                                nucleotides in the forward strand, the reverse
                                strand nucleotides are the compliments
                                    
    Program Execution:
        1.  Read in all the .homoIndex files and store the data within them in
            this script's datastructures.
        2.  Go through all the individuals and remove any SNPs that should be
            filtered out due to various criteria.  This will remove HomoIndex
            values and genotypes (NA, A, B, H) from various datastructures
        3.  If the NAIndividualFilter is on, the script will go
            through all the individuals' genotypes and make a list of ones that
            have too many invalids.  Then the script will basically start over
            again, reading in the .homoIndex files and filtering out SNPs,
            however when the .homoIndex files are read in, any .homoIndex files
            relating to individuals we want to now ignore will not be read in or
            considered in the filters
        4.  Output the new filtered data to an intitial genotype mapping file
            and various R scripts that will be used in the next parallel step
            in the pipeline.  The R scripts will be used to run DIP tests on 
            each SNP in parallel.";
        
#command line arguments
my $logFile;
my $statusFile;
my $homoIndexDir;
my $heteroCut;
my $homoCut;
my $NAsAllowedPerSNP;
my $NAsAllowedPerIndividual;
my $F1FilterOn;
my $F1FilterORorAND;
my $parentFilter;
my $parentFilterORorAND;
my $NASNPFilterOn;
my $NAIndividualFilterOn;
my $duplicateFilterOn;
my $outputF1FilterSNPs;
my $outputParentFilterSNPs;
my $outputNAFilterSNPs;
my $outputNAFilterIndividuals;
my $outputRemovedDuplicates;
my $output_dir;
my $initialGenoMappingFile;
my $preHomoIndexMapping_F;
my $preHomoIndexMapping_R;
my $postHomoIndexMapping_F;
my $postHomoIndexMapping_R;
my $genotypeOption;

#other variables
my $rank;
my $size;
my $hostname;
my @homoIndexFiles;
my $duplicatesRemoved = 0;
my $bioStatsString = "";
my $perfStatsString = "";
my $handleUserOutput = $true;
my $numPartsInProgressBar = 50;
my $progressPartsOutputSoFar = 0;
my $nucOption = 0;
my $parentOption = 1;

#hashes

#a hash reference of hash references (2d hash), 1st hash's key is indivudal
#name, second key is snp, the value stored is the homoIndex value
my $homoIndexValues_F = {};
my $homoIndexValues_R = {};

#key is snp, the value is how many NAs there are at that snp
#across all individuals
my %naCounts = ();

#a hash reference of hash references (2d hash) 1st hash's key is individual 
#name, second key is snp, the value stored is string code representing
#whether the snp across both strands is being considered homozygote
#from one or the other parent, heterozygote, or NA
my $genotypeMapping = {};

#key is individual name, value is a dummy value, this is basically a list of
#individuals to ignore / throw out the second time through these filters if the
#NAIndividualFilter is on
my %individualsToIgnore = ();

#the key is snp, the value is a boolean, true if the snp
#across all F1s has heterozygote homoIndex values in the strand, false
#otherwise
my %f1ValidSNPs_F = ();
my %f1ValidSNPs_R = ();

#the key is snp, the value is the genotype mapping allele within the
#respective parent at that snp in the forward strand
my %parentAsnps = ();
my %parentBsnps = ();

#program execution--------------------------------------------------------------

#are there valid command line arguments?
&checkUsage();
&getArguments();

#assumes these environment variables have been set
$rank = $ENV{MPI_RANK};
$size = $ENV{MPI_SIZE};
$hostname = $ENV{MPI_HOSTNAME};

if ( !defined( $hostname ))     
{
   $hostname = "local";   # executing locally ( not in parallel ) 
   $rank = 0;
   $size = 1;
}

#open the log file and set up a header
&initializeLogFile();

#open the status file
$statusFile = $statusFile . "_$hostname" . "_$rank.status";

if(!open(STATUS, ">$statusFile"))
{
    &printError("Could not open status file: $statusFile, quitting");
    close(LOG);
    exit;
}

#does the homoIndex files directory exist?
opendir(TEMP, $homoIndexDir) or
      &exitEarly("Could not open directory $homoIndexDir for input");

close(TEMP);
@homoIndexFiles = glob($homoIndexDir . "/*.homoIndex");

#are there files to process
if(scalar @homoIndexFiles < 1)
{
    &exitEarly("There are no HomoIndex files to process in the input directory");
}

#does the output directory exist?
if(!opendir(TEMP, $output_dir))
{
    print LOG "Creating directory: $output_dir\n";
    mkdir($output_dir);
}
else
{
    close(TEMP);
}

#a hack, don't want to do the parent filter no
#matter what if the genotypes are going to be the
#SNP nucleotides rather than the parent matching
#symbols
if($genotypeOption == $nucOption)
{
    $parentFilter = $false;
}

#program flow
my $firstRunThrough = $true;
&inputHomoIndexFiles();
&filterSNPs();

if($NAIndividualFilterOn || $duplicateFilterOn)
{
    &printInfo("Removing invalid individuals and running through the filters again");
    &createListOfBadIndividuals();
    
    #are there any individuals to remove?
    if(scalar keys %individualsToIgnore > 0)
    {
        #empty everything out
        $homoIndexValues_F = {};
        $homoIndexValues_R = {};
        %naCounts = ();
        $genotypeMapping = {};
        %f1ValidSNPs_F = ();
        %f1ValidSNPs_R = ();
        %parentAsnps = ();
        %parentBsnps = ();
        
        #record some statistics
        my $num_individuals_ignored = scalar keys %individualsToIgnore;
        my $filteredIndividuals = $num_individuals_ignored - $duplicatesRemoved;
        
        &recordBioStat($num_individuals_ignored,
        "NUM INDIVIDUALS REMOVED",
        "The total number of individuals removed");
        
        &recordBioStat($filteredIndividuals,
        "NUM NA INDIVIDUALS REMOVED",
        "A subset of the total number of individuals removed, the individuals removed simply because they have a high percentage of invalid genotype mapping alleles across all SNPs, has nothing to do with duplicates");
        
        &recordBioStat($duplicatesRemoved,
        "NUM DUPLICATE INDIVIDUALS REMOVED",
        "A subset of the total number of individuals removed, the individuals removed because they were one of at least 2 duplicates, and they were not the one with the lowest percentage of invalid genotype mapping alleles");
        
        $firstRunThrough = $false;
        &inputHomoIndexFiles();
        &filterSNPs();
    }
    else
    {
        &printInfo("There are no invalid individuals to remove");
    }
}

&outputFiles();
&exitSuccessfully("Successfully processed the HomoIndex files");

#-------------------------------------------------------------------------------
#
#  Open and initialize the log file for appending
#
sub initializeLogFile
{
    $logFile = $logFile . "_$hostname" . "_$rank.log";
    
    if(!open(LOG, ">>$logFile"))
    {
        &printError("Could not open log file for output: $logFile, quitting");
        exit;
    }

    #output log info about this run
    my @timeData = localtime(time);
    my $month = $timeData[4] + 1;
    my $day = $timeData[3];
    my $year = $timeData[5] + 1900;
    my $hour = $timeData[2];
    my $minute = $timeData[1];

    #header info
    print LOG "


-------------------Run on $month/$day/$year at $hour:$minute-------------------
    logFile:\t$logFile
    statusFile:\t$statusFile
    homoIndexFiles:\t$homoIndexDir
    heteroCutoff:\t$heteroCut
    homoCutoff:\t$homoCut
    NAsAllowedPerSNP:\t$NAsAllowedPerSNP
    NAsAllowedPerIndividual:\t$NAsAllowedPerIndividual
    F1Filter:\t$F1FilterOn
    F1FilterORorAND:\t$F1FilterORorAND
    parentFilter:\t$parentFilter
    parentFilterORorAND:\t$parentFilterORorAND
    NASNPFilter:\t$NASNPFilterOn
    NAIndividualFilter:\t$NAIndividualFilterOn
    duplicateFilter:\t$duplicateFilterOn
    outputF1FilterSNPs:\t$outputF1FilterSNPs
    outputParentFilterSNPs:\t$outputParentFilterSNPs
    outputNAFilterSNPs:\t$outputNAFilterSNPs
    outputNAFilterIndividuals:\t$outputNAFilterIndividuals
    outputRemovedDuplicates:\t$outputRemovedDuplicates
    outputDir:\t$output_dir
    initialGenotypeMappingFile:\t$initialGenoMappingFile
    preHomoIndexMapping_F:\t$preHomoIndexMapping_F
    preHomoIndexMapping_R:\t$preHomoIndexMapping_R
    postHomoIndexMapping_F:\t$postHomoIndexMapping_F
    postHomoIndexMapping_R:\t$postHomoIndexMapping_R
    genotypeOption:\t$genotypeOption\n\n";
}

#-------------------------------------------------------------------------------
#
#  Go through every homoIndex file and input all the info stored in the file,
#  basically filling in all the major data structures of this script, remember
#  that each individual has 2 .homoIndex files, one for the forward strand and
#  one for the reverse strand
#
sub inputHomoIndexFiles
{
    my $num_HI_files = @homoIndexFiles;
    my $num_files_ignored = scalar keys %individualsToIgnore;
    $num_HI_files = $num_HI_files - ($num_files_ignored * 2);
    
    &printInfo("Processing $num_HI_files HomoIndex files");
    &recordBioStat($num_HI_files,
    "NUM HOMOINDEX FILES",
    "The num of HomoIndex files processed");
    
    if($num_HI_files == 0)
    {
        &exitEarly("Cannot process 0 homoIndex files");
    }
    
    &outputProgressBar();
    
    my $prevFileNumSNPs;
    my $progress = 0;
    
    #go through every homoIndex file
    foreach my $file (@homoIndexFiles)
    {
        my @fileParts = split('/', $file);
        my $abrev_file = $fileParts[(scalar @fileParts) - 1];
        
        #to get rid of the _F.homoIndex or _R.homoIndex
        $abrev_file = substr($abrev_file, 0, length($abrev_file) - 12);
        
        #only input if an individual we don't want to throw out
        if(!defined $individualsToIgnore{$abrev_file})
        {
            &updateProgress($num_HI_files, $progress);
            ++$progress;
        
            #attempt to open the file
            open(HI, $file) or
            &exitEarly("Could not open file: $file for input");
            
            #is it a forward or reverse file?
            my $f_file;
            
            if($file =~ /^.*F\.homoIndex$/)
            {
                $f_file = $true;
                $homoIndexValues_F -> {$abrev_file} = {};
            }
            else
            {
                $f_file = $false;
                $homoIndexValues_R -> {$abrev_file} = {};
            }
            
            #is it an F1 file?
            my $isF1 = $false;
            
            if($abrev_file =~ /^.*_F1.*$/)
            {
                $isF1 = $true;
            }
            
            #is it one of the parent clones
            my $isParentA = $false;
            my $isParentB = $false;
            
            if($abrev_file =~ /^.*I_clone.*$/)
            {
                $isParentA = $true;
            }
            elsif($abrev_file =~ /^.*X_clone.*$/)
            {
                $isParentB = $true;
            }
            
            #initialize the genotype mapping storage for this individual
            if(!defined $genotypeMapping -> {$abrev_file})
            {
                $genotypeMapping -> {$abrev_file} = {};
            }
            
            #go through every line of the file
            my $num_snps_in_file = 0;
            
            while(my $line = <HI>)
            {
                if($line =~ /^(\w+)\s+([ACTG]{4})\s+([0-9\.]+|[0-9\.]+e-[0-9]+)\s+(HE|HO|UK)\s+(NA|H|B|A|C|T|G|AC|AT|AG|CA|CT|CG|TA|TC|TG|GA|GC|GT)$/)
                {
                    ++$num_snps_in_file;
                    my $snp = $1;
                    my $genotypeMappingAllele = $5;
                    my $hi = $3;
                    
                    #initialize certain variables
                    if(!defined $naCounts{$snp})
                    {
                        $naCounts{$snp} = 0;
                    }
                    
                    if(!defined $f1ValidSNPs_F{$snp})
                    {
                        $f1ValidSNPs_F{$snp} = $true;
                    }
                    
                    if(!defined $f1ValidSNPs_R{$snp})
                    {
                        $f1ValidSNPs_R{$snp} = $true;
                    }
                    
                    #update the mapping value
                    if(!defined $genotypeMapping -> {$abrev_file} -> {$snp})
                    {
                        $genotypeMapping -> {$abrev_file} -> {$snp} = $genotypeMappingAllele;
                    }
                    elsif($genotypeMapping -> {$abrev_file} -> {$snp} ne $genotypeMappingAllele)
                    {
                        my $error = "A SNP's genotype in the F and R ";
                        $error .= "HomoIndex files of an individual don't ";
                        $error .= "match: $abrev_file: $snp";
                        &exitEarly($error);
                    }
                    
                    #update the parent data structures
                    if($parentFilter && $isParentA)
                    {
                        $parentAsnps{$snp} = $genotypeMappingAllele;
                    }
                    elsif($parentFilter && $isParentB)
                    {
                        $parentBsnps{$snp} = $genotypeMappingAllele;
                    }
                    
                    #really small HI value
                    if($hi =~ /([0-9\.]+)e(-[0-9]+)/)
                    {
                        $hi = $1 * (10 ** $2);
                    }
                
                    #count each SNP's NA genotypes, only do it for
                    #one strand's set of files, or else you'll
                    #double count
                    if($NASNPFilterOn && $f_file &&
                       $genotypeMappingAllele eq "NA")
                    {
                        ++$naCounts{$snp};
                    }
                    
                    #update forward strand structures
                    if($f_file)
                    {
                        $homoIndexValues_F -> {$abrev_file} -> {$snp} = $hi;
                        
                        if($F1FilterOn && $isF1 && $hi < $heteroCut)
                        {
                            $f1ValidSNPs_F{$snp} = $false;
                        }
                    }
                    #update reverse strand structures
                    else
                    {
                        $homoIndexValues_R -> {$abrev_file} -> {$snp} = $hi;
                        
                        if($F1FilterOn && $isF1 && $hi < $heteroCut)
                        {
                            $f1ValidSNPs_R{$snp} = $false;
                        }
                    }
                }
                else
                {
                    &exitEarly("Invalid line in HomoIndex file: $file: $line");
                }
            }
            
            close(HI);
            
            #quality check
            if(!defined $prevFileNumSNPs)
            {
                $prevFileNumSNPs = $num_snps_in_file;
            }
            elsif($num_snps_in_file != $prevFileNumSNPs)
            {
                my $error = "The num of SNPs in each HomoIndex file are not ";
                $error .= "equal";
                &exitEarly($error);
            }
        }
    }
    
    &updateProgress($num_HI_files, $num_HI_files);
}

#-------------------------------------------------------------------------------
#
#  Print the initial, pre-filter, HomoIndex values mapping, and filter entire
#  SNPs out of the data based on what filters are on and off
#
sub filterSNPs
{
    my @fileNames = sort(keys %$genotypeMapping);
    my $numFiles = @fileNames;
    my $firstIteration = $true;
    my $numBadF1SNPs = 0;
    my $numBadParentSNPs = 0;
    my $numNAsnps = 0;
    
    #output the pre HomoIndex Mapping files?
    if($firstRunThrough)
    {
        &printInfo("Outputting the pre-FIGMDP HomoIndex mapping files");
        &outputProgressBar();
    
        #open for output the pre HomoIndex Mapping files
        my $preHIM_F = "$output_dir/$preHomoIndexMapping_F";
        my $preHIM_R = "$output_dir/$preHomoIndexMapping_R";
        
        open(PREHIMF, ">$preHIM_F") or
        &exitEarly("Could not open file: $preHIM_F for output");
        
        open(PREHIMR, ">$preHIM_R") or
        &exitEarly("Could not open file: $preHIM_R for output");
        
        #are there any individuals
        if(scalar @fileNames < 1)
        {
            my $error = "There are 0 individuals to process in the first run ";
            $error .= "through the 'filterSNPs' function";
            &exitEarly($error);
        }
        else
        {
            my $hash = $genotypeMapping -> {$fileNames[0]};
            my @snps = sort(keys %$hash);
            
            my $header = join("\t", @snps);
            print PREHIMF "\t$header";
            print PREHIMR "\t$header";
            
            my $numFiles = @fileNames;
            
            #go through each individual
            for(my $j = 0; $j < $numFiles; ++$j)
            {
                &updateProgress($numFiles, $j);
            
                my $individual = $fileNames[$j];
            
                print PREHIMF "\n$individual";
                print PREHIMR "\n$individual";
                
                #go through each snp
                foreach my $snp (@snps)
                {
                    my $hif = $homoIndexValues_F -> {$individual} -> {$snp};
                    my $hir = $homoIndexValues_R -> {$individual} -> {$snp};
                    
                    print PREHIMF "\t$hif";
                    print PREHIMR "\t$hir";
                }
            }
            
            &updateProgress($numFiles, $numFiles);
        }
        
        close(PREHIMF);
        close(PREHIMR);
    }
    
    &printInfo("Filtering invalid SNPs out of $numFiles individuals");
    &outputProgressBar();
    
    #go through every individual
    for(my $i = 0; $i < $numFiles; ++$i)
    {
        &updateProgress($numFiles, $i);
    
        my $fileName = $fileNames[$i];
        
        my $fileHash = $genotypeMapping -> {$fileName};
        my $values_F = {};
        my $values_R = {};
        my $genotypeMappingHash = {};
        
        #go through each snp 
        foreach my $snp (keys %$fileHash)
        {
            #F1 filter
            if($F1FilterOn && (($F1FilterORorAND  && (($f1ValidSNPs_F{$snp} == $false) || 
                                                      ($f1ValidSNPs_R{$snp} == $false))) ||
                               (!$F1FilterORorAND && (($f1ValidSNPs_F{$snp} == $false) && 
                                                      ($f1ValidSNPs_R{$snp} == $false)))))
            {
                if($firstIteration == $true)
                {
                    ++$numBadF1SNPs;
                    
                    if($outputF1FilterSNPs)
                    {
                        &recordBioStat($snp,
                        "F1 FILTER SNP",
                        "A SNP removed by the F1 filter");
                    }
                }
            }
            #parent filter
            elsif($parentFilter && (($parentFilterORorAND  &&
                                     (($parentAsnps{$snp} ne "A") || 
                                      ($parentBsnps{$snp} ne "B"))) ||
                                    (!$parentFilterORorAND &&
                                     (($parentAsnps{$snp} ne "A") && 
                                      ($parentBsnps{$snp} ne "B")))))
            {
                if($firstIteration == $true)
                {
                    ++$numBadParentSNPs;
                    
                    if($outputParentFilterSNPs)
                    {
                        &recordBioStat($snp,
                        "PARENT FILTER SNP",
                        "A SNP removed by the parent filter");
                    }
                }
            }
            #too many NAs filter
            elsif($NASNPFilterOn &&
                 ($naCounts{$snp} / $numFiles >= $NAsAllowedPerSNP))
            {
                if($firstIteration == $true)
                {
                    ++$numNAsnps;
                    
                    if($outputNAFilterSNPs)
                    {
                        &recordBioStat($snp,
                        "NA FILTER SNP",
                        "A SNP removed by the NA SNP filter");
                    }
                }
            }
            #add more filters here!
            #the 'else' means the snp passed all filters
            else
            {
                $values_F -> {$snp} = $homoIndexValues_F -> {$fileName} -> {$snp};
                $values_R -> {$snp} = $homoIndexValues_R -> {$fileName} -> {$snp};
                $genotypeMappingHash -> {$snp} = $genotypeMapping -> {$fileName} -> {$snp};
            }
        }
        
        $firstIteration = $false;
        
        #the new filtered hashes
        $homoIndexValues_F -> {$fileName} = $values_F;
        $homoIndexValues_R -> {$fileName} = $values_R;
        $genotypeMapping -> {$fileName} = $genotypeMappingHash;
    }
    
    &updateProgress($numFiles, $numFiles);
    
    #record some statistics
    &recordBioStat($numBadF1SNPs, "NUM INVALID F1 SNPS",
    "The num of SNPs removed because there was a non-hetero SNP HomoIndex value in an F1 individual");
    
    &recordBioStat($numBadParentSNPs, "NUM INVALID PARENT SNPS",
    "The num of SNPs removed because they had an incorrect genotype mapping allele within the parent individuals");
    
    &recordBioStat($numNAsnps, "NUM INVALID NA SNPS",
    "The num of SNPs removed because they contained a high percentage of invalid genotype mapping alleles across all individuals");
}

#-------------------------------------------------------------------------------
#
#  Go through the initial genotype mapping  and figure out which individuals 
#  have more than some percentage (an input parameter) of NAs.  Note that the 
#  function empties the list before adding to the list
#
#  The function will also add to the list the worst of duplicate individuals
#
#  This function should only be called if the NA individuals filter or the
#  remove duplicates filter is on
#
sub createListOfBadIndividuals
{
    &printInfo("Creating list of individuals to remove");
    &outputProgressBar();

    %individualsToIgnore = ();
    my @individuals = keys %$genotypeMapping;
    my $numIndividuals = @individuals;
    my $prevNumSNPs;
    my %duplicateVals;
    my %duplicateNames;

    #go through each individual
    for(my $i = 0; $i < $numIndividuals; ++$i)
    {
        &updateProgress($numIndividuals, $i);
    
        my $individual = $individuals[$i];
        my $genotypeMappingValues = $genotypeMapping -> {$individual};
        my @snps = keys %$genotypeMappingValues;
        my $numSNPs = @snps;
        my $numNAs = 0;
        
        #quality check
        if(!defined $prevNumSNPs)
        {
            $prevNumSNPs = $numSNPs;
        }
        elsif($prevNumSNPs != $numSNPs)
        {
            my $error = "Num of SNPs in each individual not equal in the ";
            $error .= "genotype mapping data structure";
            &exitEarly($error);
        }
        
        if($numSNPs < 1)
        {
            my $error = "Num of SNPs in individuals less than 1 in 'create ";
            $error .= "list of bad individuals' function";
            &exitEarly($error);
        }
        
        #go through every snp
        foreach my $snp (@snps)
        {
            if($genotypeMappingValues -> {$snp} eq 'NA')
            {
                ++$numNAs;
            }
        }
        
        my $percentNA = $numNAs / $numSNPs;
        
        #throw it out due to too many NAs?
        if($NAIndividualFilterOn && $percentNA >= $NAsAllowedPerIndividual)
        {
            $individualsToIgnore{$individual} = "dummy";
            
            if($outputNAFilterIndividuals)
            {
                &recordBioStat($individual,
                "NA FILTER INDIVIDUAL",
                "An individual removed by the NA individual filter");
            }
        }
        #now go through the if statements for if its the worse of
        #some duplicates
        elsif($individual =~ /^.*_F1.*$/)
        {
            #an F1, wanna keep it, don't do anything
        }
        elsif($individual =~ /^.*[XI]_clone.*$/)
        {
            #a parent, wanna keep it, don't do anything
        }
        elsif($duplicateFilterOn)
        {
            my @nameParts = split("_", $individual);
            my $numParts = @nameParts;
            my $num;
            
            #whats the last number in the individual name
            if($nameParts[$numParts - 1] eq "smoothed")#if smoothing occured
            {
                $num = $nameParts[$numParts - 2];
            }
            else#if smoothing didn't occur
            {
                $num = $nameParts[$numParts - 1];
            }
        
            if(!defined $duplicateVals{$num})
            {
                $duplicateVals{$num} = $percentNA;
                $duplicateNames{$num} = $individual;
            }
            elsif($duplicateVals{$num} <= $percentNA)
            {
                $individualsToIgnore{$individual} = "dummy";
                ++$duplicatesRemoved;
                
                if($outputRemovedDuplicates)
                {
                    &recordBioStat($individual,
                    "DUPLICATE FILTER INDIVIDUAL",
                    "An individual removed because it was the worst of a set of duplicates");
                }
            }
            else
            {
                $individualsToIgnore{$duplicateNames{$num}} = "dummy";
                
                if($outputRemovedDuplicates)
                {
                    &recordBioStat($duplicateNames{$num},
                    "DUPLICATE FILTER INDIVIDUAL",
                    "An individual removed because it was the worst of a set of duplicates");
                }
                
                $duplicateVals{$num} = $percentNA;
                $duplicateNames{$num} = $individual;
                ++$duplicatesRemoved;
            }
        }
    }
    
    &updateProgress($numIndividuals, $numIndividuals);
}

#-------------------------------------------------------------------------------
#
#  Output the data / results to the appropriate files, and create the dip R
#  scripts
#
sub outputFiles
{
    &printInfo("Outputting results");
    
    my $genoMapOut = "$output_dir/$initialGenoMappingFile";
    my $him_F = "$output_dir/$postHomoIndexMapping_F";
    my $him_R = "$output_dir/$postHomoIndexMapping_R";

    #open output files
    open(MAP, ">$genoMapOut") or
    &exitEarly("Could not open file: $genoMapOut for output");
    
    open(HIMF, ">$him_F") or
    &exitEarly("Could not open file for output: $him_F");
    
    open(HIMR, ">$him_R") or
    &exitEarly("Could not open file for output: $him_R");
    
    my @individuals = sort (keys %$genotypeMapping);
    my $numIndividuals = @individuals;
    
    #are there any individuals left after all the filters?
    if($numIndividuals > 0)
    {
        my $aFileHash = $genotypeMapping -> {$individuals[0]};
        my @snps = sort (keys %$aFileHash);
        my $numSNPs = @snps;
        
        #headers
        my $header = join("\t", @snps);
        print MAP "\t$header";
        print HIMF "\t$header";
        print HIMR "\t$header";
        
        &printInfo("Outputting the initial genotype and HomoIndex mapping files");
        &outputProgressBar();
        
        #output to the map
        #go through every individual
        for(my $i = 0; $i < $numIndividuals; ++$i)
        {
            &updateProgress($numIndividuals, $i);
        
            my $file = $individuals[$i];
            print MAP "\n$file";
            print HIMF "\n$file";
            print HIMR "\n$file";
            
            #print out all the values
            foreach my $snp (@snps)
            {
                my $genotypeMappingVal = $genotypeMapping -> {$file} -> {$snp};
                my $hi_F = $homoIndexValues_F -> {$file} -> {$snp};
                my $hi_R = $homoIndexValues_R -> {$file} -> {$snp};
                
                print MAP "\t$genotypeMappingVal";
                print HIMF "\t$hi_F";
                print HIMR "\t$hi_R";
            }
        }
        
        &updateProgress($numIndividuals, $numIndividuals);
        close(MAP);
        close(HIMF);
        close(HIMR);
        
        &printInfo("Outputting the R scripts");
        &outputProgressBar();
        
        #output to the R scripts
        #go through each snp
        for(my $i = 0; $i < $numSNPs; ++$i)
        {
            &updateProgress($numSNPs, $i);
            
            my $snp = $snps[$i];
        
            my $file_F = ">$output_dir/$snp" . "_F.R";
            my $file_R = ">$output_dir/$snp" . "_R.R";
            
            open(RFILEF, $file_F) or
            &exitEarly("Could not open $file_F for output");
            
            open(RFILER, $file_R) or
            &exitEarly("Could not open $file_R for output");
            
            print RFILEF "library(\"diptest\")\n$snp<-c(";
            print RFILER "library(\"diptest\")\n$snp<-c(";
            
            my $F_string = "";
            my $R_string = "";
            
            #go through each individual
            for(my $i = 0; $i < $numIndividuals; ++$i)
            {
                my $file = $individuals[$i];
                
                #we don't want the parent or F1 SNP
                #data to be used in the F2 dip tests
                if(!($file =~ /^.*(_F1|X_clone|I_clone).*$/))
                {
                    my $hi_F = $homoIndexValues_F -> {$file} -> {$snp};
                    my $hi_R = $homoIndexValues_R -> {$file} -> {$snp};
                    
                    #ignore the NA genotypes
                    if($genotypeMapping -> {$file} -> {$snp} ne "NA")
                    {
                        $F_string .= "$hi_F, ";
                        $R_string .= "$hi_R, ";
                    }
                }
            }
            
            #chop off the last ','
            my $F_length = length($F_string);
            my $R_length = length($R_string);
            
            if($F_length > 2)
            {
                $F_string = substr($F_string, 0, $F_length - 2);
            }
            else
            {
                &exitEarly("No non-NA values to run dip test on in F strand of snp: $snp");
            }
            
            if($R_length > 2)
            {
                $R_string = substr($R_string, 0, $R_length - 2);
            }
            else
            {
                &exitEarly("No non-NA values to run dip test on in R strand of snp: $snp");
            }
            
            #the rest of the R commands, assumes the command
            #'install.packages("diptest")' has been executed
            #on all the machines that will run R
            print RFILEF $F_string;
            print RFILER $R_string;
            
            print RFILEF ")\n$snp", "_diptest<-dip.test($snp)\n";
            print RFILER ")\n$snp", "_diptest<-dip.test($snp)\n";
            
            print RFILEF "$snp", "_p<-c(\"$snp\", ";
            print RFILER "$snp", "_p<-c(\"$snp\", ";
            
            print RFILEF "$snp", "_diptest", '$p.value)', "\n";
            print RFILER "$snp", "_diptest", '$p.value)', "\n";
            
            my $out_F = "$output_dir/$snp" . "_F.dip";
            my $out_R = "$output_dir/$snp" . "_R.dip";
            
            print RFILEF "write($snp", "_p, file=\"$out_F\", ";
            print RFILER "write($snp", "_p, file=\"$out_R\", ";
            
            print RFILEF "append=TRUE, ncolumns=2, sep=\"\\t\")";
            print RFILER "append=TRUE, ncolumns=2, sep=\"\\t\")";
            
            close(RFILEF);
            close(RFILER);
        }
        
        &updateProgress($numSNPs, $numSNPs);
    }
    else
    {
        &exitEarly("No individuals to output");
    }
}

#-------------------------------------------------------------------------------
#
#  Output a progress bar for the user to see, only do it if this is parallel
#  instance that should handle it
#
#  Note that this bypasses the log and print methods below that the rest of the
#  script uses
#
sub outputProgressBar
{
    if($handleUserOutput)
    {
        my $bar = "0% ";
        
        for(my $i = 0; $i < $numPartsInProgressBar; ++$i)
        {
            $bar .= "_";
        }
        
        $bar .= " 100%";
        
        print STDERR "[$processTag: $infoTag]\t$bar\n";
        print STDERR "[$processTag: $infoTag]\t   ";
        $progressPartsOutputSoFar = 0;
    }
}

#-------------------------------------------------------------------------------
#
#  Update the progress bar, this decides whether to update the progress bar,
#  pass in the total number of things that need to be done, and the number of
#  things completed so far
#
#  Note that this bypasses the log and print methods below that the rest of the
#  script uses
#
sub updateProgress
{
    if($handleUserOutput)
    {
        my ($total, $completed) = @_;
        
        my $stepSize = 100 / $numPartsInProgressBar;
        my $percentage = ($completed / $total) * 100;
        
        while($progressPartsOutputSoFar * $stepSize < $percentage)
        {
            print STDERR ">";
            ++$progressPartsOutputSoFar;
        }
        
        if($completed >= $total)
        {
            print STDERR "\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Make sure the appropriate command line arguments have been passed to this
#  program, if not, print out the usage message and quit
#
sub checkUsage
{
    if(scalar @ARGV != 26)
    {
        print STDERR "$usage_message\n\n";
        exit;
    }
}

#-------------------------------------------------------------------------------
#
#  Set variables to either their defaults or their inputted command line
#  argument
#
sub getArguments
{
    #these arguments are not optional, therefore no defaults possible
    $logFile = $ARGV[0];
    $statusFile = $ARGV[1];
    $homoIndexDir = $ARGV[2];
    $heteroCut = $ARGV[3];
    $homoCut = $ARGV[4];
    $NAsAllowedPerSNP = $ARGV[5];
    $NAsAllowedPerIndividual = $ARGV[6];
    $F1FilterOn = $ARGV[7];
    $F1FilterORorAND = $ARGV[8];
    $parentFilter = $ARGV[9];
    $parentFilterORorAND = $ARGV[10];
    $NASNPFilterOn = $ARGV[11];
    $NAIndividualFilterOn = $ARGV[12];
    $duplicateFilterOn = $ARGV[13];
    $outputF1FilterSNPs = $ARGV[14];
    $outputParentFilterSNPs = $ARGV[15];
    $outputNAFilterSNPs = $ARGV[16];
    $outputNAFilterIndividuals = $ARGV[17];
    $outputRemovedDuplicates = $ARGV[18];
    $output_dir = $ARGV[19];
    $initialGenoMappingFile = $ARGV[20];
    $preHomoIndexMapping_F = $ARGV[21];
    $preHomoIndexMapping_R = $ARGV[22];
    $postHomoIndexMapping_F = $ARGV[23];
    $postHomoIndexMapping_R = $ARGV[24];
    $genotypeOption = $ARGV[25];
}

#-------------------------------------------------------------------------------
#
#  Print some info to the screen for the user to see, note that only one
#  parallel instance is allowed to print info the screen, info is also recorded
#  in the log
#
sub printInfo
{
    my ($message) = @_;
    my @messageLines = split("\n", $message);
    
    foreach my $line (@messageLines)
    {
        print LOG "[USER $infoTag]\t$line\n";
        
        if($handleUserOutput)
        {
            print STDERR "[$processTag: $infoTag]\t$line\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print error messages for the user to see, also records the messages in the
#  log, note that only one parallel instance is allowed to print info to the
#  screen
#
sub printError
{
    my ($message) = @_;
    my @messageLines = split("\n", $message);
    
    foreach my $line (@messageLines)
    {
        print LOG "[USER $errorTag]\t$line\n";
        
        if($handleUserOutput)
        {
            print STDERR "[$processTag: $errorTag]\t$line\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print messages to the log
#
sub logInfo
{
    my ($message) = @_;
    my @messageLines = split("\n", $message);
    
    foreach my $line (@messageLines)
    {
        print LOG "[LOG]\t$line\n";
    }
}

#-------------------------------------------------------------------------------
#
#  Record biology statistics, pass in a value for the statistic, pass in a tag,
#  and pass in a descriptor
#
sub recordBioStat
{
    my ($value, $tag, $desc) = @_;
    
    $bioStatsString .= "[$tag]\t$value - $desc\n";
}

#-------------------------------------------------------------------------------
#
#  Record performance statistics, pass in a value for the statistic, pass in a
#  tag, and pass in a descriptor
#
sub recordPerfStat
{
    my ($value, $tag, $desc) = @_;
    
    $perfStatsString .= "[$tag]\t$value - $desc\n";
}

#-------------------------------------------------------------------------------
#
#  Exit due to an error, close all open files and update the status file with an
#  error, also output the statistics to the status file
#
sub exitEarly
{
    my ($message) = @_;
    
    print STATUS "$errorTag\n$message\n$perfStatsString";
    print STATUS "$bioTag\n$bioStatsString";
    print LOG "[$errorTag]\t$message\n";
    print STDERR "[$errorTag]\t$message\n";  # rdb added
    
    close(LOG);
    close(STATUS);
    close(TEMP);
    close(HI);
    close(RFILEF);
    close(RFILER);
    close(MAP);
    close(PREHIMF);
    close(PREHIMR);
    
    exit;
}

#-------------------------------------------------------------------------------
#
#  Finish the script by outputting success to the log and the status files, and
#  close both files, should be the last thing called, and then the program
#  should exit correctly, statistics will also be output to the status file
#
sub exitSuccessfully
{
    my ($message) = @_;
    
    print STATUS "$successTag\n$message\n$perfStatsString";
    print STATUS "$bioTag\n$bioStatsString";
    print LOG "[$successTag]\t$message\n";
    
    close(LOG);
    close(STATUS);
    close(TEMP);
    close(HI);
    close(RFILEF);
    close(RFILER);
    close(MAP);
    close(PREHIMF);
    close(PREHIMR);
}
