#!/usr/bin/perl
#
#  Brian Albere
#
#  A script that is NOT run in parallel.  Takes in an original genotype mapping
#  and a directory of dip result files.  All snps will be removed from the
#  mapping that don't have significant p values in both strands.
#
#  The script also takes in a file listing individuals to ignore and remove
#  from the genotype mapping
#
#  Output is just a new genotype mapping in the same format as the original that
#  was created in previous steps
#
#  NOTE - info messages are printed to STDERR for the user to see because
#  anything printed to STDOUT during MPI is not put to the screen, but STDERR
#  is...In other words, not everything printed to STDERR in this file is
#  necessarily an error...This may not be entirely true, it could be that STDOUT
#  is being buffered and everything would be printed eventually, but STDERR
#  seems to work well, so just keeping this
#
#  Edit history
#  06/05/12 received from Brian.
#  06/12/12 rdb
#           -changed exitEarly to also send error message to STDERR
#           -handle uniprocessor case, where MPI env variables aren't set

use warnings;
use strict;
use List::Util qw(sum ) ;

#variables----------------------------------------------------------------------

my $true = 1;
my $false = 0;
my $processTag = "FINALGENOTYPEMAPPING";
my $infoTag = "INFO";
my $errorTag = "ERROR";
my $successTag = "SUCCESS";
my $bioTag = "BIOLOGY";
my $usageMessage = "
Usage: FinalGenotypeMapping.pl [logFile] [statusFile] [dipFiles]
                               [originalGenotypeMappingFile]
                               [originalHomoIndexMappingFile_F]
                               [originalHomoIndexMappingFile_R]
                               [individualsToIgnore] [pValuesFilter]
                               [pValueFilterORorAND] [outputPValueFilterSNPs]
                               [outputIgnoredIndividuals] [pValueCutoff]
                               [heteroCutoff] [homoCutoff]
                               [homoIndexQualityChoice] [outputDir]
                               [finalGenotypeMappingFile] [homoIndexMapping_F]
                               [homoIndexMapping_R] [homoIndexQualityMapping]
                               [genotypeOption]

    logFile                         The name of the file where log information
                                    should be stored during this script, this is
                                    only the beginning of the file name, the
                                    hostname and the mpi rank will be added to
                                    the name along with .log
    statusFile                      The name of a status file where this script
                                    will put information about the overall
                                    success of its execution, this is
                                    only the beginning of the file name, the
                                    hostname and the mpi rank will be added to
                                    the name along with .status
    dipFiles                        The directory containing all the dip result
                                    files to read in
    originalGenotypeMappingFile     The genotype mapping file before the dip
                                    analysis was run
    originalHomoIndexMappingFile_F  The HomoIndex mapping file from the FIGMDP
                                    step for the forward strand
    originalHomoIndexMappingFile_R  The HomoIndex mapping file from the FIGMDP
                                    step for the reverse strand
    individualsToIgnore             A list of individuals to remove from the
                                    final genotype mapping (F1's, or maybe the
    pValueFilter                    1 (on )  means SNPs will be removed if they
                                    have invalid pValues computed by the
                                    previous pipeline step, 0 (off )  means the
                                    pValues will not be considered
    pValueFilterORorAND             1 (OR )  means the pValue filter will remove
                                    SNPs if they have bad pValues in either
                                    strand, 0 (AND )  means the pValue filter
                                    will only remove SNPs if they have bad
                                    pValues in both strands
    outputPValueFilterSNPs          1 (on )  means output the SNP Identifiers 
                                    removed by the pValue filter to the bio
                                    stats, 0 (off )  means don't do this output
    outputIgnoredIndividuals        1 (on )  means output the individuals
                                    removed because they were listed in the
                                    ignore individuals file to the bio stats, 0
                                    (off )  means don't do this output
    pValueCutoff                    pValues greater than this number are
                                    considered invalid
    heteroCutoff                    HomoIndex values above this number are
                                    considered heterozygous
    homoCutoff                      HomoIndex values below this number are
                                    considered homozygous
    homoIndexQualityChoice          Each genotype mapping allele has two
                                    associated HomoIndex values for the two
                                    DNA strands, and therefore two associated
                                    HomoIndex quality scores.  This parameter
                                    determines which of the two is used as the
                                    HomoIndex quality score for each genotype
                                    mapping allele.  Three possible values
                                    (1, 2, or 3 ) , 1 means use the worst of the
                                    two, 2 means use the best of the two, and 3
                                    means use the average of the two
    outputDir                       The directory to store the resulting
                                    final genotype mapping files
    finalGenotypeMappingFile        The name of the new genotype mapping file
    homoIndexMapping_F              The name of the text file in the format of a
                                    table, similar to the format of the
                                    pipeline's final genotype mapping file, has
                                    all the HomoIndex values of every SNP in
                                    every individual in the forward strand after
                                    the final filters are applied
    homoIndexMapping_R              The name of the text file in the format of a
                                    table, similar to the format of the
                                    pipeline's final genotype mapping file, has
                                    all the HomoIndex values of every SNP in
                                    every individual in the reverse strand after
                                    the final filters are applied
    homoIndexQualityMapping         The name of the text file in the format of a
                                    table that has all the homoIndex quality
                                    scores for each individual and each SNP in
                                    the final genotype mapping.
    genotypeOption                  1 means the genotype mapping alleles will
                                    either be A, B, H, or NA depending on how
                                    each SNP's nucleotides match up with the
                                    parent nucleotides, 0 means the genotype
                                    mapping alleles will either be A, C, T, G,
                                    any combination of two, or NA, which is
                                    basically each SNP's nucleotides on the
                                    corresponding set of homologous chromosomes,
                                    if the SNP is homo the genotype will be one
                                    nucleotide, and if the SNP is hetero the
                                    genotype will be two nucleotides, note that
                                    the nucleotides are the nucleotides in the
                                    forward strand, the reverse
                                    strand nucleotides are the compliments
                 
    Program Execution:
        1.  Read in all the dip result files, and store the p values of each
            snp in the forward and reverse strands
        2.  Go through every individual of the original genotype mapping and
            remove any snps that don't have significant p values in both strands
        3.  Output the new genotype mapping files";
        
#command line arguments
my $logFile;
my $statusFile;
my $dipDir;
my $origGenoMapFile;
my $origHomoIndexMapping_F;
my $origHomoIndexMapping_R;
my $ignoreFile;
my $pValueFilterOn;
my $pValueFilterORorAND;
my $outputPValueFilterSNPs;
my $outputIgnoredIndividuals;
my $pValueCut;
my $heteroCut;
my $homoCut;
my $hiQualChoice;
my $outputDir;
my $finalGenoMapFile;
my $homoIndexMapping_F;
my $homoIndexMapping_R;
my $homoIndexQualityMapping;
my $genotypeOption;

#other variables
my $rank;
my $size;
my $hostname;
my @dipFilesF;
my @dipFilesR;
my $significant = 1;
my $notSignificant = 0;
my $bioStatsString = "";
my $perfStatsString = "";
my $handleUserOutput = $true;
my $numPartsInProgressBar = 50;
my $progressPartsOutputSoFar = 0;
my $nucOption = 0;
my $parentOption = 1;

#a hash reference of hash references (2d hash )  1st hash's key is individual 
#name, second key is snp, the value stored is string code representing
#whether the snp across both strands is being considered homozygote
#from one or the other parent, heterozygote, or NA
my $finalGenoMap;

#a hash reference of hash references (2d hash )  1st hash's key is individual 
#name, second key is snp, the value stored is the HomoIndex value
my $finalHomoIndexValues_F;
my $finalHomoIndexValues_R;

#keys are snp, values are significant or not
my %dipF;
my %dipR;

#key is individuals to ignore, values are
#dummy variables
my %individualsToIgnore;

#program execution--------------------------------------------------------------

#are there valid command line arguments?
&checkUsage( ) ;
&getArguments( ) ;

#assumes these environment variables have been set
$rank = $ENV{MPI_RANK};
$size = $ENV{MPI_SIZE};
$hostname = $ENV{MPI_HOSTNAME};

if ( !defined( $hostname  )  )      
{
   $hostname = "local";   # executing locally ( not in parallel  )  
   $rank = 0;
   $size = 1;
}

#open the log file and set up a header
&initializeLogFile( ) ;

#open the status file
$statusFile = $statusFile . "_$hostname" . "_$rank.status";

if ( !open(STATUS, ">$statusFile" )  ) 
{
    &printError("Could not open status file: $statusFile, quitting" ) ;
    close(LOG ) ;
    exit;
}

#does the dip files directory exist?
opendir(TEMP, $dipDir )  or
&exitEarly("Could not open directory $dipDir for input" ) ;

close(TEMP ) ;
@dipFilesF = glob( $dipDir . "/*F.dip" ) ;
@dipFilesR = glob( $dipDir . "/*R.dip" ) ;

#are there files to process?
if ( $pValueFilterOn && (scalar @dipFilesF < 1 || scalar @dipFilesR < 1 ||
                       scalar @dipFilesF != scalar @dipFilesR )  ) 
{
    &exitEarly( "Invalid # of dip files. F strand: " . scalar( @dipFilesF ) .
                "   R strand: " . scalar( @dipFilesR ) . "\n" );
}

#get the list of individuals to ignore
open(IG, $ignoreFile )  or
&exitEarly("Could not open $ignoreFile for input" ) ;

while(my $line = <IG> ) 
{
    chop($line ) ;
    $individualsToIgnore{$line} = "dummy";
}
close(IG ) ;

#does the output directory exist?
if ( !opendir(TEMP, $outputDir )  ) 
{
    print LOG "Creating directory: $outputDir\n";
    mkdir($outputDir ) ;
}
else
{
    close(TEMP ) ;
}

#program flow
if ( $pValueFilterOn ) 
{
    %dipF = %{&inputDipResults(\@dipFilesF, "Forward" ) };
    %dipR = %{&inputDipResults(\@dipFilesR, "Reverse" ) };
}

#quality check
my $numDipFKeys = keys %dipF;
my $numDipRKeys = keys %dipR;
if ( $numDipFKeys != $numDipRKeys ) 
{
    &exitEarly("Unequal number of snps in dip result input files" ) ;
}

&createNewMappings( ) ;
&outputResults( ) ;
&exitSuccessfully("Successfully generated the final genotype mapping files" ) ;

#-------------------------------------------------------------------------------
#
#  Open and initialize the log file for appending
#
sub initializeLogFile
{
    $logFile = $logFile . "_$hostname" . "_$rank.log";
    
    if ( !open(LOG, ">>$logFile" )  ) 
    {
        &printError("Could not open log file for output: $logFile, quitting" ) ;
        exit;
    }

    #output log info about this run
    my @timeData = localtime(time ) ;
    my $month = $timeData[4] + 1;
    my $day = $timeData[3];
    my $year = $timeData[5] + 1900;
    my $hour = $timeData[2];
    my $minute = $timeData[1];

    #header info
    print LOG "


-------------------Run on $month/$day/$year at $hour:$minute-------------------
    logFile:\t$logFile
    statusFile:\t$statusFile
    dipFiles:\t$dipDir
    originalGenotypeMappingFile:\t$origGenoMapFile
    originalHomoIndexMapping_F:\t$origHomoIndexMapping_F
    originalHomoIndexMapping_R:\t$origHomoIndexMapping_R
    individualsToIgnore:\t$ignoreFile
    pValueFilter:\t$pValueFilterOn
    pValueFilterORorAND:\t$pValueFilterORorAND
    outputPValueFilterSNPs:\t$outputPValueFilterSNPs
    outputIgnoredIndividuals:\t$outputIgnoredIndividuals
    pValueCutoff:\t$pValueCut
    heteroCutoff:\t$heteroCut
    homoCutoff:\t$homoCut
    homoIndexQualityChoice:\t$hiQualChoice
    outputDir:\t$outputDir
    finalGenotypeMappingFile:\t$finalGenoMapFile
    homoIndexMapping_F:\t$homoIndexMapping_F
    homoIndexMapping_R:\t$homoIndexMapping_R
    homoIndexQualityMapping:\t$homoIndexQualityMapping
    genotypeOption:\t$genotypeOption\n\n";
}

#-------------------------------------------------------------------------------
#
#  Pass in an array of .dip files and a name to associate with the batch of
#  dip files.  The function will go through all the dip files and record which
#  snps have significant p values and which ones don't in a hash, and the hash
#  will be returned.
#
sub inputDipResults
{
    my ($ar, $batch )  = @_;
    my @dipFiles = @{$ar};

    my $numFiles = @dipFiles;
    my %hash;
    
    &printInfo("Processing $numFiles $batch dip result files" ) ;
    &recordBioStat($numFiles,
    "NUM DIP TEST FILES",
    "The num of $batch dip test files processed for pValues" ) ;
    &outputProgressBar( ) ;
    
    #input go through all the files
    for ( my $i = 0; $i < $numFiles; ++$i ) 
    {
        &updateProgress($numFiles, $i ) ;
        
        #attempt to open the file
        my $file = $dipFiles[$i];
        open(DIP, $file )  or
        &exitEarly("Could not open file: $file for input" ) ;
        
        #this should only iterate once, there should
        #only be one line in each file
        my $iterations = 0;
        while(my $inputLine = <DIP> ) 
        {
            #quality check
            if ( $iterations > 0 ) 
            {
                &exitEarly("More than one line in dip results file: $file" ) ;
            }
            
            if ( $inputLine =~ /^(\w+)\s+([0-9\.]+|[0-9\.]+e-[0-9]+|NA)$/ ) 
            {
                my $snp = $1;
                my $pVal = $2;
                
                #reall small pValue?
                if ( $pVal =~ /([0-9\.]+)e(-[0-9]+)/ ) 
                {
                    $pVal = $1 * (10 ** $2 ) ;
                }
                
                #quality check
                if ( $pVal ne "NA" && ($pVal > 1 || $pVal < 0 )  ) 
                {
                   &exitEarly("Out of range pVal in dip results file: $file: $inputLine: $pVal" ) ;
                }
                elsif ( $pVal eq "NA" || $pVal >= $pValueCut ) 
                {
                    $hash{$snp} = $notSignificant;
                }
                else
                {
                    $hash{$snp} = $significant;
                }
            }
            else
            {
                &exitEarly("Incorrectly formatted dip results file: $file: $inputLine" ) ;
            }
            
            ++$iterations;
        }
        
        close(DIP ) ;
    }
    
    &updateProgress($numFiles, $numFiles ) ;
    
    \%hash;
}

#-------------------------------------------------------------------------------
#
#  Go through every individual in the original mapping files and take out any
#  snps that have invalid p values.
#
sub createNewMappings
{    
    my $info = "Processing the original genotype and HomoIndex mapping files ";
    $info .= "and removing SNPs with invalid pValues if that filter is on";
    &printInfo($info ) ;
    
    #attempt to open the original files for input
    open(MAP, $origGenoMapFile )  or
    &exitEarly("Could not open file: $origGenoMapFile for input" ) ;
    
    open(HIMFIN, $origHomoIndexMapping_F )  or
    &exitEarly("Could not open file: $origHomoIndexMapping_F for input" ) ;
    
    open (HIMRIN, $origHomoIndexMapping_R )  or
    &exitEarly("Could not open file: $origHomoIndexMapping_R for input" ) ;
    
    #there should be the same number of rows (lines ) 
    #and columns in the three initial mapping input
    #files
    
    my $firstLine = <MAP>;
    chomp($firstLine ) ;
    my @snps = split('\t', $firstLine ) ;
    my $numSNPs = @snps;
    
    #quality check
    my $firstLine_hiF = <HIMFIN>;
    my $firstLine_hiR = <HIMRIN>;
    chomp($firstLine_hiF ) ;
    chomp($firstLine_hiR ) ;
    my @snps_F = split('\t', $firstLine_hiF ) ;
    my @snps_R = split('\t', $firstLine_hiR ) ;
    
    if ( scalar @snps != scalar @snps_F ||
       scalar @snps != scalar @snps_R ) 
    {
        my $error = "The num of SNPs in the three initial mapping files are ";
        $error .= "not equal";
        &exitEarly($error ) ;
    }
    
    for ( my $i = 0; $i < scalar @snps; ++$i ) 
    {
        if ( $snps[$i] ne $snps_F[$i] ||
           $snps[$i] ne $snps_R[$i] ) 
        {
            my $error = "The order of the SNP columns in the three initial ";
            $error .= "mapping files are not the same";
            &exitEarly($error ) ;
        }
    }
    
    my $numDIP_F_snps = scalar keys %dipF;
    
    #counters
    my $numPValueSNPs = 0;
    my $individualsIgnored = 0;
    
    my $firstIteration = $true;
    
    #go through every line/individual of the orig genotype mapping
    while(my $inputLine = <MAP> ) 
    {
        my $hiLineF = <HIMFIN>;
        my $hiLineR = <HIMRIN>;
    
        chomp($inputLine ) ;
        chomp($hiLineF ) ;
        chomp($hiLineR ) ;
        
        my @mapValues = split('\t', $inputLine ) ;
        my @hiValuesF = split('\t', $hiLineF ) ;
        my @hiValuesR = split('\t', $hiLineR ) ;
        
        my $numValues = @mapValues;
        my $numValuesF = @hiValuesF;
        my $numValuesR = @hiValuesR;
        
        #quality check
        #NOTE - $numValues and $numSNPs are actually one
        #more than there are snps because of the individual
        #name on the line, and the leading blank before the
        #tab on the header
        if ( $numValues != $numSNPs ||
           ($pValueFilterOn && $numValues != $numDIP_F_snps + 1 )  ||
            $numValues != $numValuesF || $numValues != $numValuesR ) 
        {   
            my $error = "Mismatch between the num of snp values on a line in ";
            $error .= "the initial mapping files or the num of snp dip ";
            $error .= "results";
            &exitEarly($error ) ;
        }
        
        if ( $mapValues[0] ne $hiValuesF[0] || $mapValues[0] ne $hiValuesR[0] ) 
        {
            my $error = "The order of the individuals in the initial ";
            $error .= "mapping files are not the same";
            &exitEarly($error ) ;
        }
        
        #are we ignoring this individual?
        if ( !defined $individualsToIgnore{$mapValues[0]} ) 
        {
            $finalGenoMap -> {$mapValues[0]} = {};
            $finalHomoIndexValues_F -> {$mapValues[0]} = {};
            $finalHomoIndexValues_R -> {$mapValues[0]} = {};
            
            #go through every snp
            for ( my $i = 1; $i < $numValues; ++$i ) 
            {
                my $snp = $snps[$i];
                
                #the pValue filter
                if ( $pValueFilterOn && (($pValueFilterORorAND &&
                                        (($dipF{$snp} == $notSignificant )  || 
                                         ($dipR{$snp} == $notSignificant )  )  )  ||
                                       (!$pValueFilterORorAND &&
                                        (($dipF{$snp} == $notSignificant )  && 
                                         ($dipR{$snp} == $notSignificant )  )  )  )  ) 
                {
                    if ( $firstIteration ) 
                    {
                        ++$numPValueSNPs;
                        
                        if ( $outputPValueFilterSNPs ) 
                        {
                            &recordBioStat($snp, "PVALUE FILTER SNP",
                            "A SNP removed because of invalid pValues from its dip test" ) ;
                        }
                    }
                }
                #SNP passed the filter(s ) 
                else
                {
                    $finalGenoMap -> {$mapValues[0]} -> {$snp} = $mapValues[$i];
                    $finalHomoIndexValues_F -> {$mapValues[0]} -> {$snp} = $hiValuesF[$i];
                    $finalHomoIndexValues_R -> {$mapValues[0]} -> {$snp} = $hiValuesR[$i];
                }
            }
        }
        else
        {
            ++$individualsIgnored;
            
            if ( $outputIgnoredIndividuals ) 
            {
                &recordBioStat($mapValues[0], "IGNORED INDIVIDUAL",
                "An individual removed because it was listed in the ignore individuals file" ) ;
            }
        }
        
        $firstIteration = $false;
    }
    
    #quality check
    if ( (defined (my $temp1 = <HIMFIN> )  )  || (defined (my $temp2 = <HIMRIN> )  )  ) 
    {
        my $error = "There are a different num of individuals in the initial ";
        $error .= "mapping files";
        &exitEarly($error ) ;
    }
    
    close(MAP ) ;
    close(HIMFIN ) ;
    close(HIMRIN ) ;
    
    #record stats
    &recordBioStat($individualsIgnored, "NUM INDIVIDUALS IGNORED",
    "The num of individuals manually removed from the final genotype mapping (usually the parent clones or the F1s ) , relates to the individuals to ignore parameter" ) ;
    
    &recordBioStat($numPValueSNPs, "NUM PVALUE REMOVED SNPS",
    "The num of SNPs removed by the pValue filter" ) ;
}

#-------------------------------------------------------------------------------
#
#  Output the final results, and compute many final statistics and output them
#
sub outputResults
{
    &printInfo("Outputting the final mapping files" ) ;
    &outputProgressBar( ) ;

    #attempt to open files for output
    my $genoOut = $outputDir . "/" . $finalGenoMapFile;
    my $hiF = $outputDir . "/" . $homoIndexMapping_F;
    my $hiR = $outputDir . "/" . $homoIndexMapping_R;
    my $hiQual = $outputDir . "/" . $homoIndexQualityMapping;
    
    open(OUT, ">$genoOut" )  or
    &exitEarly("Could not open $genoOut for output" ) ;
    
    open(HIMF, ">$hiF" )  or
    &exitEarly("Could not open $hiF for output" ) ;
    
    open(HIMR, ">$hiR" )  or
    &exitEarly("Could not open $hiR for output" ) ;
    
    open(HIQUAL, ">$hiQual" )  or
    &exitEarly("Could not open $hiQual for output" ) ;
    
    #final num of SNPs and Individuals
    my @individuals = keys %$finalGenoMap;
    my $numIndividuals = @individuals;
    
    if ( $numIndividuals < 1 ) 
    {
        &exitEarly("0 individuals in the final genotype mapping" ) ;
    }
    
    my $hash = $finalGenoMap -> {$individuals[0]};
    my @snps = keys %$hash;
    my $numSNPs = @snps;
    
    if ( $numSNPs < 1 ) 
    {
        &exitEarly("0 individuals in the final genotype mapping" ) ;
    }
    
    #output a few statistics
    &recordBioStat($numIndividuals, "FINAL NUM INDIVIDUALS",
    "The num of individuals in the final genotype mapping file" ) ;
    
    &recordBioStat($numSNPs, "FINAL NUM SNPs",
    "The num of SNPs in the final genotype mapping file" ) ;
    
    #some statistic variables
    my $numH = 0;
    my $numB = 0;
    my $numNA = 0;
    my $numA = 0;
    my $numC = 0;
    my $numT = 0;
    my $numG = 0;
    my $numAC = 0;
    my $numAG = 0;
    my $numAT = 0;
    my $numCA = 0;
    my $numCT = 0;
    my $numCG = 0;
    my $numTA = 0;
    my $numTG = 0;
    my $numTC = 0;
    my $numGA = 0;
    my $numGT = 0;
    my $numGC = 0;
    my $totalNonNA = 0;
    
    my $hiQualSum_total = 0;
    my $hiQualSum_B = 0;
    my $hiQualSum_H = 0;
    my $hiQualSum_A = 0;
    my $hiQualSum_C = 0;
    my $hiQualSum_T = 0;
    my $hiQualSum_G = 0;
    my $hiQualSum_AC = 0;
    my $hiQualSum_AG = 0;
    my $hiQualSum_AT = 0;
    my $hiQualSum_CA = 0;
    my $hiQualSum_CT = 0;
    my $hiQualSum_CG = 0;
    my $hiQualSum_TA = 0;
    my $hiQualSum_TC = 0;
    my $hiQualSum_TG = 0;
    my $hiQualSum_GA = 0;
    my $hiQualSum_GC = 0;
    my $hiQualSum_GT = 0;
    my $hiQualSum_NA = 0;
    
    #variables for a quality score
    my %contigConsistency;
    my $numContigsWithMultipleSNPs = 0;
    my $goodContigs = 0;
    
    my %snpsSharingAContig;
    my $firstIteration = $true;
    
    #the header
    my $header = join("\t", @snps ) ;
    
    print OUT "\t$header";
    print HIMF "\t$header";
    print HIMR "\t$header";
    print HIQUAL "\t$header";
    
    #go through each individual
    for ( my $k = 0; $k < $numIndividuals; ++$k ) 
    {
        &updateProgress($numIndividuals, $k ) ;
    
        my $file = $individuals[$k];
    
        print OUT "\n$file";
        print HIMF "\n$file";
        print HIMR "\n$file";   
        print HIQUAL "\n$file";     
        
        %contigConsistency = ( ) ;
        
        #go through every snp
        foreach my $snp (@snps ) 
        {
            my $mapVal = $finalGenoMap -> {$file} -> {$snp};
            my $hiVal_F = $finalHomoIndexValues_F -> {$file} -> {$snp};
            my $hiVal_R = $finalHomoIndexValues_R -> {$file} -> {$snp};
            my $hiQ_F = &homoIndexQualityScore($hiVal_F ) ;
            my $hiQ_R = &homoIndexQualityScore($hiVal_R ) ;
            my $hiQualScore = &selectHomoIndexQualityScore($hiQ_F, $hiQ_R ) ;
            $hiQualSum_total += $hiQualScore;
            
            print OUT "\t$mapVal";
            print HIMF "\t$hiVal_F";
            print HIMR "\t$hiVal_R";
            print HIQUAL "\t$hiQualScore";
            
            #record statistic counts
            
            #for the distribution statistics
            if ( $mapVal eq "B" ) 
            {
                ++$numB;
                ++$totalNonNA;
                $hiQualSum_B += $hiQualScore;
            }
            elsif ( $mapVal eq "H" ) 
            {
                ++$numH;
                ++$totalNonNA;
                $hiQualSum_H += $hiQualScore;
            }
            elsif ( $mapVal eq "NA" ) 
            {
                ++$numNA;
                $hiQualSum_NA += $hiQualScore;
            }
            elsif ( $mapVal eq "A" ) 
            {
                ++$numA;
                ++$totalNonNA;
                $hiQualSum_A += $hiQualScore;
            }
            elsif ( $mapVal eq "C" ) 
            {
                ++$numC;
                ++$totalNonNA;
                $hiQualSum_C += $hiQualScore;
            }
            elsif ( $mapVal eq "T" ) 
            {
                ++$numT;
                ++$totalNonNA;
                $hiQualSum_T += $hiQualScore;
            }
            elsif ( $mapVal eq "G" ) 
            {
                ++$numG;
                ++$totalNonNA;
                $hiQualSum_G += $hiQualScore;
            }
            elsif ( $mapVal eq "AC" ) 
            {
                ++$numAC;
                ++$totalNonNA;
                $hiQualSum_AC += $hiQualScore;
            }
            elsif ( $mapVal eq "AT" ) 
            {
                ++$numAT;
                ++$totalNonNA;
                $hiQualSum_AT += $hiQualScore;
            }
            elsif ( $mapVal eq "AG" ) 
            {
                ++$numAG;
                ++$totalNonNA;
                $hiQualSum_AG += $hiQualScore;
            }
            elsif ( $mapVal eq "CA" ) 
            {
                ++$numCA;
                ++$totalNonNA;
                $hiQualSum_CA += $hiQualScore;
            }
            elsif ( $mapVal eq "CT" ) 
            {
                ++$numCT;
                ++$totalNonNA;
                $hiQualSum_CT += $hiQualScore;
            }
            elsif ( $mapVal eq "CG" ) 
            {
                ++$numCG;
                ++$totalNonNA;
                $hiQualSum_CG += $hiQualScore;
            }
            elsif ( $mapVal eq "GA" ) 
            {
                ++$numGA;
                ++$totalNonNA;
                $hiQualSum_GA += $hiQualScore;
            }
            elsif ( $mapVal eq "GC" ) 
            {
                ++$numGC;
                ++$totalNonNA;
                $hiQualSum_GC += $hiQualScore;
            }
            elsif ( $mapVal eq "GT" ) 
            {
                ++$numGT;
                ++$totalNonNA;
                $hiQualSum_GT += $hiQualScore;
            }
            elsif ( $mapVal eq "TA" ) 
            {
                ++$numTA;
                ++$totalNonNA;
                $hiQualSum_TA += $hiQualScore;
            }
            elsif ( $mapVal eq "TC" ) 
            {
                ++$numTC;
                ++$totalNonNA;
                $hiQualSum_TC += $hiQualScore;
            }
            elsif ( $mapVal eq "TG" ) 
            {
                ++$numTG;
                ++$totalNonNA;
                $hiQualSum_TG += $hiQualScore;
            }
            else
            {
                my $error = "Invalid genotype mapping allele in the ";
                $error .= "outputResults function of the Final ";
                $error .= "Genotype Mapping step: $mapVal";
                &exitEarly($error ) ;
            }
            
            my $contig;
            my $position;
            
            #for the contig statistics
            if ( $snp =~ /^([A-Za-z0-9_]+)_([0-9]+)$/ ) 
            {
                $contig = $1;
                $position = $2;
            }
            else
            {
                &exitEarly("Invalid snp identifier in output function: $snp" ) ;
            }
            
            #how many SNPs share a contig?
            if ( $firstIteration ) 
            {
                $snpsSharingAContig{$contig} .= "$position:";
            }
            
            #contig quality score
            if ( $mapVal ne "NA" ) 
            {
                $contigConsistency{$contig} .= "$mapVal:";
            }
        }
        
        #Extra computations for the quality score statistic
        my @contigKeys = keys %contigConsistency;
        foreach my $contig (@contigKeys ) 
        {
            my @values = split(":", $contigConsistency{$contig} ) ;
            
            #if it's a contig with more than one SNP
            if ( scalar @values > 1 ) 
            {
                ++$numContigsWithMultipleSNPs;
                
                my $val;
                my $keepGoing = $true;
            
                for ( my $k = 0; $k < scalar @values && $keepGoing; ++$k ) 
                {
                    if ( !defined $val ) 
                    {
                        $val = $values[$k];
                    }
                    elsif ( $val ne $values[$k] ) 
                    {
                        $keepGoing = $false;
                    }
                }
                
                if ( $keepGoing ) 
                {
                    ++$goodContigs;
                }
            }
        }
        
        $firstIteration = $false;
    }
    
    &updateProgress($numIndividuals, $numIndividuals ) ;
    
    #record statistics
    my $percent;
    my $average;
    
    my $homoPercent = 0;
    my $heteroPercent = 0;
    
    #the distribution statistics
    if ( $totalNonNA > 0  ) 
    {
        my $prefix = "The percent of";
        my $suffix = "genotype mapping alleles out of non-NA alleles\n"
        . "\t\tacross all individuals and SNPs in the final genotype mapping";
        
        $percent = $numH / $totalNonNA;
        &recordBioStat( $percent, "PERCENT H", "$prefix H $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numB / $totalNonNA;
        &recordBioStat( $percent, "PERCENT B", "$prefix B $suffix"  ) ;
        $homoPercent += $percent;
        
        $percent = $numA / $totalNonNA;
        &recordBioStat( $percent, "PERCENT A", "$prefix A $suffix"  ) ;
        $homoPercent += $percent;
        
        $percent = $numC / $totalNonNA;
        &recordBioStat( $percent, "PERCENT C", "$prefix C $suffix"  ) ;
        $homoPercent += $percent;
        
        $percent = $numT / $totalNonNA;
        &recordBioStat( $percent, "PERCENT T", "$prefix T $suffix"  ) ;
        $homoPercent += $percent;
        
        $percent = $numG / $totalNonNA;
        &recordBioStat( $percent, "PERCENT G", "$prefix G $suffix"  ) ;
        $homoPercent += $percent;
        
        $percent = $numAC / $totalNonNA;
        &recordBioStat( $percent, "PERCENT AC", "$prefix AC $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numAT / $totalNonNA;
        &recordBioStat( $percent, "PERCENT AT", "$prefix AT $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numAG / $totalNonNA;
        &recordBioStat($percent, "PERCENT AG", "$prefix AG $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numCA / $totalNonNA;
        &recordBioStat($percent, "PERCENT CA", "$prefix CA $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numCT / $totalNonNA;
        &recordBioStat($percent, "PERCENT CT", "$prefix CT $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numCG / $totalNonNA;
        &recordBioStat($percent, "PERCENT CG", "$prefix CG $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numTA / $totalNonNA;
        &recordBioStat($percent, "PERCENT TA", "$prefix TA $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numTC / $totalNonNA;
        &recordBioStat($percent, "PERCENT TC", "$prefix TC $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numTG / $totalNonNA;
        &recordBioStat($percent, "PERCENT TG", "$prefix TG $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numGA / $totalNonNA;
        &recordBioStat($percent, "PERCENT GA", "$prefix GA $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numGC / $totalNonNA;
        &recordBioStat($percent, "PERCENT GC", "$prefix GC $suffix"  ) ;
        $heteroPercent += $percent;
        
        $percent = $numGT / $totalNonNA;
        &recordBioStat($percent, "PERCENT GT", "$prefix GT $suffix"  ) ;
        $heteroPercent += $percent;
        
        &recordBioStat($heteroPercent, "TOTAL PERCENT HETERO",
        "The total percent of heterozygous genotype mapping alleles" ) ;
        
        &recordBioStat($homoPercent, "TOTAL PERCENT HOMO",
        "The total percent of homozygous genotype mapping alleles" ) ;
    }
    else
    {
        my $error = "There were 0 non NA alleles in the final genotype ";
        $error .= "mapping across all individuals";
        &exitEarly($error ) ;
    }
    
    #average HomoIndex Quality Scores
    $average = $hiQualSum_total / ($numIndividuals * $numSNPs ) ;
    &recordBioStat($average, "AVERAGE HOMOINDEX QUALITY SCORE",
    "The average HomoIndex quality score across all genotype mapping alleles" ) ;
    
    if ( $numB > 0 ) 
    {
        $average = $hiQualSum_B / $numB;
        &recordBioStat($average, "AVERAGE B HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all B genotype mapping alleles" ) ;
    }
    
    if ( $numH > 0 ) 
    {
        $average = $hiQualSum_H / $numH;
        &recordBioStat($average, "AVERAGE H HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all H genotype mapping alleles" ) ;
    }
    
    if ( $numA > 0 ) 
    {
        $average = $hiQualSum_A / $numA;
        &recordBioStat($average, "AVERAGE A HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all A genotype mapping alleles" ) ;
    }
    
    if ( $numC > 0 ) 
    {
        $average = $hiQualSum_C / $numC;
        &recordBioStat($average, "AVERAGE C HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all C genotype mapping alleles" ) ;
    }
    
    if ( $numT > 0 ) 
    {
        $average = $hiQualSum_T / $numT;
        &recordBioStat($average, "AVERAGE T HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all T genotype mapping alleles" ) ;
    }
    
    if ( $numG > 0 ) 
    {
        $average = $hiQualSum_G / $numG;
        &recordBioStat($average, "AVERAGE G HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all G genotype mapping alleles" ) ;
    }
    
    if ( $numAC > 0 ) 
    {
        $average = $hiQualSum_AC / $numAC;
        &recordBioStat($average, "AVERAGE AC HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all AC genotype mapping alleles" ) ;
    }
    
    if ( $numAT > 0 ) 
    {
        $average = $hiQualSum_AT / $numAT;
        &recordBioStat($average, "AVERAGE AT HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all AT genotype mapping alleles" ) ;
    }
    
    if ( $numAG > 0 ) 
    {
        $average = $hiQualSum_AG / $numAG;
        &recordBioStat($average, "AVERAGE AG HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all AG genotype mapping alleles" ) ;
    }
    
    if ( $numCA > 0 ) 
    {
        $average = $hiQualSum_CA / $numCA;
        &recordBioStat($average, "AVERAGE CA HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all CA genotype mapping alleles" ) ;
    }
    
    if ( $numCT > 0 ) 
    {
        $average = $hiQualSum_CT / $numCT;
        &recordBioStat($average, "AVERAGE CT HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all CT genotype mapping alleles" ) ;
    }
    
    if ( $numCG > 0 ) 
    {
        $average = $hiQualSum_CG / $numCG;
        &recordBioStat($average, "AVERAGE CG HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all CG genotype mapping alleles" ) ;
    }
    
    if ( $numTA > 0 ) 
    {
        $average = $hiQualSum_TA / $numTA;
        &recordBioStat($average, "AVERAGE TA HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all TA genotype mapping alleles" ) ;
    }
    
    if ( $numTC > 0 ) 
    {
        $average = $hiQualSum_TC / $numTC;
        &recordBioStat($average, "AVERAGE TC HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all TC genotype mapping alleles" ) ;
    }
    
    if ( $numTG > 0 ) 
    {
        $average = $hiQualSum_TG / $numTG;
        &recordBioStat($average, "AVERAGE TG HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all TG genotype mapping alleles" ) ;
    }
    
    if ( $numGA > 0 ) 
    {
        $average = $hiQualSum_GA / $numGA;
        &recordBioStat($average, "AVERAGE GA HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all GA genotype mapping alleles" ) ;
    }
    
    if ( $numGC > 0 ) 
    {
        $average = $hiQualSum_GC / $numGC;
        &recordBioStat($average, "AVERAGE GC HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all GC genotype mapping alleles" ) ;
    }
    
    if ( $numGT > 0 ) 
    {
        $average = $hiQualSum_GT / $numG;
        &recordBioStat($average, "AVERAGE GT HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all GT genotype mapping alleles" ) ;
    }

    if ( $numNA > 0 ) 
    {
        $average = $hiQualSum_NA / $numNA;
        &recordBioStat($average, "AVERAGE NA HOMOINDEX QUALITY SCORE",
        "The average HomoIndex quality score across all NA genotype mapping alleles" ) ;
    }
    
    #the average gap between two consecutive snps on the same contig
    $percent = 0;
    my @gaps;
    
    foreach my $contig (keys %snpsSharingAContig ) 
    {
        my @positions = sort {$a <=> $b} (split(":", $snpsSharingAContig{$contig} )  ) ;
    
        if ( scalar @positions > 1 ) 
        {
            $percent += scalar @positions;
            
            my $p1 = $positions[0];
            my $p2 = $positions[1];
            
            push(@gaps, $p2 - $p1 ) ;
            $p1 = $p2;
            
            for ( my $i = 2; $i < scalar @positions; ++$i ) 
            {
                $p2 = $positions[$i];
                push(@gaps, $p2 - $p1 ) ;
                $p1 = $p2;
            }
        }
    }
    
    #percent SNPs that share a contig
    $percent = $percent / $numSNPs;
    
    #average gap between SNPs on same contig
    $average = -1;
    
    if ( scalar @gaps > 0 ) 
    {
        $average = sum(@gaps )  / scalar @gaps;
    }
    
    &recordBioStat( $percent,
    "PERCENT SNPS ON CONTIGS WITH ANOTHER",
    "The percent of SNPs, out of the final SNPs, that are located on a \n"
    . "\t\tcontig with at least one other SNP"  ) ;
    
    &recordBioStat($average,
    "AVERAGE SNP GAP",
    "The average distance between two consecutive SNPs, out of the final \n"
    . "\t\tSNPs, on the same contig"  ) ;
    
    &recordBioStat($numContigsWithMultipleSNPs, "NUM MULTIPLE NON NA SNP CONTIGS",
    "The num of contigs across all individuals, out of the final individuals,\n"
    . "\t\tthat contian multiple non-NA SNP genotype mapping alleles" ) ;
    
    &recordBioStat($goodContigs, "NUM CONSISTENT CONTIGS",
    "The num of contigs within the NUM MULTIPLE NON NA SNP CONTIGS that have\n"
    . "\t\tthe same SNP genotype mapping alleles (excluding NA alleles )  across\n"
    . "\t\ttheir contig" ) ;
    
    if ( $numContigsWithMultipleSNPs > 0 ) 
    {
        $percent = $goodContigs / $numContigsWithMultipleSNPs;
        &recordBioStat($percent, "CONSISTENT CONTIG QUALITY SCORE",
        "(NUM CONSISTENT CONTIGS )  / (NUM MULTIPLE NON NA SNP CONTIGS ) " ) ;
    }
    
    close(OUT ) ;
    close(HIMF ) ;
    close(HIMR ) ;
    close(HIQUAL ) ;
}

#-------------------------------------------------------------------------------
#
#  Compute a HomoIndex quality score based on a HomoIndex value
#
sub homoIndexQualityScore
{
    my ($homoIndexValue )  = @_;
    
    #homo HI qual score
    if ( $homoIndexValue <= $homoCut ) 
    {
        return $homoIndexValue;
    }
    #hetero HI qual score
    elsif ( $homoIndexValue >= $heteroCut ) 
    {
        return (1 - $homoIndexValue ) ;
    }
    #NA qual score
    #This is just an arbitrary way to figure out
    #HI qual scores for NA HI values
    else
    {
        my $middle = ($heteroCut + $homoCut )  / 2;
        
        if ( $homoIndexValue < $middle ) 
        {
            return $homoIndexValue;
        }
        else
        {
            return (1 - $homoIndexValue ) ;
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Which of two HomoIndex values should be used
#
sub selectHomoIndexQualityScore
{
    my ($hi1, $hi2 )  = @_;

    #pick the worst
    if ( $hiQualChoice == 1 ) 
    {
        if ( $hi1 >= $hi2 ) 
        {
            return $hi1;
        }
        else
        {
            return $hi2;
        }
    }
    #pick the best
    elsif ( $hiQualChoice == 2 ) 
    {
        if ( $hi1 <= $hi2 ) 
        {
            return $hi1;
        }
        else
        {
            return $hi2;
        }
    }
    #return the average
    elsif ( $hiQualChoice == 3 ) 
    {
        return (($hi1 + $hi2 )  / 2 ) ;
    }
    #invalid
    else
    {
        my $error = "The HomoIndexQualityChoice parameter has an invalid ";
        $error .= "value: $hiQualChoice";
        &exitEarly($error ) ;
    }
}

#-------------------------------------------------------------------------------
#
#  Output a progress bar for the user to see, only do it if this is parallel
#  instance that should handle it
#
#  Note that this bypasses the log and print methods below that the rest of the
#  script uses
#
sub outputProgressBar
{
    if ( $handleUserOutput ) 
    {
        my $bar = "0% ";
        
        for ( my $i = 0; $i < $numPartsInProgressBar; ++$i ) 
        {
            $bar .= "_";
        }
        
        $bar .= " 100%";
        
        print STDERR "[$processTag: $infoTag]\t$bar\n";
        print STDERR "[$processTag: $infoTag]\t   ";
        $progressPartsOutputSoFar = 0;
    }
}

#-------------------------------------------------------------------------------
#
#  Update the progress bar, this decides whether to update the progress bar,
#  pass in the total number of things that need to be done, and the number of
#  things completed so far
#
#  Note that this bypasses the log and print methods below that the rest of the
#  script uses
#
sub updateProgress
{
    if ( $handleUserOutput ) 
    {
        my ($total, $completed )  = @_;
        
        my $stepSize = 100 / $numPartsInProgressBar;
        my $percentage = ($completed / $total )  * 100;
        
        while($progressPartsOutputSoFar * $stepSize < $percentage ) 
        {
            print STDERR ">";
            ++$progressPartsOutputSoFar;
        }
        
        if ( $completed >= $total ) 
        {
            print STDERR "\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Make sure the appropriate command line arguments have been passed to this
#  program, if not, print out the usage message and quit
#
sub checkUsage
{
    if ( scalar @ARGV != 21 ) 
    {
        print STDERR "$usageMessage\n\n";
        exit;
    }
}

#-------------------------------------------------------------------------------
#
#  Set variables to either their defaults or their inputted command line
#  argument
#
sub getArguments
{
    #these arguments are not optional, therefore no defaults possible
    $logFile = $ARGV[0];
    $statusFile = $ARGV[1];
    $dipDir = $ARGV[2];
    $origGenoMapFile = $ARGV[3];
    $origHomoIndexMapping_F = $ARGV[4];
    $origHomoIndexMapping_R = $ARGV[5];
    $ignoreFile = $ARGV[6];
    $pValueFilterOn = $ARGV[7];
    $pValueFilterORorAND = $ARGV[8];
    $outputPValueFilterSNPs = $ARGV[9];
    $outputIgnoredIndividuals = $ARGV[10];
    $pValueCut = $ARGV[11];
    $heteroCut = $ARGV[12];
    $homoCut = $ARGV[13];
    $hiQualChoice = $ARGV[14];
    $outputDir = $ARGV[15];
    $finalGenoMapFile = $ARGV[16];
    $homoIndexMapping_F = $ARGV[17];
    $homoIndexMapping_R = $ARGV[18];
    $homoIndexQualityMapping = $ARGV[19];
    $genotypeOption = $ARGV[20];
}

#-------------------------------------------------------------------------------
#
#  Print some info to the screen for the user to see, note that only one
#  parallel instance is allowed to print info the screen, info is also recorded
#  in the log
#
sub printInfo
{
    my ($message )  = @_;
    my @messageLines = split("\n", $message ) ;
    
    foreach my $line (@messageLines ) 
    {
        print LOG "[USER $infoTag]\t$line\n";
        
        if ( $handleUserOutput ) 
        {
            print STDERR "[$processTag: $infoTag]\t$line\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print error messages for the user to see, also records the messages in the
#  log, note that only one parallel instance is allowed to print info to the
#  screen
#
sub printError
{
    my ($message )  = @_;
    my @messageLines = split("\n", $message ) ;
    
    foreach my $line (@messageLines ) 
    {
        print LOG "[USER $errorTag]\t$line\n";
        
        if ( $handleUserOutput ) 
        {
            print STDERR "[$processTag: $errorTag]\t$line\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print messages to the log
#
sub logInfo
{
    my ($message )  = @_;
    my @messageLines = split("\n", $message ) ;
    
    foreach my $line (@messageLines ) 
    {
        print LOG "[LOG]\t$line\n";
    }
}

#-------------------------------------------------------------------------------
#
#  Record biology statistics, pass in a value for the statistic, pass in a tag,
#  and pass in a descriptor
#
sub recordBioStat
{
    my ($value, $tag, $desc )  = @_;
    
    $bioStatsString .= "[$tag]\t$value - $desc\n";
}

#-------------------------------------------------------------------------------
#
#  Record performance statistics, pass in a value for the statistic, pass in a
#  tag, and pass in a descriptor
#
sub recordPerfStat
{
    my ($value, $tag, $desc )  = @_;
    
    $perfStatsString .= "[$tag]\t$value - $desc\n";
}

#-------------------------------------------------------------------------------
#
#  Exit due to an error, close all open files and update the status file with an
#  error, also output the statistics to the status file
#
sub exitEarly
{
    my ($message )  = @_;
    
    print STATUS "$errorTag\n$message\n$perfStatsString";
    print STATUS "$bioTag\n$bioStatsString";
    print LOG "[$errorTag]\t$message\n";
    
    print STDERR "[$errorTag]\t$message\n";  # rdb added
    
    close(LOG ) ;
    close(STATUS ) ;
    close(OUT ) ;
    close(TEMP ) ;
    close(IG ) ;
    close(DIP ) ;
    close(MAP ) ;
    close(HIMFIN ) ;
    close(HIMRIN ) ;
    close(HIMF ) ;
    close(HIMR ) ;
    
    exit;
}

#-------------------------------------------------------------------------------
#
#  Finish the script by outputting success to the log and the status files, and
#  close both files, should be the last thing called, and then the program
#  should exit correctly, statistics will also be output to the status file
#
sub exitSuccessfully
{
    my ($message )  = @_;
    
    print STATUS "$successTag\n$message\n$perfStatsString";
    print STATUS "$bioTag\n$bioStatsString";
    print LOG "[$successTag]\t$message\n";
    
    close(LOG ) ;
    close(STATUS ) ;
    close(OUT ) ;
    close(TEMP ) ;
    close(IG ) ;
    close(DIP ) ;
    close(MAP ) ;
    close(HIMFIN ) ;
    close(HIMRIN ) ;
    close(HIMF ) ;
    close(HIMR ) ;
}
