#!/usr/bin/perl
#
#  Brian Albere
#
#  Generates HomoIndex values and genotype mapping alleles for the SNPs in a
#  subset of the total pipeline's bases files since this is a paralle step.
#  Prefilter computations are applied that mark certain genotype mapping
#  alleles invalid, and the data is output to various files.
#
#  This script should be used in parallel using MPI because it only attempts to
#  process a subset of all the base files since there should be other instances
#  of the script processing the rest.  The script will create a status file to
#  communicate back up the parallel stack the outcome of the script and the
#  scripts statistics
#
#  NOTE - info messages are printed to STDERR for the user to see because
#  anything printed to STDOUT during MPI is not put to the screen, but STDERR
#  is...In other words, not everything printed to STDERR in this file is
#  necessarily an error...This may not be entirely true, it could be that STDOUT
#  is being buffered and everything would be printed eventually, but STDERR
#  seems to work well, so just keeping this
#
#  output format:
#  snp  nucOrder HI class   genotype
#
#  'snp' is the name of the snp / loci, the SNP identifier
#  'nucOrder' an ordering of ACTG in the order of greatest luminescent reading
#       to least
#  'HI' the HomoIndex value
#  'class' the classification for this snp based on the HI value, could be 'HE'
#       (heterozygous), 'HO' (homozygous), or 'UK' (unknown)
#  'genotype' the allele based on matching up nucleotides with the
#       parent nucleotides, can be 'NA', 'H', 'A', or 'B'
#
#  TODO NOTE that a hack has been added to this code to remove data points 
#  based on their mean and std.  The data we received from NimbleGen very
#  distinguished clusters in the their standard deviation vs mean plots, and
#  code has been hacked in in an attempt to remove data in the smaller clusters.
#  This pre-filter is not currently controllable from the XML parameters.
#
#  Edit history
#  06/05/12 received from Brian.
#  06/12/12 rdb
#           -changed exitEarly to also send error message to STDERR
#           -handle uniprocessor case, where MPI env variables aren't set

use warnings;
use strict;
use List::Util qw(sum);

#variables----------------------------------------------------------------------

my $true = 1;
my $false = 0;
my $processTag = "HOMOINDEX";
my $infoTag = "INFO";
my $errorTag = "ERROR";
my $successTag = "SUCCESS";
my $bioTag = "BIOLOGY";
my $usage_message = "
Usage: HomoIndex.pl [logFile] [statusFile] [parentAlleleFile] [basesFiles]
                    [hiQualityCutoff] [heteroCutoff] [homoCutoff]
                    [hiQualityPreFilter] [genotypeOption]
                    [logProcessedFiles] [outputDir]

    logFile                 The name of the file where log information should be
                            stored during this script, this is only the
                            beginning of the file name, the hostname and the mpi
                            rank will be added to the name along with .log
    statusFile              The name of a status file where this script will put
                            information about the overall success of its
                            execution, this is only the beginning of the file
                            name, the hostname and the mpi rank will be added to
                            the name along with .status
    parentAlleleFile        A listing of the SNPs this experiment is dealing
                            with, along with the SNP nucleotides in the parent
                            clones
    basesFiles              The directory of bases files to process
    hiQualityCutoff         HomoIndex quality scores greater than this number
                            are considered invalid
    heteroCutoff            HomoIndex values greater than or equal to this value
                            are considered heterozygous
    homoCutoff              HomoIndex values less than or equal to this value
                            are considered homozygous
    hiQualityPreFilter      1 (on) means SNP genotype mapping alles will be
                            marked as invalid if at least one of the
                            contributing HomoIndex values has an invalid
                            HomoIndex quality score, 0 (off) will not
                            take the quality scores into consideration
    genotypeOption          1 means the genotype mapping alleles will either be
                            A, B, H, or NA depending on how each SNP's
                            nucleotides match up with the parent nucleotides, 0
                            means the genotype mapping alleles will either be
                            A, C, T, G, any combination of two, or NA, which is
                            basically each SNP's nucleotides on the
                            corresponding set of homologous chromosomes, if the
                            SNP is homo the genotype will be one nucleotide, and
                            if the SNP is hetero the genotype will be two
                            nucleotides, note that the nucleotides are the
                            nucleotides in the forward strand, the reverse
                            strand nucleotides are the compliments
    logProcessedFiles       1 (on) means log which files this parallel
                            instance handles, 0 (off) means don't do this
                            output
    outputDir               The directory to store the resulting .homoIndex
                            files
                                    
    Program Execution:
        1.  Open the parent allele file, read in and store the snp identifiers
            we're concerned with, and the parent clone SNP nucleotides
        2.  Compute which base files this process should be concerned with based
            on how many base files there are, how many MPI processes there are
            overall, and which MPI process this is (its rank)
        3.  Process each base files by computing homoIndex values,
            homoIndex classifications, genotype mapping alleles, etc.  Output the
            results.";

#command line arguments
my $logFile;
my $statusFile;
my $snpFile;
my $basesFilesDir;
my $hiQualCut;
my $heteroCut;
my $homoCut;
my $hiQualPreFilterOn;
my $genotypeOption;
my $logProcessedFiles;
my $outputDir;

#other variables
my $rank;
my $size;
my $hostname;
my @baseFiles;
my @myBaseFiles;
my $heteroClass = "HE";
my $homoClass = "HO";
my $unknownClass = "UK";
my $bioStatsString = "";
my $perfStatsString = "";
my $handleUserOutput = $false;
my $numPartsInProgressBar = 50;
my $progressPartsOutputSoFar = 0;
my $nucOption = 0;
my $parentOption = 1;

#error counts
my $numAllelesBadStdOverMean = 0;
my $numAllelesBadBothStrandHIQualScores = 0;
my $numAllelesBadFStrandHIQualScores = 0;
my $numAllelesBadRStrandHIQualScores = 0;
my $numAllelesUnknownBothStrandHIClass = 0;
my $numAllelesUnknownFStrandHIClass = 0;
my $numAllelesUnknownRStrandHIClass = 0;
my $numAllelesBadHeteroClassBothStrands = 0;
my $numAllelesBadHeteroClassFStrand = 0;
my $numAllelesBadHeteroClassRStrand = 0;
my $numAllelesHIClassesDontMatch = 0;
my $numAllelesHomoDontMatchEitherParent = 0;
my $numAllelesHeteroNuc1MatchesParentANuc2DoesntMatchParentB = 0;
my $numAllelesHeteroNuc1MatchesParentBNuc2DoesntMatchParentA = 0;
my $numAllelesHeteroNuc1DoesntMatchParentAOrB = 0;
my $numAllelesComplimentsDontMatch = 0;

#a hash reference of hash references (2d hash), 1st hash's key is snp,
#2nd hash's key is either 'A' or 'B' representing parentA or parentB, and value
#stored is the nucleotide at that snp in that parent in the forward
#strand
my $parentSNPs;

#program execution--------------------------------------------------------------

#are there valid command line arguments?
&checkUsage();
&getArguments();

#assumes these environment variables have been set
$rank = $ENV{MPI_RANK};
$size = $ENV{MPI_SIZE};
$hostname = $ENV{MPI_HOSTNAME};

if ( !defined( $hostname ))     
{
   $hostname = "local";   # executing locally ( not in parallel ) 
   $rank = 0;
   $size = 1;
}

#open the log file and set up a header
&initializeLogFile();

#open the status file
$statusFile = $statusFile . "_$hostname" . "_$rank.status";

if(!open(STATUS, ">$statusFile"))
{
    &printError("Could not open status file: $statusFile, quitting");
    close(LOG);
    exit;
}

#does the base files directory exist?
opendir(TEMP, $basesFilesDir) or
&exitEarly("Could not open directory $basesFilesDir for input");

close(TEMP);
@baseFiles = glob($basesFilesDir . "/*.bases");

#is there even anything to smooth?
if( scalar @baseFiles < 1 )
{
    &exitEarly("No bases files in $basesFilesDir directory");
}


#randomize which parallel instance handles the user output
my @timeData = localtime(time);
my $day = $timeData[3];
srand($day);

my $randIndex;
if($size > scalar @baseFiles)
{
    $randIndex = scalar @baseFiles;
}
else
{
    $randIndex = $size;
}

my $i = int(rand($randIndex));
if($i == $rank)
{
    $handleUserOutput = $true;
    &printInfo("The parallel instance handling the user output: $rank");
}

#does the output directory exist?
if(!opendir(TEMP, $outputDir))
{
    print LOG "Creating directory: $outputDir\n";
    mkdir($outputDir);
}
else
{
    close(TEMP);
}

#is there anything this process should do?
&computeMyFiles();

if(scalar @myBaseFiles > 0)
{
    #input the parent SNP data
    &inputParentSNPs();

    #generate the homoIndex data
    &processBaseFiles();

    #successful execution of program
    &exitSuccessfully("Successfully processed all bases files");
}
else
{
    &exitSuccessfully("No files processed, more cores available than needed");
}

#-------------------------------------------------------------------------------
#
#  Open and initialize the log file for appending
#
sub initializeLogFile
{
    $logFile = $logFile . "_$hostname" . "_$rank.log";
    
    if(!open(LOG, ">>$logFile"))
    {
        &printError("Could not open log file for output: $logFile, quitting");
        exit;
    }

    #output log info about this run
    my @timeData = localtime(time);
    my $month = $timeData[4] + 1;
    my $day = $timeData[3];
    my $year = $timeData[5] + 1900;
    my $hour = $timeData[2];
    my $minute = $timeData[1];

    #header info
    print LOG "


-------------------Run on $month/$day/$year at $hour:$minute-------------------
    logFile:\t$logFile
    statusFile:\t$statusFile
    parentAlleleFile:\t$snpFile
    basesFiles:\t$basesFilesDir
    hiQualityCutoff:\t$hiQualCut
    heteroCutoff:\t$heteroCut
    homoCutoff:\t$homoCut
    hiQualPreFilter:\t$hiQualPreFilterOn
    genotypeOption:\t$genotypeOption
    logProcessedFiles:\t$logProcessedFiles
    outputDir:\t$outputDir\n\n";
}

#-------------------------------------------------------------------------------
#
#  Open the parent allele file and store its info in a hash
#
sub inputParentSNPs
{
    my $info = "Processing the Parent Allele File to input the parent SNP ";
    $info .= "nucleotides";
    &printInfo($info);

    open(SNP, $snpFile) or
    &exitEarly("Could not open file: $snpFile for input");
    
    my $numSNPs = 0;
    
    #go through every line of the file
    while(my $inputLine = <SNP>)
    {
        if($inputLine =~ /^([a-zA-Z0-9_]+)\s+([0-9]+)\s+[ATCGactg]\s+([ATCG])\s+([ATCG]).*$/)
        {
            ++$numSNPs;
            my $snp = $1 . "_" . $2;
            $parentSNPs -> {$snp} = {};
            
            #store the parent nucleotides
            $parentSNPs -> {$snp} -> {"B"} = $3;
            $parentSNPs -> {$snp} -> {"A"} = $4;
        }
        else
        {
            my $error = "Incorrectly formatted Parent Allele File line: ";
            $error .= "$inputLine";
            &exitEarly($error);
        }
    }
    
    close(SNP);
    
    &recordBioStat($numSNPs, "NUM SNPS",
    "The num of SNP Identifiers in the Parent Allele File");
}

#-------------------------------------------------------------------------------
#
#  Call this function to fill in the array that contains which files this
#  process should handle since there will be multiple instances of this
#  process running in parallel
#
sub computeMyFiles
{
    my $numOfBaseFiles = @baseFiles;
    
    #more cores than there are files to process
    if($size > $numOfBaseFiles)
    {
        if($rank < $numOfBaseFiles)
        {
            $myBaseFiles[0] = $baseFiles[$rank];
        }
    }
    else
    {
        my $filesPerCore = int($numOfBaseFiles / $size);
        my $startIndex = $rank * $filesPerCore;
        
        for(my $i = 0; $i < $filesPerCore; ++$i)
        {
            $myBaseFiles[$i] = $baseFiles[$startIndex];
            ++$startIndex;
        }
        
        if($rank < $numOfBaseFiles - ($size * $filesPerCore))
        {
            $myBaseFiles[$filesPerCore] = $baseFiles[($size * $filesPerCore) + $rank];
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Process the bases files
#
sub processBaseFiles
{
    my $total = @baseFiles;
    my $n = @myBaseFiles;
    
    &printInfo("Processing $total bases files in total");
    &recordBioStat($n,
    "MY NUM BASES FILES",
    "This parallel instance processed $n bases files");
    &outputProgressBar();
    
    if($logProcessedFiles)
    {
        &logInfo("This parallel instance is processing the following $n bases files");
    }
    
    #go through each bases file
    for(my $i = 0; $i < $n; ++$i)
    {
        &updateProgress($n, $i);
        &processBaseFile($myBaseFiles[$i]);
    }
    
    #done
    &updateProgress($n, $n);
    
    #output some statistics
    my $prefix = "The num of invalid genotype mapping alleles across this parallel"
    . "\n\t\tinstance's SNPs and individuals caused by: \n\t\t\t";
    &recordBioStat($numAllelesBadStdOverMean,
    "NUM BAD MEAN VS STD ALLELES",
    "$prefix at least one of the eight pairs of means and stddevs associated\n"
    . "\t\t\twith each allele being in an invalid cluster in the mean vs std\n"
    . "\t\t\t2d graph, HACKED IN FOR OUR SPECIFIC EXPERIMENT!!!");
    
    &recordBioStat($numAllelesBadBothStrandHIQualScores,
    "NUM BAD HI QUAL ALLELES BOTH STRANDS",
    "prefix both strand HomoIndex quality scores being invalid");
    
    &recordBioStat($numAllelesBadFStrandHIQualScores,
    "NUM BAD HI QUAL ALLELES F STRAND",
    "$prefix the forward strand HomoIndex quality score being invalid");
    
    &recordBioStat($numAllelesBadRStrandHIQualScores,
    "NUM BAD HI QUAL ALLELES R STRAND",
    "$prefix the reverse strand HomoIndex quality score being invalid");
    
    &recordBioStat($numAllelesUnknownBothStrandHIClass,
    "NUM UK ALLELES BOTH STRANDS",
    "$prefix both HomoIndex classifications being unknown (the HomoIndex\n"
    . "\t\t\tvalues are not heterozygous or homozygous)");
    
    &recordBioStat($numAllelesUnknownFStrandHIClass,
    "NUM UK ALLELES F STRAND",
    "$prefix the forward strand HomoIndex classification being unknown");
    
    &recordBioStat($numAllelesUnknownRStrandHIClass,
    "NUM UK ALLELES R STRAND",
    "$prefix the reverse strand HomoIndex classification being unknown");

    &recordBioStat($numAllelesBadHeteroClassBothStrands,
    "NUM BAD HETERO ALLELES BOTH STRANDS",
    "$prefix an invalid heterozygous classification of a HomoIndex value in\n"
    . "\t\t\tboth strands (a set of four readings looked something like the example: 20, 19, 18, 2)");
    
    &recordBioStat($numAllelesBadHeteroClassFStrand,
    "NUM BAD HETERO ALLELES F STRANDS",
    "$prefix an invalid heterozygous classification of a HomoIndex value in\n"
    . "\t\t\tthe forward strand (a set of four readings looked something\n"
    . "\t\t\tlike the example: 20, 19, 18, 2)");
    
    &recordBioStat($numAllelesBadHeteroClassRStrand,
    "NUM BAD HETERO ALLELES R STRANDS",
    "$prefix an invalid heterozygous classification of a HomoIndex value in\n"
    . "\t\t\tthe reverse strand (a set of four readings looked something\n"
    . "\t\t\tlike the example: 20, 19, 18, 2)");
    
    &recordBioStat($numAllelesHIClassesDontMatch,
    "NUM ALLELES MISMATCH CLASS",
    "$prefix the two strand HomoIndex classifications not matching");
    
    &recordBioStat($numAllelesHomoDontMatchEitherParent,
    "NUM ALLELES HOMO MISMATCH A AND B",
    "$prefix the SNP nucleotide not matching the A or B parent nucleotide");
    
    &recordBioStat($numAllelesHeteroNuc1MatchesParentANuc2DoesntMatchParentB,
    "NUM ALLELES HETERO NUC2 MISMATCH B",
    "$prefix the nucleotide with the highest oligo luminescent reading matching\n"
    . "\t\t\tthe A parent nucleotide but the nucleotide with the second highest\n"
    . "\t\t\toligo luminescent reading not matching the B parent nucleotide");
    
    &recordBioStat($numAllelesHeteroNuc1MatchesParentANuc2DoesntMatchParentB,
    "NUM ALLELES HETERO NUC2 MISMATCH B",
    "$prefix the nucleotide with the highest oligo luminescent reading matching\n"
    . "\t\t\tthe B parent nucleotide but the nucleotide with the second highest\n"
    . "\t\t\toligo luminescent reading not matching the A parent nucleotide");
    
    &recordBioStat($numAllelesHeteroNuc1DoesntMatchParentAOrB,
    "NUM ALLELES HETERO NUC1 MISMATCH A AND B",
    "$prefix the nucleotide with the highest oligo luminescent reading not\n"
    . "\t\t\tmatching the A or B parent nucleotide");
    
    &recordBioStat($numAllelesComplimentsDontMatch,
    "NUM ALLELES COMPLIMENTS MISMATCH",
    "$prefix the nucleotides in both strands (the two nucleotides with the\n"
    . "\t\t\thighest oligo luminescent readings are taken from each strand\n"
    . "\t\t\tif the SNP is hetero, and only the single nucleotide with the\n"
    . "\t\t\thighest oligo luminescent reading is taken from each strand if\n"
    . "\t\t\tthe SNP is homo) not being each other's compliments");
}

#-------------------------------------------------------------------------------
#
#  Go through a bases file and store all the means and standard deviations for
#  every snp, and then call the function that will compute the
#  homoIndex values, homoIndex classifications, genotypes,
#  homoIndex quality scores, etc.
#
sub processBaseFile
{
    my ($file_name) = @_;
    my @fileParts = split('/', $file_name);
    my $abrev_file_name = $fileParts[(scalar @fileParts) - 1];
    
    if($logProcessedFiles)
    {
        &logInfo("Processing $abrev_file_name");
    }
    
    #attempt to open the base file
    open(BASE, $file_name) or
    &exitEarly("Could not open file: $file_name for input");
    
    #first five lines of every .base file we don't need
    for(my $i = 0; $i < 5; ++$i)
    {
        my $inputLine = <BASE>;
    }
    
    my %meanHash;
    my %stdHash;
    
    #go through every line of the .base file
    while(my $inputLine = <BASE>)
    {
        if($inputLine =~ /^(\w+)\s+\d+\s+[AT]\s+[ACTG]\s+([\d\.]+)\s+([\d\.]+)\s+\d+.*$/)
        {
            $meanHash{$1} .= "$2:";
            $stdHash{$1} .= "$3:";
        }
        else
        {
            my $error = "Incorrectly formatted line in bases file ";
            $error .= "$abrev_file_name: $inputLine";
            &exitEarly($error);
        }
    }
    
    close(BASE);
    &createValues(\%meanHash, \%stdHash, $abrev_file_name);
}

#-------------------------------------------------------------------------------
#
#  Compute homoIndex values, homoIndex classifications, genotype
#  mapping alleles, etc., and output the results
#
sub createValues
{
    my ($mhr, $shr, $file_name) = @_;
    my %meanHash = %{$mhr};
    my %stdHash = %{$shr};
    
    #attempt to open the output file
    my @fileParts = split(/\./, $file_name);
    my $outFile_F = "$outputDir/$fileParts[0]_F.homoIndex";
    my $outFile_R = "$outputDir/$fileParts[0]_R.homoIndex";
    
    open(OUTF, ">$outFile_F") or
    &exitEarly("Could not open file: $outFile_F for output");
    
    open(OUTR, ">$outFile_R") or
    &exitEarly("Could not open file: $outFile_R for output");
    
    #only create values for the snps we care about
    foreach my $snp (keys %$parentSNPs)
    {
        my @snpMeans = split(':', $meanHash{$snp});
        my @snpSTDs = split(':', $stdHash{$snp});
        
        my @means_F = ($snpMeans[0], $snpMeans[1], $snpMeans[2], $snpMeans[3]);
        my @means_R = ($snpMeans[4], $snpMeans[5], $snpMeans[6], $snpMeans[7]);
        my @stds_F = ($snpSTDs[0], $snpSTDs[1], $snpSTDs[2], $snpSTDs[3]);
        my @stds_R = ($snpSTDs[4], $snpSTDs[5], $snpSTDs[6], $snpSTDs[7]);
        
        #TODO this was an experiment to try and fix the means by moving the "A" and
        #"C" clusters in the distribution of means vs. standard deviations
        #to be part of the "B" cluster, described in a word document emailed
        #by dieter
        #&fix_means(\@means_F, \@stds_F);
        #&fix_means(\@means_R, \@stds_R);
        
        #TODO this is hacked in for our specific experiment to try
        #and filter out bad data in the bases files based on a
        #comparison between each data point's mean and std readings.
        #The graphs have very specific clusterings, and we're attempting
        #to throw out the clusters other than the cluster that contains
        #the majority of the data
        my $badStdOverMean = &checkStdsOverMeans(\@snpMeans, \@snpSTDs);
        
        my $sortedNucs_F = &sort4(\@means_F);
        my $sortedNucs_R = &sort4(\@means_R);
        
        my $homoIndex_F = &computeHomoIndex(\@means_F);
        my $homoIndex_R = &computeHomoIndex(\@means_R);
        
        my $class_F;
        my $class_R;
        
        my $genotypeMappingAllele = "";
        
        #quality check
        if($homoIndex_F < 0 || $homoIndex_F > 1 ||
           $homoIndex_R < 0 || $homoIndex_R > 1)
        {
            my $error = "A homoIndex value is out of range in bases file ";
            $error .= "$file_name at SNP $snp: hi_F: $homoIndex_F: ";
            $error .= "hi_R: $homoIndex_R";
            &exitEarly($error);
        }

        #check the means vs stds
        if($badStdOverMean)
        {
            $genotypeMappingAllele = "NA";
            $class_F = $unknownClass;
            $class_R = $unknownClass;
            ++$numAllelesBadStdOverMean;
        }
        #good means vs stds
        else
        {
            #heterozygous, homozygous, or unknown (in the middle of 0 and 1)
            $class_F = &makeCall($homoIndex_F);
            $class_R = &makeCall($homoIndex_R);

            if($class_F eq $unknownClass && $class_R eq $unknownClass)
            {
                $genotypeMappingAllele = "NA";
                ++$numAllelesUnknownBothStrandHIClass;
            }
            elsif($class_F eq $unknownClass)
            {
                $genotypeMappingAllele = "NA";
                ++$numAllelesUnknownFStrandHIClass;
            }
            elsif($class_R eq $unknownClass)
            {
                $genotypeMappingAllele = "NA";
                ++$numAllelesUnknownRStrandHIClass;
            }
            #both HomoIndex values are hetero or homo
            else
            {
                #are any of the hetero classifications invalid,
                #look at the computeExtraHomoIndex comments
                my $badHetero_F = ($class_F eq $heteroClass) &&
                                  (&makeCall(&computeExtraHeteroHomoIndex(\@means_F)) eq $heteroClass);
                my $badHetero_R = ($class_R eq $heteroClass) &&
                                  (&makeCall(&computeExtraHeteroHomoIndex(\@means_R)) eq $heteroClass);
                
                #comment out these two lines, and uncomment
                #the above lines to actually use the second
                #hetero HomoIndex equation
                #my $badHetero_F = $false;
                #my $badHetero_R = $false;

                if($badHetero_F && $badHetero_R)
                {
                    $genotypeMappingAllele = "NA";
                    ++$numAllelesBadHeteroClassBothStrands;
                }
                elsif($badHetero_F)
                {
                    $genotypeMappingAllele = "NA";
                    ++$numAllelesBadHeteroClassFStrand;
                }
                elsif($badHetero_R)
                {
                    $genotypeMappingAllele = "NA";
                    ++$numAllelesBadHeteroClassRStrand;
                }
                #both HomoIndex values are either homo or a valid hetero
                else
                {
                    #check the HomoIndex quality scores
                    my $invalidHIQualF;
                    my $invalidHIQualR;
                    
                    if($class_F eq $heteroClass)
                    {
                        $invalidHIQualF = $hiQualPreFilterOn &&
                                          (1.0 - $homoIndex_F > $hiQualCut);
                    }
                    else
                    {
                        $invalidHIQualF = $hiQualPreFilterOn &&
                                          ($homoIndex_F > $hiQualCut);
                    }
                    
                    if($class_R eq $heteroClass)
                    {
                        $invalidHIQualR = $hiQualPreFilterOn &&
                                          (1.0 - $homoIndex_R > $hiQualCut);
                    }
                    else
                    {
                        $invalidHIQualR = $hiQualPreFilterOn &&
                                          ($homoIndex_R > $hiQualCut);
                    }

                    if($invalidHIQualF && $invalidHIQualR)
                    {
                        $genotypeMappingAllele = "NA";
                        ++$numAllelesBadBothStrandHIQualScores;
                    }
                    elsif($invalidHIQualF)
                    {
                        $genotypeMappingAllele = "NA";
                        ++$numAllelesBadFStrandHIQualScores;
                    }
                    elsif($invalidHIQualR)
                    {
                        $genotypeMappingAllele = "NA";
                        ++$numAllelesBadRStrandHIQualScores;
                    }
                    #everything good so far
                    else
                    {
                        #A, B, H, or NA?
                        $genotypeMappingAllele = &genotypeMappingClassification(
                                                 $class_F, $class_R,
                                                 $sortedNucs_F, $sortedNucs_R,
                                                 $snp);
                    }
                }
            }
        }

        #output the data, not all of it is used by the next
        #step, but maybe it will be in the future
        print OUTF "$snp\t$sortedNucs_F\t$homoIndex_F\t$class_F\t$genotypeMappingAllele\n";
        print OUTR "$snp\t$sortedNucs_R\t$homoIndex_R\t$class_R\t$genotypeMappingAllele\n";
    }
    
    close(OUTF);
    close(OUTR);
}

#-------------------------------------------------------------------------------
#
#  Return true or false saying whether any of the data points fall in the bad
#  std over mean clusters
#
#  TODO hacked in for the mean vs std fix
#
sub checkStdsOverMeans
{
    my ($m, $s) = @_;
    my @means = @{$m};
    my @stds = @{$s};
    my $zero = 0.00000000001;

    if(scalar @means != 8 || scalar @stds != 8)
    {
        my $error = "Incorrect num of means or stds passed to 'checkStdsOverMeans'";
        &exitEarly($error);
    }
    
    #go through each mean and std
    for(my $i = 0; $i < 8; ++$i)
    {
        my $mean = $means[$i];
        my $std = $stds[$i];
        
        #two lines that define the acceptable values
        my $maxSTD = 0.7 * $mean + 2500;
        my $minSTD = 0.6 * $mean - 5000;
        
        if($std > 25000)
        {
            return $true;
        }
        elsif($mean > 40000)
        {
            return $true;
        }
        elsif($std > $maxSTD)
        {
            return $true;
        }
        elsif($std < $minSTD)
        {
            return $true;
        }
    }
    
    #all of them are good
    return $false;
}

#-------------------------------------------------------------------------------
#
#  Based on a snp's classification (hetero or homo), the snp's
#  nucleotides, and the parents' nucleotides at that snp, make a matching
#  classification
#
sub genotypeMappingClassification
{
    my ($class_F, $class_R, $nucs_F, $nucs_R, $snp) = @_;
    
    if(length($nucs_F) != 4 || length($nucs_R) != 4)
    {
        my $error = "One of the \"order\" of nucleotides passed to ";
        $error .= "genotypeMappingClassification was not of length 4, which ";
        $error .= "is required";
        &exitEarly($error);
    }
    
    my $nuc1_F = substr($nucs_F, 0, 1);
    my $nuc2_F = substr($nucs_F, 1, 1);
    my $nuc1_R = substr($nucs_R, 0, 1);
    my $nuc2_R = substr($nucs_R, 1, 1);
    
    my $parentA_nuc = $parentSNPs -> {$snp} -> {"A"};
    my $parentB_nuc = $parentSNPs -> {$snp} -> {"B"};
    
    #sanity check
    if($class_F eq $unknownClass || $class_R eq $unknownClass)
    {
        my $error = "One of the HomoIndex classifications passed to ";
        $error .= "genotypeMappingClassification was the unknown ";
        $error .= "classification, which is invalid";
        &exitEarly($error);
    }
    #classifications don't match
    elsif($class_F ne $class_R)
    {
        ++$numAllelesHIClassesDontMatch;
        return "NA";
    }
    #both calls are homozygote, and the F and R strands have
    #compliment nucleotides
    elsif($class_F eq $homoClass && $nuc1_F eq &complimentNuc($nuc1_R))
    {
        #return the homo nucleotide
        if($genotypeOption == $nucOption)
        {
            return "$nuc1_F";
        }
        #A homozygote
        elsif($nuc1_F eq $parentA_nuc)
        {
            return "A";
        }
        #B homozygote
        elsif($nuc1_F eq $parentB_nuc)
        {
            return "B";
        }
        #NA
        else
        {
            ++$numAllelesHomoDontMatchEitherParent;
            return "NA";
        }
    }
    #both calls are heterozygote, and both the first 2 nucs in both
    #strands are compliments of eachother
    elsif($class_F eq $heteroClass &&
          (($nuc1_F eq &complimentNuc($nuc1_R) &&
            $nuc2_F eq &complimentNuc($nuc2_R)) ||
           ($nuc1_F eq &complimentNuc($nuc2_R) &&
            $nuc2_F eq &complimentNuc($nuc1_R))))
    {
        #return the hetero nucleotides
        if($genotypeOption == $nucOption)
        {
            return "$nuc1_F" . "$nuc2_F";
        }
        elsif($nuc1_F eq $parentA_nuc)
        {
            #H heterozygote
            if($nuc2_F eq $parentB_nuc)
            {
                return "H";
            }
            else
            {
                ++$numAllelesHeteroNuc1MatchesParentANuc2DoesntMatchParentB;
                return "NA";
            }
        }
        elsif($nuc1_F eq $parentB_nuc)
        {
            #H heterozygote
            if($nuc2_F eq $parentA_nuc)
            {
                return "H";
            }
            else
            {
                ++$numAllelesHeteroNuc1MatchesParentBNuc2DoesntMatchParentA;
                return "NA"
            }
        }
        #NA
        else
        {
            ++$numAllelesHeteroNuc1DoesntMatchParentAOrB;
            return "NA";
        }
    }
    #NA - compliments don't match up
    else
    {
        ++$numAllelesComplimentsDontMatch;
        return "NA";
    }
}

#-------------------------------------------------------------------------------
#
#  Take in a string of length one (a nucletodie of A, C, T, or G), and return
#  it's nucleotide compliment
#
sub complimentNuc
{
    my ($nuc) = @_;
    
    if($nuc eq "A")
    {
        "T";
    }
    elsif($nuc eq "T")
    {
        "A";
    }
    elsif($nuc eq "C")
    {
        "G";
    }
    elsif($nuc eq "G")
    {
        "C";
    }
    else
    {
        &exitEarly("Invalid nucleotide passed to 'complimentNuc': $nuc");
    }
}

#-------------------------------------------------------------------------------
#
#  Take in a homoIndex value, return the classification
#
sub makeCall
{
    my ($hi) = @_;
    
    if($hi < $heteroCut && $hi > $homoCut)
    {
        return $unknownClass;
    }
    elsif($hi <= $homoCut)
    {
        return $homoClass;
    }
    elsif($hi >= $heteroCut)
    {
        return $heteroClass;
    }
    else
    {
        &exitEarly("Invalid HomoIndex value passed to 'makeCall': $hi");
    }
}

#-------------------------------------------------------------------------------
#
#  Take in an array of 4 rankings, sorted from greatest to least, return a
#  homoIndex value
#
sub computeHomoIndex
{
    my ($rr) = @_;
    my @ranks = @{$rr};
    my $numRanks = @ranks;
    my $zero = 0.00000001;
    
    if($numRanks != 4)
    {
        my $error = "The num of ranks passed to computeHomoIndex was not ";
        $error .= "equal to 4, which is required";
        &exitEarly($error);
    }
    
    if($ranks[0] < $ranks[1] || $ranks[1] < $ranks[2] || $ranks[2] < $ranks[3])
    {
        my $error = "The ranks passed to computeHomoIndex were in the ";
        $error .= "incorrect order";
        &exitEarly($error);
    }
    
    my $homoIndex = ($ranks[1] - $ranks[3]) / (($ranks[0] - $ranks[3]) + $zero);
    
    #my $homoIndex = ($ranks[1]-(($ranks[2]+$ranks[3])/2)) / (($ranks[0]-(($ranks[2]+$ranks[3])/2))+$zero);
    
    $homoIndex;
}

#-------------------------------------------------------------------------------
#
#  The original HomoIndex equation as of 04-11-2012 will classify the following
#  example four readings as hetero, which is undesirable
#
#  20, 19, 18, 3
#
#  So this test will look at the four values again, and should be called after
#  the initial HomoIndex equation classifies a set of readings as hetero.  The
#  returned value should be classified as hetero, homo, or unknown again, but
#  if the value is hetero, the set of four readings should be marked as invalid
#  because it means the second and third readings are too close together, if
#  it's not hetero, than the readings are fine, and use the original HomoIndex
#  value throughout the rest of the pipeline
#
sub computeExtraHeteroHomoIndex
{
    my ($rr) = @_;
    my @ranks = @{$rr};
    my $numRanks = @ranks;
    my $zero = 0.00000001;
    
    if($numRanks != 4)
    {
        my $error = "The num of ranks passed to computeExtraHeteroHomoIndex ";
        $error .= "was not equal to 4, which is required";
        &exitEarly($error);
    }
    
    if($ranks[0] < $ranks[1] || $ranks[1] < $ranks[2] || $ranks[2] < $ranks[3])
    {
        my $error = "The ranks passed to computeExtraHeteroHomoIndex were in ";
        $error .= "the incorrect order";
        &exitEarly($error);
    }
    
    my $secondHomoIndex = ($ranks[2] - $ranks[3]) /
                         (($ranks[1] - $ranks[3]) + $zero);
    
    #my $secondHomoIndex = ($ranks[2] - (($ranks[2] + $ranks[3]) / 2)) /
    #                      (($ranks[1] - (($ranks[2] + $ranks[3]) / 2)) + $zero);
    
    $secondHomoIndex;
}

#-------------------------------------------------------------------------------
# Sort an array of 4 numbers -- in place
#      Assume the numbers represent nucleotides in order ACGT
#      return a string with A, C, G, T in the order reflected by the
#      sorted array.  Array is passed by reference, so it will then be
#      sorted
#
sub sort4()
{
   my ( $nums ) = @_;
   
   my @nucOrder = ( "A", "C", "G", "T" );
   
   # dumb selection sort:
   for ( my $i = 0; $i < 3; $i++ )
   {
      my $max = $i;
      for ( my $k = $i + 1; $k < 4; $k++ )
      {
         if ( $$nums[ $k ] > $$nums[ $max ] )
         {
            $max = $k;
         }
      }
      my $tmp = $$nums[ $max ];
      $$nums[ $max ] = $$nums[ $i ];
      $$nums[ $i ] = $tmp;
         
      my $tmpNuc  = $nucOrder[ $max ];
      $nucOrder[ $max ] = $nucOrder[ $i ];
      $nucOrder[ $i ] = $tmpNuc; 
   }
   my $nucString = join( "", @nucOrder );
   return $nucString;
}

#-------------------------------------------------------------------------------
#
#  TODO This was hacked in when atempting to fix bad data points based on the
#  mean vs standard deviation relationship, but I don't think it's used anymore
#
sub fix_means
{
    my ($means, $stds) = @_;
    
    for(my $i = 0; $i < scalar @$means; ++$i)
    {
        my $m = $$means[$i];
        my $s = $$stds[$i];
        my $slope = $s / $m;
        
        if($slope < 0.4)
        {
            $$means[$i] = $m / 2;
        }
        elsif($slope > 0.7)
        {
            $$means[$i] = $m * 3;
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Output a progress bar for the user to see, only do it if this is parallel
#  instance that should handle it
#
#  Note that this bypasses the log and print methods below that the rest of the
#  script uses
#
sub outputProgressBar
{
    if($handleUserOutput)
    {
        my $bar = "0% ";
        
        for(my $i = 0; $i < $numPartsInProgressBar; ++$i)
        {
            $bar .= "_";
        }
        
        $bar .= " 100%";
        
        print STDERR "[$processTag: $infoTag]\t$bar\n";
        print STDERR "[$processTag: $infoTag]\t   ";
        $progressPartsOutputSoFar = 0;
    }
}

#-------------------------------------------------------------------------------
#
#  Update the progress bar, this decides whether to update the progress bar,
#  pass in the total number of things that need to be done, and the number of
#  things completed so far
#
#  Note that this bypasses the log and print methods below that the rest of the
#  script uses
#
sub updateProgress
{
    if($handleUserOutput)
    {
        my ($total, $completed) = @_;
        
        my $stepSize = 100 / $numPartsInProgressBar;
        my $percentage = ($completed / $total) * 100;
        
        while($progressPartsOutputSoFar * $stepSize < $percentage)
        {
            print STDERR ">";
            ++$progressPartsOutputSoFar;
        }
        
        if($completed >= $total)
        {
            print STDERR "\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Make sure the appropriate command line arguments have been passed to this
#  program, if not, print out the usage message and quit
#
sub checkUsage
{
    if(scalar @ARGV != 11)
    {
        print STDERR "$usage_message\n\n";
        exit;
    }
}

#-------------------------------------------------------------------------------
#
#  Set variables to either their defaults or their inputted command line
#  argument
#
sub getArguments
{
    #these arguments are not optional, therefore no defaults possible
    $logFile = $ARGV[0];
    $statusFile = $ARGV[1];
    $snpFile = $ARGV[2];
    $basesFilesDir = $ARGV[3];
    $hiQualCut = $ARGV[4];
    $heteroCut = $ARGV[5];
    $homoCut = $ARGV[6];
    $hiQualPreFilterOn = $ARGV[7];
    $genotypeOption = $ARGV[8];
    $logProcessedFiles = $ARGV[9];
    $outputDir = $ARGV[10];
}

#-------------------------------------------------------------------------------
#
#  Print some info to the screen for the user to see, note that only one
#  parallel instance is allowed to print info the screen, info is also recorded
#  in the log
#
sub printInfo
{
    my ($message) = @_;
    my @messageLines = split("\n", $message);
    
    foreach my $line (@messageLines)
    {
        print LOG "[USER $infoTag]\t$line\n";
        
        if($handleUserOutput)
        {
            print STDERR "[$processTag: $infoTag]\t$line\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print error messages for the user to see, also records the messages in the
#  log, note that only one parallel instance is allowed to print info to the
#  screen
#
sub printError
{
    my ($message) = @_;
    my @messageLines = split("\n", $message);
    
    foreach my $line (@messageLines)
    {
        print LOG "[USER $errorTag]\t$line\n";
        
        if($handleUserOutput)
        {
            print STDERR "[$processTag: $errorTag]\t$line\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print messages to the log
#
sub logInfo
{
    my ($message) = @_;
    my @messageLines = split("\n", $message);
    
    foreach my $line (@messageLines)
    {
        print LOG "[LOG]\t$line\n";
    }
}

#-------------------------------------------------------------------------------
#
#  Record biology statistics, pass in a value for the statistic, pass in a tag,
#  and pass in a descriptor
#
sub recordBioStat
{
    my ($value, $tag, $desc) = @_;
    
    $bioStatsString .= "[$tag]\t$value - $desc\n";
}

#-------------------------------------------------------------------------------
#
#  Record performance statistics, pass in a value for the statistic, pass in a
#  tag, and pass in a descriptor
#
sub recordPerfStat
{
    my ($value, $tag, $desc) = @_;
    
    $perfStatsString .= "[$tag]\t$value - $desc\n";
}

#-------------------------------------------------------------------------------
#
#  Exit due to an error, close all open files and update the status file with an
#  error, also output the statistics to the status file
#
sub exitEarly
{
    my ($message) = @_;
    
    print STATUS "$errorTag\n$message\n$perfStatsString";
    print STATUS "$bioTag\n$bioStatsString";
    print LOG "[$errorTag]\t$message\n";
    print STDERR "[$errorTag]\t$message\n";  # rdb added
    
    close(LOG);
    close(STATUS);
    close(SNPINPUT);
    close(BASE);
    close(OUTF);
    close(OUTR);
    close(TEMP);
    
    exit;
}

#-------------------------------------------------------------------------------
#
#  Finish the script by outputting success to the log and the status files, and
#  close both files, should be the last thing called, and then the program
#  should exit normally, statistics will also be output to the status file
#
sub exitSuccessfully
{
    my ($message) = @_;
    
    print STATUS "$successTag\n$message\n$perfStatsString";
    print STATUS "$bioTag\n$bioStatsString";
    print LOG "[$successTag]\t$message\n";
    
    close(LOG);
    close(STATUS);
    close(SNPINPUT);
    close(BASE);
    close(OUTF);
    close(OUTR);
    close(TEMP);
}
