#!/usr/bin/perl
#
#  Brian Albere
#
#  The SpatialSmoothing step is parallel, smooths a subset of the bases files
#  based on what rank this parallel instance is.  It also creates a status file
#  to communicate back up the parallel stack the success of this process and
#  its statistics
#
#  Output is a bunch of base files in the same format as they came in, but with
#  slightly different standard deviation and mean values
#
#  NOTE - info messages are printed to STDERR for the user to see because
#  anything printed to STDOUT during MPI is not put to the screen, but STDERR
#  is...In other words, not everything printed to STDERR in this file is
#  necessarily an error...This may not be entirely true, it could be that STDOUT
#  is being buffered and everything would be printed eventually, but STDERR
#  seems to work well, so just keeping this
#
#  Edit history
#  06/05/12 received from Brian.
#  06/12/12 rdb
#           -changed exitEarly to also send error message to STDERR
#           -handle uniprocessor case, where MPI env variables aren't set

use warnings;
use strict;

#variables----------------------------------------------------------------------

my $true = 1;
my $false = 0;
my $processTag = "SPATIALSMOOTH";
my $infoTag = "INFO";
my $errorTag = "ERROR";
my $successTag = "SUCCESS";
my $bioTag = "BIOLOGY";
my $usage_message = "
Usage: SpatialSmooth.pl [logFile] [statusFile] [parentAlleleFile] [ndfFile]
                        [basesFiles] [xSlideDimensions] [ySlideDimensions]
                        [numZones] [smoothNumber] [logProcessedFiles]
                        [outputDirectory]

    logFile             The name of the file where log information should be
                        stored during this script, this is only the beginning of
                        the file name, the hostname and the mpi rank will be
                        added to the name along with .log
    statusFile          The name of a status file where this script will put
                        information about the overall success of its execution,
                        this is only the beginning of the file name, the
                        hostname and the mpi rank will be added to the name
                        along with .status
    parentAlleleFile    A listing of the SNPs this experiment is dealing with,
                        along with the SNP nucleotides in the parent clones
    ndfFile             The NDF NimbleGen file
    basesFiles          The directory of bases files to spatially smooth
    xSlideDimensions    The x dimensions of the NimbleGen slides, must be evenly
                        divided by the square root of the numZones parameter
    ySlideDimensions    The y dimensions of the NimbleGen slides, must be evenly
                        divided by the square root of the numZones parameter
    numZones            The num of zones to divide the slides into during the
                        spatial smooth algorithm, the square root of the number
                        must be a whole number
    smoothNumber        A trivial number that is only used to prevent divisions
                        by zero during the spatial smooth algorithm
    logProcessedFiles   1 ( on )  means log which files this parallel instance
                        handles, 0 ( off )  means don't do this output
    outputDir           The directory to store the resulting smoothed bases
                        files.
    maxSnps             For testing: if >0 only process this many snps
    maxBasesFiles       For testing: if >0 only process this many bases files
                                
    NOTE - if any of the \"required\" aspects of the command line arguments are
           not met, the behavior of the program is unknown
           
    NOTE - only the means and stds of the snps we care about ( listed in the file
           designated by the parentAlleleFile )  are smoothened, any other lines
           in the .base files that have means and stds are not
           smoothened/changed and don't affect the smoothing
           
    NOTE - this has not been tested with input files of small length, it has
           been used with .bases and .ndf files with thousand of entries, the
           behavior of the program with small files is unknown
                                    
    Program Execution:
        1.  Read in the list of snp ids and positions that this program should
            care about from the parentAlleleFile, store them in a hash
        2.  Read in the ( x,y )  coordinates of every snp that we care about, store
            them in a hash
        3.  For every .bases file
            a.  For every line of the file
                i.  if the line is of a snp that the program should care about,
                    store the mean and std intensity in a hash, store everything
                    else on the line in a hash
            b.  For every zone
                i.  compute numbers for that zone based on the mean intensities
                    inputted for this file
            c.  For every snp the program cares about
                i.  alter the snp's mean and std intensities according to
                    computations that take into account the snp's distance from
                    all other zones and each zone's computed numbers
                ii. store the new mean, std, and everything else needed for that
                    snp in a .base file in a hash
            d.  Output the new .bases file";

#command line arguments
my $logFile;
my $statusFile;
my $snpFile;
my $ndfFile;
my $baseDir;
my $outputDir;
my $xSlideDimensions;
my $ySlideDimensions;
my $numOfZones;
my $smoothNumber;
my $logProcessedFiles;
my $maxSnps;
my $maxBasesFiles;

#other variables
my $rank;
my $size;
my $hostname;
my @baseFiles;
my @myBaseFiles;
my $zonesPerSide;
my $bioStatsString = "";
my $perfStatsString = "";
my $handleUserOutput = $false;
my $numPartsInProgressBar = 50;
my $progressPartsOutputSoFar = 0;
my $averageChangeOverall = 0;
my $numReadingsOverall = 0;

#all the keys in this hash will be the
#snp identifiers the program cares about, all
#the values will be dummy values, a snp
#identifier is snpContig_snpPosition
my %snpNames;

#the hash that stores all the x:y coordinates
#for every nucleotide / oligo of each strand of
#every snp, key is
#snpContig_snpPosition_strand_nucleotide
#note that the snpContig and the snpPosition combine
#to make the snp Identifier
my %snpCoordinates;

#program execution--------------------------------------------------------------

#are there valid command line arguments?
&checkUsage(); 
&getArguments(); 

#assumes these environment variables have been set
$rank = $ENV{MPI_RANK};
$size = $ENV{MPI_SIZE};
$hostname = $ENV{MPI_HOSTNAME};

if ( !defined( $hostname ))     
{
   $hostname = "local";   # executing locally ( not in parallel ) 
   $rank = 0;
   $size = 1;
}
#open the log file and set up a header
&initializeLogFile(); 

#open the status file
$statusFile = $statusFile . "_$hostname" . "_$rank.status";

if( !open( STATUS, ">$statusFile" ))   
{
    &printError( "SpatialSmooth: can't open status file: $statusFile, quitting" ); 
    close( LOG ); 
    exit;
}

#set other global variables if possible
$zonesPerSide = sqrt( $numOfZones ); 

#attempt to open files for input
open( SNPINPUT, $snpFile )  
        or &exitEarly( "SpatialSmooth: can't open snp file: $snpFile for input" ); 

open( NDFINPUT, $ndfFile )  
        or &exitEarly( "SpatialSmooth: can't open ndf file: $ndfFile for input" ); 

#does the base files directory exists?
opendir( TEMP, $baseDir )  
        or &exitEarly( "SpatialSmooth: can't open directory $baseDir for input" ); 

close( TEMP ); 
@baseFiles = glob( $baseDir . "/*.bases" ); 

if ( $maxBasesFiles > 0  )
{
    while ( $maxBasesFiles < @baseFiles )
    {
        pop( @baseFiles );
    }
}

#randomize which parallel instance handles the user output
my @timeData = localtime( time ); 
my $day = $timeData[3];
srand( $day ); 

my $randIndex;
if( $size > scalar @baseFiles ) 
{
    $randIndex = scalar @baseFiles;
}
else
{
    $randIndex = $size;
}

my $i = int( rand( $randIndex )) ; 
if( $i == $rank ) 
{
    $handleUserOutput = $true;
    &printInfo( "The parallel instance handling the user output: $rank" ); 
}

#does the output directory exist?
if( !opendir( TEMP, $outputDir ))   
{
    &logInfo( "Creating directory: $outputDir" ); 
    mkdir( $outputDir ); 
}
else
{
    close( TEMP ); 
}

#is there even anything to smooth?
if( scalar @baseFiles < 1 ) 
{
    &exitEarly( "There are no bases files in the input directory" ); 
}

#are there any file this parallel instance should handle?
&computeMyFiles(); 

if( scalar @myBaseFiles > 0 ) 
{
    #input SNPs the program cares about
    &inputSnpNames(); 
    close( SNPINPUT ); 

    #input the x:y coordinates
    &inputXYCoordinates(); 
    close( NDFINPUT ); 

    #smooth the .bases files
    &processBaseFiles(); 

    #quality check
    if( $numReadingsOverall !=
       ( scalar keys %snpCoordinates )  * ( scalar @myBaseFiles ))   
    {
        my $error = "The number of readings processed across all bases files ";
        $error .= "does not equal the number of (x,y) coordinates * the num";
        $error .= "of bases files";
        &exitEarly( $error ); 
    }

    #bio stat
    $averageChangeOverall /= $numReadingsOverall;
    my $desc = "The average percentage each reading was increased or ";
    $desc .= "decreased by across all individuals processed by this parallel ";
    $desc .= "instance";
    &recordBioStat( $averageChangeOverall, "AVERAGE CHANGE", $desc ); 

    #successful execution of program
    &exitSuccessfully( "Successfully Spatially Smoothed Files" ); 
}
else
{
    &exitSuccessfully( "No files processed, more cores available than needed" ); 
}

#-------------------------------------------------------------------------------
#
#  Open and initialize the log file for appending
#
sub initializeLogFile
{
    $logFile = $logFile . "_$hostname" . "_$rank.log";
    
    if( !open( LOG, ">>$logFile" ))   
    {
        &printError( "Could not open log file for output: $logFile, quitting" ); 
        exit;
    }

    #output log info about this run
    my @timeData = localtime( time ); 
    my $month = $timeData[4] + 1;
    my $day = $timeData[3];
    my $year = $timeData[5] + 1900;
    my $hour = $timeData[2];
    my $minute = $timeData[1];

    #header info
    print LOG "


-------------------Run on $month/$day/$year at $hour:$minute-------------------
    logFile:\t$logFile
    statusFile:\t$statusFile
    parentAlleleFile:\t$snpFile
    maxSnps:\t$maxSnps
    ndfFile:\t$ndfFile
    basesFiles:\t$baseDir
    maxBasesFiles:\t$maxBasesFiles
    xSlideDimensions:\t$xSlideDimensions
    ySlideDimensions:\t$ySlideDimensions
    numZones:\t$numOfZones
    smoothNumber:\t$smoothNumber
    logProcessedFiles:\t$logProcessedFiles
    outputDir:\t$outputDir\n\n";
}

#-------------------------------------------------------------------------------
#
#  Go through the parent allele file and create an entry in the snpNames table
#  for every snp we care about, where the key is the snp identifier and the
#  value is a dummy value
#
sub inputSnpNames
{
    my $info = "Processing the Parent Allele File to determine SNP ";
    $info .= "Identifiers";
    &printInfo( $info ); 

    #go through every line of snp input 
    #   except: if $maxSnps is specified, quit early
    #
    my $count = 0;
    while( my $inputLine = <SNPINPUT> ) 
    {
        my @lineElements = split( '\t', $inputLine ); 
        my $snpId_position = $lineElements[0] . "_" . $lineElements[1];

        #insert the dummy value
        $snpNames{ $snpId_position } = "dummyValue";
        $count++;
        if ( $maxSnps != 0 && $count >= $maxSnps )
        {
            print STDERR "--------------- SNP subset $maxSnps -------------\n";
            last;
        }
    }
    
    my $numSNPs = scalar keys %snpNames;
    &recordBioStat( $numSNPs, "NUM SNPS",
                    "The num of SNP Identifiers in the Parent Allele File" );
    print STDERR "---------------Processed $count snps------------------\n"; 
}

#-------------------------------------------------------------------------------
#
#  Go through the NDF file and store x:y coordinates for each snp that has an
#  entry in the snpNames hash
#
sub inputXYCoordinates
{
    my $info = "Processing the NDF File to determine oligo ( x,y )  ";
    $info .= "coordinates";
    &printInfo( $info ); 

    #go through every line of input
    while( my $inputLine = <NDFINPUT> ) 
    {
        my @lineElements = split( '\t', $inputLine ); 
        my $snpId_position = $lineElements[4];

        #is this a snp the program cares about?
        if( exists $snpNames{$snpId_position} ) 
        {
            my $strand;
            my $nucleotide;
        
            #get the strand and nucleotide
            if( $lineElements[3] =~ /([AT])\/([atcg])/ ) 
            {
                $strand = $1;
                $nucleotide = $2;
            }
            else
            {
                my $error = "Incorrectly formatted SELECTION_CRITERIA column ";
                $error .= "in the NDF file: $lineElements[3] in\n" 
                          . "$inputLine---------------";
                &exitEarly( $error ); 
            }

            #get the x, y values
            my $x = $lineElements[15];
            my $y = $lineElements[16];
            chomp( $y ); 

            #input into the x:y hash
            my $key = $snpId_position . "_" . $strand . "_" . $nucleotide;
            $snpCoordinates{$key} = $x . ":" . $y;
        }
    }
    
    #are there the correct number of entries?
    my $numOfSnpNames = keys %snpNames;
    my $numOfSnpCoords = keys %snpCoordinates;
    
    if( $numOfSnpNames * 8 != $numOfSnpCoords ) 
    {
        my $error = "# of (x,y) coords in NDF file ($numOfSnpCoords) should ";
        $error .= "be 8 times the # SNPs in parent allele file ($numOfSnpNames); ";
        $error .= "it isn't.";
        
        &exitEarly( $error ); 
    }
}

#-------------------------------------------------------------------------------
#
#  Call this function to fill in the array that contains which files this
#  process should handle since there will be multiple instances of this
#  process running in parallel
#
sub computeMyFiles
{
    my $numOfBaseFiles = @baseFiles;
    
    #more cores than there are files to process
    if( $size > $numOfBaseFiles ) 
    {
        if( $rank < $numOfBaseFiles ) 
        {
            $myBaseFiles[0] = $baseFiles[$rank];
        }
    }
    else
    {
        my $filesPerCore = int( $numOfBaseFiles / $size ); 
        my $startIndex = $rank * $filesPerCore;
        
        for( my $i = 0; $i < $filesPerCore; ++$i ) 
        {
            $myBaseFiles[$i] = $baseFiles[$startIndex];
            ++$startIndex;
        }
        
        if( $rank < $numOfBaseFiles - ( $size * $filesPerCore ))   
        {
            $myBaseFiles[$filesPerCore] = $baseFiles[( $size * $filesPerCore )  + $rank];
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Execute step 3 of the usage message
#
sub processBaseFiles
{    
    #key is x:y
    my %meanIntensities;
    
    #key is x:y
    my %stdIntensities;
    
    #key is x:y
    my %logMeanIntensities;
    
    #key is x:y, values are the elements of
    #a .bases file line, without the mean and std,
    #separated by :
    my %restOfLine;
    
    my $total = @baseFiles;
    my $n = @myBaseFiles;
    
    &printInfo( "Processing $n out of $total bases files in total" ); 
    &recordBioStat( $n,
                    "MY NUM BASES FILES",
                    "This parallel instance processed $n bases files" ); 
    &outputProgressBar(); 
    
    if( $logProcessedFiles ) 
    {
        &logInfo( "This parallel instance is processing the following $n bases files" ); 
    }
    
    #go through every .bases file that needs smoothing,
    #smooth each one independently, they don't affect
    #each other
    for( my $j = 0; $j < $n; $j++ ) 
    {    
        my $fileName = $myBaseFiles[ $j ];
        #print STDERR "-----SpatialSmooth file: $fileName\n";
        my @fileNameElements = split( '/', $fileName ); 
        my $numOfFileNameElements = @fileNameElements;
        my $adjustedFileName = $fileNameElements[ $numOfFileNameElements - 1 ];

        &updateProgress( $n, $j ); 
        
        if( $logProcessedFiles ) 
        {
            &logInfo( "Processing $adjustedFileName" ); 
        }
        
        #attempt to open the file
        open( BASE, $fileName )  or
             &exitEarly( "SpatialSmooth: Can't open: $fileName for input" ); 
        
        #empty out the hashes
        %meanIntensities = (); 
        %stdIntensities = (); 
        %logMeanIntensities = (); 
        %restOfLine = (); 
        
        #there are beginning lines in the file we dont care about
        my @headerLines;
        my $i = 0;
        
        #go through every line of the .base file
        while( my $inputLine = <BASE> ) 
        {
            my @lineElements = split( '\t', $inputLine ); 
            my $snpId_position = $lineElements[0];
            
            #is this a line that we care about?
            if( exists $snpNames{$snpId_position} ) 
            {
                #get the mean and std
                my $mean = $lineElements[4];
                my $std = $lineElements[5];
                
                #get the x:y for this line
                my $coordinatesKey = $snpId_position . "_" .
                                     $lineElements[2] . "_" .
                                     lc( $lineElements[3] ); 
                my $coordinates = $snpCoordinates{$coordinatesKey};
                
                if ( defined( $coordinates )) # rdb added for testing
                {
                    #store the info for the later computations
                    $meanIntensities{$coordinates} = $mean;
                    $stdIntensities{$coordinates} = $std;
                    $logMeanIntensities{$coordinates} = log( $mean )  / log( 2 ); 
                    $restOfLine{$coordinates} = $lineElements[0] . ":" .
                                                $lineElements[1] . ":" .
                                                $lineElements[2] . ":" .
                                                $lineElements[3] . ":" .
                                                $lineElements[6];
                }
            }
            else
            {
                $headerLines[$i] = $inputLine;
                ++$i;
            }
        }
        
        #close the file
        close( BASE ); 
        
        #are there the correct number of entries in each hash?
        my $numOfSnpNames = keys %snpNames;
        
        if( $numOfSnpNames * 8 != keys %snpCoordinates ||
           $numOfSnpNames * 8 != keys %meanIntensities ||
           $numOfSnpNames * 8 != keys %stdIntensities ||
           $numOfSnpNames * 8 != keys %logMeanIntensities ||
           $numOfSnpNames * 8 != keys %restOfLine ) 
        {
            my $error = "The number of SNP entries from .bases file: ";
            $error .= "$adjustedFileName within the the various hashes\n";
            $error .= "used throughout the code do not match.  Examine ";
            $error .= "the .bases file and the code";
            
            &exitEarly( $error ); 
        }
        
        #smooth the file
        &smoothFile( \%logMeanIntensities,\%meanIntensities,\%stdIntensities,
                    \%restOfLine, \@headerLines, $adjustedFileName ); 
    }
    
    #done with all files
    &updateProgress( $n, $n ); 
}

#-------------------------------------------------------------------------------
#
#  Execute step 3-b of the usage message
#
#  Creates three new hashes, commented below, and passes those to another
#  subroutine that actually smooths the means and stds by altering them
#  accordingly
#
sub smoothFile
{
    #variables
    my ( $lmR, $mR, $sR, $lR, $hR, $fileName )  = @_;
    my %logMeanIntensities = %{$lmR};
    
    my $numOfZoneRows = sqrt( $numOfZones ); 
    my $numOfZoneColumns = sqrt( $numOfZones ); 
    my $numOfPixelRowsPerZone = $ySlideDimensions / $numOfZoneRows;
    my $numOfPixelColumnsPerZone = $xSlideDimensions / $numOfZoneColumns;
    
    #key is x:y, an entry for each zone,
    #the median value of all
    #the log mean intensities in the zone, the
    #key is the center x:y of the zone
    my %zoneMedian;
    
    #key is x:y, holds all the
    #log means within a zone, gets
    #wiped clean for each zone
    my %zoneLogMeans;
    
    #go through each row of zones
    for( my $r = 0; $r < $numOfZoneRows; ++$r ) 
    {
        #go through each column of zones
        for( my $c = 0; $c < $numOfZoneColumns; ++$c ) 
        {
            #empty out the hash to get ready for this zone
            %zoneLogMeans = (); 
        
            #store all of the pixel locations' means
            
            #go through each row of pixels in a zone
            for( my $pr = 1; $pr <= $numOfPixelRowsPerZone; ++$pr ) 
            {
                my $absoluteY = $r * $numOfPixelRowsPerZone + $pr;
            
                #go through each columns of pixels in a zone
                for( my $pc = 1; $pc <= $numOfPixelColumnsPerZone; ++$pc ) 
                {
                    my $absoluteX = $c * $numOfPixelColumnsPerZone + $pc;                
                    my $key = $absoluteX . ":" . $absoluteY;
                    
                    if( defined $logMeanIntensities{$key} ) 
                    {
                        $zoneLogMeans{$key} = $logMeanIntensities{$key};
                    }
                }
            }
            
            #only consider this zone if there are any
            #values within it
            if( ( keys %zoneLogMeans )  > 0 ) 
            {            
                #the center of the zone coordinates
                my $absoluteZoneCenterX = int( $c * $numOfPixelColumnsPerZone +
                                              .5 * $numOfPixelColumnsPerZone ); 
                my $absoluteZoneCenterY = int( $r * $numOfPixelRowsPerZone +
                                              .5 * $numOfPixelRowsPerZone ); 
                my $key = $absoluteZoneCenterX . ":" . $absoluteZoneCenterY;
                
                #get and store the median mean of the zone
                my @sortedZoneLogMeans = sort{$a <=> $b} values %zoneLogMeans;
                $zoneMedian{$key} = &computeMedian( \@sortedZoneLogMeans ); 
            }
        }
    }
    
    #compute the median of all the zone medians
    my @sortedZoneLogMedians = sort {$a <= $b} values %zoneMedian;
    my $overallZoneLogMedian = &computeMedian( \@sortedZoneLogMedians ); 

    #all zone data has been computed by here,
    #now actually alter the numbers
    &alterMeansAndStds( $lmR, $mR, $sR, $lR, $hR, $fileName,
                       \%zoneMedian, $overallZoneLogMedian ); 
}

#-------------------------------------------------------------------------------
#
#  Execute step 3-c of the usage message
#
sub alterMeansAndStds
{
    #variables
    my ( $lmR, $mR, $sR, $lR, $hR, $fileName, $medR, $overallZoneLogMedian )  = @_;
    my %logMeanIntensities = %{$lmR};
    my %meanIntensities = %{$mR};
    my %stdIntensities = %{$sR};
    my %restOfLine = %{$lR};
    my %zoneMedian = %{$medR};
    
    #key is snp identifier, the value is the tab
    #separted line to output
    my %adjustedOutput;
    
    my $a = scalar keys %logMeanIntensities;
    my $b = scalar keys %zoneMedian;
    
    #go through every location's mean
    foreach my $key ( keys %logMeanIntensities ) 
    {
        my $logMean = $logMeanIntensities{$key};
        my $mean = $meanIntensities{$key};
        my $std = $stdIntensities{$key};
        my @coordinates = split( ':', $key ); 
        my $x = $coordinates[0];
        my $y = $coordinates[1];
        
        my $totalMedianDifference = 0;
        my $totalWeight;
        
        #go through the medians of each zone ( 1 median per zone ) 
        
        #NOTE - the original code went through these in a sorted order
        #( by location )  but I don't think that is necessary
        foreach my $mKey ( keys %zoneMedian ) 
        {
            my @zoneCoordinates = split( ':', $mKey ); 
            my $zoneCenterX = $zoneCoordinates[0];
            my $zoneCenterY = $zoneCoordinates[1];
            
            my $zoneMed = $zoneMedian{$mKey};
            my $medianDifference = $overallZoneLogMedian - $zoneMed;
            
            my $weight = 1 / ( ( ( ( $zoneCenterX - $x )  ** 2 )   +
                               ( ( $zoneCenterY - $y )  ** 2 ))    + $smoothNumber ); 
            
            $totalMedianDifference += $medianDifference * $weight;
            $totalWeight += $weight;
        }
        
        my $md = $totalMedianDifference / $totalWeight;
        
        #the smoothened mean
        my $logMeanAdjusted = $logMean + $md;
        $logMeanAdjusted = exp( $logMeanAdjusted * log( 2 )) ; 
        $logMeanAdjusted = sprintf( "%.2f", $logMeanAdjusted ); 
        
        #the smoothened std
        my $adjustedStd = ( $logMeanAdjusted / $mean )  * $std;
        $adjustedStd = sprintf( "%.2f", $adjustedStd ); 
        
        #create the new line of output
        my @lineElements = split( ':', $restOfLine{$key} ); 
        my $outputKey = $lineElements[0] . $lineElements[2] . $lineElements[3];
        $adjustedOutput{$outputKey} = $lineElements[0] . "\t" .
                                      $lineElements[1] . "\t" .
                                      $lineElements[2] . "\t" .
                                      $lineElements[3] . "\t" .
                                      $logMeanAdjusted . "\t" .
                                      $adjustedStd     . "\t" .
                                      $lineElements[4];
        
        #statistics
        my $percentChange = abs( ( $logMeanAdjusted / $mean )  - 1 ); 
        $averageChangeOverall += $percentChange;
        ++$numReadingsOverall;
    }
    
    #everything has been smoothened, print the new file
    &printAdjustedBaseFile( \%adjustedOutput, $hR, $fileName ); 
}

#-------------------------------------------------------------------------------
#
#  Print the new values to a new file
#
sub printAdjustedBaseFile
{
    my ( $aR, $hR, $fileName )  = @_;
    my %adjustedOutput = %{$aR};
    my @headerLines = @{$hR};
    
    #there should be the same number of output lines as input lines
    my $numOfOutputLines = keys %adjustedOutput;
    my $numOfSnpNames = keys %snpNames;
    
    if( $numOfSnpNames * 8 != $numOfOutputLines ) 
    {
        my $error = "The number of output lines does not match the ";
        $error .= "number of input lines for .bases file: $fileName";
        &exitEarly( $error ); 
    }
    
    #whats the new file name
    my @fileElements = split( /\./, $fileName ); 
    $fileName = "$outputDir/$fileElements[0]_smoothed.$fileElements[1]";
    
    #attempt to open the file for output
    open( OUT, ">$fileName" )  or
    &exitEarly( "Could not open file: $fileName for output" ); 
    
    #print out the header lines
    foreach my $line ( @headerLines ) 
    {
        print OUT $line;
    }
    
    #print the adjusted output
    foreach my $key ( sort keys %adjustedOutput ) 
    {
        print OUT $adjustedOutput{$key};
    }
    
    close( OUT ); 
}

#-------------------------------------------------------------------------------
#
#  Returns the median of the sorted ascending array of numbers passed in
#
sub computeMedian
{
    my ( $aR )  = @_;
    my @numbers = @{$aR};
    my $numOfEntries = @numbers;
    
    if( $numOfEntries % 2 == 1 ) 
    {
        $numbers[( ( $numOfEntries + 1 )  / 2 )  - 1];
    }
    else
    {                
        my $med1 = $numbers[( $numOfEntries / 2 )  - 1];
        my $med2 = $numbers[( ( $numOfEntries + 2 )  / 2 )  - 1];
        ( $med1 + $med2 )  / 2;
    }
}

#-------------------------------------------------------------------------------
#
#  Output a progress bar for the user to see, only do it if this is parallel
#  instance that should handle it
#
#  Note that this bypasses the log and print methods below that the rest of the
#  script uses
#
sub outputProgressBar
{
    if( $handleUserOutput ) 
    {
        my $bar = "0% ";
        
        for( my $i = 0; $i < $numPartsInProgressBar; ++$i ) 
        {
            $bar .= "_";
        }
        
        $bar .= " 100%";
        
        print STDERR "[$processTag: $infoTag]\t$bar\n";
        print STDERR "[$processTag: $infoTag]\t   ";
        $progressPartsOutputSoFar = 0;
    }
}

#-------------------------------------------------------------------------------
#
#  Update the progress bar, this decides whether to update the progress bar,
#  pass in the total number of things that need to be done, and the number of
#  things completed so far
#
#  Note that this bypasses the log and print methods below that the rest of the
#  script uses
#
sub updateProgress
{
    if( $handleUserOutput ) 
    {
        my ( $total, $completed )  = @_;
        
        my $stepSize = 100 / $numPartsInProgressBar;
        my $percentage = ( $completed / $total )  * 100;
        
        while( $progressPartsOutputSoFar * $stepSize < $percentage ) 
        {
            print STDERR ">";
            ++$progressPartsOutputSoFar;
        }
        
        if( $completed >= $total ) 
        {
            print STDERR "\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Make sure the appropriate command line arguments have been passed to this
#  program, if not, print out the usage message and quit
#
sub checkUsage
{
    if( scalar @ARGV != 13 ) 
    {
        print STDERR "$usage_message\n\n";
        exit;
    }
}

#-------------------------------------------------------------------------------
#
#  Set variables to either their defaults or their inputted command line
#  argument
#
sub getArguments
{
    #these arguments are not optional, therefore no defaults possible
    $logFile = $ARGV[0];
    $statusFile = $ARGV[1];
    $snpFile = $ARGV[2];
    $ndfFile = $ARGV[3];
    $baseDir = $ARGV[4];
    $xSlideDimensions = $ARGV[5];
    $ySlideDimensions = $ARGV[6];
    $numOfZones = $ARGV[7];
    $smoothNumber = $ARGV[8];
    $logProcessedFiles = $ARGV[9];
    $outputDir = $ARGV[10];
    $maxSnps = $ARGV[11];
    $maxBasesFiles = $ARGV[12];
}

#-------------------------------------------------------------------------------
#
#  Print some info to the screen for the user to see, note that only one
#  parallel instance is allowed to print info the screen, info is also recorded
#  in the log
#
sub printInfo
{
    my ( $message )  = @_;
    my @messageLines = split( "\n", $message ); 
    
    foreach my $line ( @messageLines ) 
    {
        print LOG "[USER $infoTag]\t$line\n";
        
        if( $handleUserOutput ) 
        {
            print STDERR "[$processTag: $infoTag]\t$line\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print error messages for the user to see, also records the messages in the
#  log, note that only one parallel instance is allowed to print info to the
#  screen
#
sub printError
{
    my ( $message )  = @_;
    my @messageLines = split( "\n", $message ); 
    
    foreach my $line ( @messageLines ) 
    {
        print LOG "[USER $errorTag]\t$line\n";
        
        if( $handleUserOutput ) 
        {
            print STDERR "[$processTag: $errorTag]\t$line\n";
        }
    }
}

#-------------------------------------------------------------------------------
#
#  Print messages to the log
#
sub logInfo
{
    my ( $message )  = @_;
    my @messageLines = split( "\n", $message ); 
    
    foreach my $line ( @messageLines ) 
    {
        print LOG "[LOG]\t$line\n";
    }
}

#-------------------------------------------------------------------------------
#
#  Record biology statistics, pass in a value for the statistic, pass in a tag,
#  and pass in a descriptor
#
sub recordBioStat
{
    my ( $value, $tag, $desc )  = @_;
    
    $bioStatsString .= "[$tag]\t$value - $desc\n";
}

#-------------------------------------------------------------------------------
#
#  Record performance statistics, pass in a value for the statistic, pass in a
#  tag, and pass in a descriptor
#
sub recordPerfStat
{
    my ( $value, $tag, $desc )  = @_;
    
    $perfStatsString .= "[$tag]\t$value - $desc\n";
}

#-------------------------------------------------------------------------------
#
#  Exit due to an error, close all open files and update the status file with an
#  error, also output the statistics to the status file
#
sub exitEarly
{
    my ( $message )  = @_;
    
    print STATUS "$errorTag\n$message\n$perfStatsString";
    print STATUS "$bioTag\n$bioStatsString";
    print LOG "[$errorTag]\t$message\n";
    print STDERR "[$errorTag]\t$message\n";  # rdb added
    
    close( LOG ); 
    close( STATUS ); 
    close( SNPINPUT ); 
    close( NDFINPUT ); 
    close( TEMP ); 
    close( BASE ); 
    close( OUT ); 
    
    exit;
}

#-------------------------------------------------------------------------------
#
#  Finish the script by outputting success to the log and the status files, and
#  close both files, should be the last thing called, and then the program
#  should exit correctly, statistics will also be output to the status file
#
sub exitSuccessfully
{
    my ( $message )  = @_;
    
    print STATUS "$successTag\n$message\n$perfStatsString";
    print STATUS "$bioTag\n$bioStatsString";
    print LOG "[$successTag]\t$message\n";
    
    close( LOG ); 
    close( STATUS ); 
    close( SNPINPUT ); 
    close( NDFINPUT ); 
    close( TEMP ); 
    close( BASE ); 
    close( OUT ); 
}
